/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.module;

import com.revrobotics.CANSparkMax;
import com.revrobotics.CANSparkMax.IdleMode;
import com.revrobotics.CANSparkMaxLowLevel.MotorType;

import net.bancino.log.Log;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.encoder.AnalogEncoder;
import net.bancino.robotics.swerveio.encoder.Encoder;
import net.bancino.robotics.swerveio.encoder.SparkMaxEncoder;

import net.bancino.robotics.swerveio.geometry.Length;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

/**
 * A swerve module implementation that uses RevRobotics Neo motors and Spark Max
 * motor controllers. This was designed for Swerve Drive Specialties' MK2
 * Module.
 * 
 * <p>
 * To use this module implementation, you must have the following hardware
 * setup:
 * </p>
 * <ul>
 * <li>2 RevRobotics' NEO brushless motors</li>
 * <li>8.33:1 gear ratio drive gearbox</li>
 * <li>4-inch wheels</li>
 * <li>A 1:1 (analog) encoder</li>
 * </ul>
 * 
 * <p>
 * This implementation <b>requires</b> a 1:1 encoder. You <b>must</b> attach an
 * encoder to the dedicated location on the MK2 kit in order to use this module
 * implementation. The Neo internal encoders are absolute garbage.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.1
 * @since 1.0.0
 */
@Loggable
public class MK2SwerveModule extends GenericSwerveModule {

    private final CANSparkMax pivotMotor;

    /**
     * The swerve module is constructed to allow the pivot motor to coast, this
     * allows for adjustments, but as soon as the module is driven, it switches to
     * brake mode to prevent outside modifications.
     */
    @Log
    private boolean setPivotIdleMode = false;

    /**
     * Create a new MK2 swerve module composed of Neo brushless motors that use the
     * Spark Max motor controllers.
     * 
     * @param driveCanId   The CAN ID of the drive motor controller for this module.
     * @param pivotCanId   The CAN ID of the pivot motor controller for this module.
     * @param pivotEncoder The encoder to use as the pivot encoder.
     */
    public MK2SwerveModule(int driveCanId, int pivotCanId, Encoder pivotEncoder) {
        super(new GenericSwerveModule.Builder()
                .setDriveMotor(new CANSparkMax(driveCanId, MotorType.kBrushless),
                (GenericSwerveModule.Builder builder, CANSparkMax motor) -> {
                    motor.setIdleMode(IdleMode.kBrake);
                    builder.setDriveEncoder(new SparkMaxEncoder(motor));
                }).setPivotMotor(new CANSparkMax(pivotCanId, MotorType.kBrushless),
                        (GenericSwerveModule.Builder builder, CANSparkMax motor) -> {
                            motor.setIdleMode(IdleMode.kCoast); // This is set to brake before we drive
                            builder.setPivotEncoder(pivotEncoder);
                        })
                // Constants for the MK2 kit and Neo
                .setDriveGearRatio(8.33).setDriveMaxRPM(5680).setWheelDiameter(new Length(4, Unit.INCHES)));

        pivotMotor = (CANSparkMax) getPivotMotor();
    }

    /**
     * Create a new MK2 swerve module composed of Neo brushless motors that use the
     * Spark Max motor controllers.
     * 
     * @param driveCanId        The CAN ID of the drive motor for this module.
     * @param pivotCanId        The CAN ID of the pivot motor for this module.
     * @param analogEncoderPort The port on the roboRIO that the encoder to use as
     *                          the pivot encoder is on.
     */
    public MK2SwerveModule(int driveCanId, int pivotCanId, int analogEncoderPort) {
        this(driveCanId, pivotCanId, new AnalogEncoder(analogEncoderPort));
    }

    @Override
    public void setAngle(double angle) {
        /*
         * Enable brake mode so that the module doesn't move due to external force after
         * starting.
         */
        if (!setPivotIdleMode) {
            pivotMotor.setIdleMode(IdleMode.kBrake);
            setPivotIdleMode = true;
        }
        super.setAngle(angle);
    }
}
