/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.module;

import com.ctre.phoenix.motorcontrol.can.WPI_TalonFX;
import com.ctre.phoenix.sensors.SensorInitializationStrategy;

import net.bancino.robotics.swerveio.pid.PIDController;

import com.ctre.phoenix.motorcontrol.FeedbackDevice;
//import com.ctre.phoenix.motorcontrol.ControlMode;
//import com.ctre.phoenix.motorcontrol.StatusFrameEnhanced;
import com.ctre.phoenix.motorcontrol.NeutralMode;

import net.bancino.robotics.swerveio.encoder.Encoder;
import net.bancino.robotics.swerveio.encoder.PhoenixCANCoder;
import net.bancino.robotics.swerveio.encoder.PhoenixEncoder;
import net.bancino.robotics.swerveio.geometry.Length;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

/**
 * A swerve module implementation that used CTRE Falcon 500 motors and TalonFX
 * motor controllers. This implementation was designed for Swerve Drive
 * Specialties' MK3 Module, and provides support for the internal Falcon 500
 * encoder as well as the external 1:1 encoder recommended by Swerve Drive
 * Specialties.
 *
 * <p>
 * To use this module implementation, you must have the following hardware
 * setup:
 * </p>
 * <ul>
 * <li>2 CTRE Falcon 500 brushless motors</li>
 * <li>8.16:1 gear ratio gearbox</li>
 * <li>4-inch wheels</li>
 * </ul>
 * 
 * <p>
 * Optional hardware:
 * </p>
 * 
 * <ul>
 * <li>CTRE CANCoder or other external encoder</li>
 * </ul>
 *
 * @author Jordan Bancino
 * @version 6.0.2
 * @since 4.2.0
 */
public class MK3SwerveModule extends GenericSwerveModule {

    private final WPI_TalonFX pivotMotor;
    private final WPI_TalonFX driveMotor;
    private boolean setPivotIdleMode = false;

    /**
     * Create a new MK3 swerve module. This constructor assumes a profile slot of
     * zero and that you're using the CTRE CANCoder.
     * 
     * @param driveCanId        The CAN ID of the drive motor controller for this
     *                          module.
     * @param pivotCanId        The CAN ID of the pivot motor controller for this
     *                          module.
     * @param pivotEncoderCanId The CAN ID of the pivot motor encoder (CANCoder).
     */
    public MK3SwerveModule(int driveCanId, int pivotCanId, int pivotEncoderCanId) {
        this(driveCanId, pivotCanId, new PhoenixCANCoder(pivotEncoderCanId));
    }

    /**
     * Create a new MK3 swerve module. This constructor assumes a profile slot of
     * zero and the encoder is 1:1.
     * 
     * @param driveCanId   The CAN ID of the drive motor controller for this module.
     * @param pivotCanId   The CAN ID of the pivot motor controller for this module.
     * @param pivotEncoder The encoder that monitors the pivot motor at a ratio of
     *                     1:1.
     */
    public MK3SwerveModule(int driveCanId, int pivotCanId, Encoder pivotEncoder) {
        this(driveCanId, pivotCanId, pivotEncoder, 1);
    }

    /**
     * Create a new MK3 swerve module. This constructor assumes a profile slot of
     * zero and uses the internal encoder.
     * 
     * @param driveCanId The CAN ID of the drive motor controller for this module.
     * @param pivotCanId The CAN ID of the pivot motor controller for this module.
     */
    public MK3SwerveModule(int driveCanId, int pivotCanId) {
        this(driveCanId, pivotCanId, null, 6);
    }

    /**
     * Create a new MK3 swerve module. This constructor assumes a profile slot of
     * zero.
     * 
     * @param driveCanId     The CAN ID of the drive motor controller for this
     *                       module.
     * @param pivotCanId     The CAN ID of the pivot motor controller for this
     *                       module.
     * @param pivotEncoder   The encoder that monitors the pivot motor.
     * @param pivotGearRatio The gear ratio between the pivot motor and the pivot
     *                       encoder.
     */
    public MK3SwerveModule(int driveCanId, int pivotCanId, Encoder pivotEncoder, double pivotGearRatio) {
        this(driveCanId, pivotCanId, 0, pivotEncoder, pivotGearRatio);
    }

    /**
     * Create a new MK3 swerve module composed of Falcon 500 motors that use the
     * TalonFX motor controllers.
     * 
     * @param driveCanId     The CAN ID of the drive motor controller for this
     *                       module.
     * @param pivotCanId     The CAN ID of the pivot motor controller for this
     *                       module.
     * @param pidSlot        CTRE motor controllers support multiple PID controller
     *                       slots. Specify the slot you want to use for both the
     *                       drive and pivot encoders, as well as the drive and
     *                       pivot PID controllers. This value is used in this
     *                       module implementation wherever Phoenix requires a slot
     *                       number, so make sure all your motors and encoders use
     *                       the same slot number.
     * @param pivotEncoder   The encoder that monitors the pivot motor.
     * @param pivotGearRatio The gear ratio between the pivot motor and the pivot
     *                       encoder.
     */
    public MK3SwerveModule(int driveCanId, int pivotCanId, int pidSlot, Encoder pivotEncoder, double pivotGearRatio) {
        super(new GenericSwerveModule.Builder().setDriveMotor(new WPI_TalonFX(driveCanId),
                (GenericSwerveModule.Builder builder, WPI_TalonFX motor) -> {
                    motor.configFactoryDefault();
                    builder.setDriveEncoder(new PhoenixEncoder(motor, FeedbackDevice.IntegratedSensor, pidSlot));
                    motor.setNeutralMode(NeutralMode.Brake);
                }).setPivotMotor(new WPI_TalonFX(pivotCanId),
                        (GenericSwerveModule.Builder builder, WPI_TalonFX motor) -> {
                            motor.configIntegratedSensorInitializationStrategy(
                                    SensorInitializationStrategy.BootToAbsolutePosition);
                            motor.setNeutralMode(NeutralMode.Coast); // This is set to brake before we drive
                            if (pivotEncoder != null) {
                                builder.setPivotEncoder(pivotEncoder);
                            } else {
                                builder.setPivotEncoder(
                                        new PhoenixEncoder(motor, FeedbackDevice.IntegratedSensor, pidSlot));
                            }
                            motor.setInverted(true);
                        })
                .setDriveGearRatio(8.16).setPivotGearRatio(pivotGearRatio)
                /* 6380 RPM x 75% = 4785 */
                .setDriveMaxRPM(4785).setWheelDiameter(new Length(3.875, Unit.INCHES)));

        pivotMotor = (WPI_TalonFX) getPivotMotor();
        driveMotor = (WPI_TalonFX) getDriveMotor();

        /* Sensible defaults for the Falcon 500 using the integrated encoder. */
        PIDController drivePIDController = getDrivePIDController();
        drivePIDController.setP(0.00008);
        drivePIDController.setI(0.000007);
        drivePIDController.setD(0.0);
        drivePIDController.setF(0.00002);
        drivePIDController.setOutputLimits(1);
    }

    @Override
    public void setAngle(double angle) {
        /*
         * Enable brake mode so that the module doesn't move due to external force after
         * starting.
         */
        if (!setPivotIdleMode) {
            pivotMotor.setNeutralMode(NeutralMode.Brake);
            setPivotIdleMode = true;
        }
        super.setAngle(angle);
    }

    @Override
    public void setSpeed(double speed) {
        /*
         * The CTRE motor controller expects a velocity in encoder counts per 100ms.
         * This converts RPMS to counts per 100ms based on the number of encoder counts
         * there are in a revolution of the motor shaft, then scales it on the percent
         * output reference we're given.
         * 
         * In hindsight, we should have converted the sensor velocity into RPMs so we
         * could create a unified velocity loop API, but once upon a time we tried to
         * use the motor controller's PID loop, hence this conversion.
         */
        double setpoint = speed * (getDriveMaxRPM() * getDriveEncoder().countsPerRevolution()) / 600.0;
        double feedback = driveMotor.getSelectedSensorVelocity();
        double output = getDrivePIDController().getOutput(feedback, setpoint);

        driveMotor.set(output);
    }

    @Override
    public double getOutputThreshhold() {
        return 0.02;
    }
}
