/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio;

import net.bancino.robotics.swerveio.module.SwerveModule;

import java.util.Map;

import net.bancino.robotics.swerveio.geometry.ChassisDimension;
import net.bancino.robotics.swerveio.geometry.Length;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

import net.bancino.robotics.swerveio.module.VirtualSwerveModule;

/**
 * Static methods dealing with a virtual swerve drive (see {@link #create()}).
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class VirtualSwerveDrive {

    /**
     * Create a virtual swerve drive that utilizes all SwerveIO code, but doesn't
     * map to any hardware. This is used primarily for testing SwerveIO code. It
     * directs all outputs to an entirely virtual swerve module, which is backed by
     * virtual motors and encoders that are updated using arbitrary characteristics.
     * 
     * @return A swerve drive that does not map to any hardware. It can be used in
     *         unit tests.
     */
    public static SwerveDrive create() {
        return new SwerveDrive.Builder().useDefaultKinematics(new ChassisDimension(new Length(25, Unit.INCHES)))
                .setModuleMap((Map<SwerveModule.Location, SwerveModule> map) -> {
                    map.put(SwerveModule.Location.FRONT_RIGHT, new VirtualSwerveModule());
                    map.put(SwerveModule.Location.FRONT_LEFT, new VirtualSwerveModule());
                    map.put(SwerveModule.Location.REAR_LEFT, new VirtualSwerveModule());
                    map.put(SwerveModule.Location.REAR_RIGHT, new VirtualSwerveModule());
                }).build();
    }
}
