/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * A virtual encoder that implements the SwerveIO encoder interface. This does
 * not map to any hardware so it can be used for testing. Note, however, that
 * this must be manually updated by some method using {@code set()} because
 * otherwise, the reading will not change.
 * 
 * <p>
 * This emulates a quadrature encoder, so there are 4096 counts per revolution
 * as reflected by {@link #countsPerRevolution()}.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
@Loggable
public class VirtualEncoder implements Encoder {

    @Log(as = "rawValue", atLevel = LogLevel.IMPORTANT)
    private double value;

    /**
     * Set this encoder's value to an arbitary number.
     * 
     * @param value The raw value to set the encoder to.
     */
    public void set(double value) {
        this.value = value;
    }

    @Override
    public double get() {
        return value;
    }

    @Override
    public double countsPerRevolution() {
        return 4096;
    }
}
