/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.module;

import net.bancino.robotics.swerveio.geometry.Length;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

import net.bancino.robotics.swerveio.VirtualMotor;
import net.bancino.robotics.swerveio.encoder.VirtualEncoder;

/**
 * A virtual swerve module that implements the SwerveIO swerve module interface.
 * THis does not map to any hardware so it can be used for testing. The methods
 * in this interface automatically updates the encoders based on the motor
 * speeds.
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class VirtualSwerveModule extends GenericSwerveModule {

    /**
     * Create an virtual swerve module with arbitrary parameters using virtual
     * motors and encoders.
     */
    public VirtualSwerveModule() {
        super(new GenericSwerveModule.Builder().setDriveMotor(new VirtualMotor()).setDriveEncoder(new VirtualEncoder())
                .setPivotMotor(new VirtualMotor()).setPivotEncoder(new VirtualEncoder()).setAngleOffset(0)
                .setDriveGearRatio(10).setPivotGearRatio(5).setDriveMaxRPM(5000)
                .setWheelDiameter(new Length(4, Unit.INCHES)));
    }

    @Override
    public void setAngle(double angle) {
        super.setAngle(angle);

        /* Update the encoder based on the motor speed. */
        VirtualEncoder encoder = (VirtualEncoder) getPivotEncoder();
        encoder.set(encoder.get() + (encoder.countsPerRevolution() * getPivotMotor().get()));
    }

    @Override
    public void setSpeed(double speed) {
        super.setSpeed(speed);

        /* Update the encoder based on the motor speed. */
        VirtualEncoder encoder = (VirtualEncoder) getDriveEncoder();
        encoder.set(encoder.get() + (encoder.countsPerRevolution() * getDriveMotor().get()));
    }

    @Override
    public double getOutputThreshhold() {
        return 0.0;
    }
}
