/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.List;

/**
 * A log hander that generates JSON output directly mirroring
 * the code hierarchy.
 *
 * @author Jordan Bancino
 * @version 7.0.0
 * @since 7.0.0
 */
public class JsonLog implements LogHandler {

    public static String toJson(LogEntry entry) {
        StringBuilder json = new StringBuilder();
        if (entry != null) {
            switch (entry.getType()) {
            case LIST:
                List<LogEntry> values = entry.getAsList();
                json.append('[');
                for (int i = 0; i < values.size(); i++) {
                    json.append(toJson(values.get(i)));
                    if (i < values.size() - 1) {
                        json.append(',');
                    }
                }
                json.append(']');
                break;
            case NODE:
                json.append('{');
                for (int i = 0; i < entry.getChildren().size(); i++) {
                    LogEntry child = entry.getChildren().get(i);
                    json.append('"').append(child.getName()).append("\":");
                    json.append(toJson(child));
                    if (i < entry.getChildren().size() - 1) {
                        json.append(',');
                    }
                }
                json.append('}');
                break;
            case VALUE:
                Object raw = entry.getValue();
                String str;

                if (raw == null) {
                    str = "null";
                } else if (raw instanceof Number || raw instanceof Boolean) {
                    str = raw.toString();
                } else {
                    /* TODO: do proper JSON escaping. */
                    str = String.format("\"%s\"", raw.toString());
                }
                json.append(str);
                break;
            default:
                break;
            }
        }
        return json.toString();
    }

    private final PrintStream out;
    private boolean doComma = false;

    public JsonLog(String out) throws FileNotFoundException {
        this(new File(out));
    }

    public JsonLog(File out) throws FileNotFoundException {
        this(new PrintStream(out));
    }

    public JsonLog(PrintStream out) {
        this.out = out;
    }

    @Override
    public void open() {
        out.print('[');
    }

    @Override
    public void accept(LogEntry entry) {
        String json = toJson(entry);
        if (doComma && !json.isBlank()) {
            out.print(',');
        } else {
            doComma = true;
        }
        out.print(json);
    }

    @Override
    public void close() {
        out.println(']');
        out.close();
    }
}
