/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * The Log annotation. This is used to annotate fields that should have their values
 * logged. LogIO will automatically collect primitives and basic object information.
 * If a field is of a custom type, make sure that custom type is annotated with
 * {@link Loggable} to ensure that it actually produces useful log information.
 *
 * @author Jordan bancino
 * @version 7.0.0
 * @since 7.0.0
 */
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.FIELD)
public @interface Log {

    /**
     * Set the key name for this field. By default, the key name is the name of
     * the field variable, but this can be used to change that if the field variable
     * name is not appropriate or descriptive enough.
     * @return The key name for this field, or a blank string if the name of the field
     * variable should be used.
     */
    public String as() default "";

    /**
     * Set the level at which this field should be logged.
     * @return The log level at which this field should be logged.
     * The default is {@code TRACE}.
     */
    public LogLevel atLevel() default LogLevel.TRACE;
}
