/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

import java.util.ArrayList;
import java.util.List;

/**
 * A representation of a log enry. LogIO log entries are of a certain type, and
 * may optionally have a parent and any number of children. This class creates
 * the log tree hierarchy.
 *
 * @author Jordan Bancino
 * @version 7.0.0
 * @since 7.0.0
 */
public class LogEntry {

    /**
     * The possible log entry types.
     */
    public static enum Type {
        /**
         * Indicates a log entry is a list of loggable objects.
         */
        LIST,

        /**
         * Indicates a log entry is the parent to a number of child
         * log entries.
         */
        NODE,

        /**
         * Indicates a log entry is a value that should be logged.
         */
        VALUE
    }

    private final LogEntry parent;
    private final List<LogEntry> children;
    private String name;
    private Object value;

    /**
     * Create a new log entry. This constructor should only be used by
     * {@link Logger}. Any other uses may result in undefined behavior.
     *
     * @param parent The parent log entry, or {@code null} if this is the
     *               root entry.
     * @param name The name or "key" of this entry.
     * @param value The value of this entry.
     */
    public LogEntry(LogEntry parent, String name, Object value) {
        this.parent = parent;
        this.children = new ArrayList<>();
        this.name = name;
        this.value = value;

        if (parent != null) {
            parent.getChildren().add(this);
        }
    }

    /**
     * Get the parent log entry of this log entry.
     *
     * @return A parent log entry, or {@code null} if this log entry is a root entry.
     */
    public LogEntry getParent() {
        return parent;
    }

    /**
     * Get the child log entries of this log entry.
     *
     * @return A list of child log entries. This may be an empty list if there
     * are no children, but it will never be null.
     */
    public List<LogEntry> getChildren() {
        return children;
    }

    /**
     * Get the name of this log entry.
     *
     * @return The name of this log entry.
     */
    public String getName() {
        return name;
    }

    /**
     * Set the name of this log entry.
     *
     * @param name The new name of this log entry.
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Get the value of this log entry.
     *
     * @return The raw value of this log entry.
     */
    public Object getValue() {
        return value;
    }

    /**
     * Set the value of this log entry.
     *
     * @param value The new value of this log entry.
     */
    public void setValue(Object value) {
        this.value = value;
    }

    /**
     * Get entry type of this log entry. This is determined by evaluating the type of
     * the current value at the time of calling, and should be used to guide decisions
     * of what to do with the value.
     *
     * <ul>
     *     <li>
     *         If this returns {@code Type.LIST}, then this log entry's value should be
     *         processed by calling {@link #getAsList()}.
     *     </li>
     *     <li>
     *         If this returns {@code Type.NODE}, then this log entry's value should be
     *         processed by calling {@link #getChildren()}.
     *     </li>
     *     <li>
     *         If this returns {@code Type.VALUE}, then this log entry's value should be
     *         processed by calling {@link #getValue()}.
     *     </li>
     * </ul>
     *
     * Any other invocation usage of this log entry object is undefined.
     *
     * @return The type that this log entry's value represents.
     */
    public Type getType() {
        if (getValue() != null) {
            if (getValue() instanceof List<?>) {
                return Type.LIST;
            } else if (getValue().getClass().getAnnotation(Loggable.class) != null || getChildren().size() > 0) {
                return Type.NODE;
            }
        }
        return Type.VALUE;
    }

    /**
     * Get the value of this log entry as a list. This will only with if the type
     * of this log entry is set to {@code Type.LIST}.
     *
     * @return The value of this log entry as a list of log entries, or null if this
     * log entry does not represent a list.
     */
    @SuppressWarnings("unchecked")
    public List<LogEntry> getAsList() {
        return getType() == Type.LIST ? (List<LogEntry>) getValue() : null;
    }
}
