/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

import java.io.Closeable;
import java.util.function.Consumer;

/**
 * An interface for generating human-readable logs from {@link LogEntry} trees.
 * This interface provides a way for LogIO users to generate their own custom
 * logs. It is extremely basic in definition, but implementing it can be quite
 * tricky, due to the recursive nature of LogIO. To understand a bit more how
 * this interface works, you may wish to look at the code for {@link JsonLog}.
 *
 * @author Jordan Bancino
 * @version 7.0.0
 * @since 7.0.0
 */
/* Consumer provides accept(LogEntry), Closeable provides close() */
/* This is still technically a functional interface because it only has one method that is mandatory. */
@FunctionalInterface
public interface LogHandler extends Consumer<LogEntry>, Closeable {
    /**
     * This method is called upon the first invocation of {@link Logger#log(Object)}.
     * It can be used for setting up log files, by, for instance, writing log headers
     * or other information.
     */
    public default void open() {
    }

    /**
     * This method is called upon the invocation of {@link Logger#close()}.
     * It can be used for writing log footers and closing resource streams.
     */
    @Override
    public default void close() {
    }
}
