/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

/**
 * An enumeration that defines the support log levels. LogIO allows every field
 * to declare the verbosity level at which it is logged. The user can then set the
 * log level to include or omit log information depending on his or her needs.
 *
 * @author Jordan Bancino
 * @version 7.0.0
 * @since 7.0.0
 */
public enum LogLevel {
    /**
     * Log nothing. This effectively disables LogIO altogether by bypassing all
     * the log routines. This can be useful for enhancing the performance of
     * production code when no log data is needed.
     */
    NOTHING,

    /**
     * Log only the most important data. Fields that are set to log at this
     * level will be logged every time, no matter the log level, unless it
     * is set to {@code NOTHING}.
     */
    IMPORTANT,

    /**
     * This value indicates a field is informational in nature, and can be useful
     * in many scenarios. This value just below {@code IMPORTANT}, which means
     * that fields that are set to log at this level will be logged in every scenario
     * that the log level is not set to {@code IMPORTANT} or {@code NOTHING}.
     */
    INFO,

    /**
     * Log most data. This value indicates a verbose log, which can be useful for
     * debugging. Fields that are set to log at this level are generally unimportant
     * to the normal user and only useful for developers. They are only logged if the
     * log level is set to {@code DEBUG} or {@code TRACE}.
     */
    DEBUG,

    /**
     * Log everything. By default, {@link Log} sets the field log level to {@code TRACE}.
     * Setting the log level to {@code TRACE} guarantees that all fields annotated with
     * {@link Log} and are reachable from an object are indeed logged.
     */
    TRACE
}
