/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.log;

import java.io.Closeable;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * The main LogIO class that is responsible for using reflection to recursively
 * process all {@link Log} and {@link Loggable} annotations to generate a tree
 * of {@link LogEntry}s.
 *
 * @author Jordan Bancino
 * @version 7.0.0
 * @since 7.0.0
 */
public class Logger implements Closeable {

    private LogLevel level = LogLevel.INFO;
    private List<LogHandler> to = new ArrayList<>();
    private boolean calledOpen = false;

    /**
     * Get whether or not the field annotated with the provided annotation
     * instance should be logged or not. This takes in to account the currently
     * set log level, and the log level of the annotation.
     *
     * @param annot The log annotation to check.
     * @return Whether or not the field annotated with this annotation should be
     * added to the {@link LogEntry} tree.
     */
    private boolean shouldLog(Log annot) {
        return level != LogLevel.NOTHING && annot.atLevel().ordinal() <= level.ordinal();
    }

    /**
     * Set the log level on this logger. This determines what fields get logged
     * and what fields are skipped, based on each field's log level.
     *
     * @param level The level, above which, fields should be logged at.
     * @return This instance, for method chaining.
     */
    public Logger setLevel(LogLevel level) {
        this.level = level;
        return this;
    }

    /**
     * Add log handers to process the log entries created by this logger.
     *
     * @param outputs Custom log output handlers.
     * @return This instance, for method chaining.
     */
    public Logger outputTo(LogHandler... outputs) {
        for (LogHandler output : outputs) {
            if (output != null && !to.contains(output)) {
                to.add(output);
            }
        }
        return this;
    }

    /**
     * Get the log level that is currently set.
     *
     * @return The current log level.
     */
    public LogLevel getLevel() {
        return level;
    }

    /**
     * Recursively log an object. This method processes all annotated fields
     * and generates a log entry tree. It then automatically calls all the
     * log handlers that are associated with this class.
     *
     * @param o The object to log all values for.
     * @throws UnsupportedOperationException If a reflection operation fails.
     */
    public void log(Object o) throws UnsupportedOperationException {
        if (level != LogLevel.NOTHING) {
            try {
                LogEntry rootEntry = recursiveLog(null, null, o);
                for (LogHandler output : to) {
                    if (output != null) {
                        if (!calledOpen) {
                            output.open();
                            calledOpen = true;
                        }
                        output.accept(rootEntry);
                    }
                }
            } catch (IllegalAccessException e) {
                throw new UnsupportedOperationException(e);
            }
        }
    }

    private void extractFields(LogEntry parent, Class<?> clazz, Object obj) throws IllegalAccessException {
        /*
         * If the annotation is not null, the user wants us to log values in this
         * object's class, not just the object itself.
         */
        if (clazz.getAnnotation(Loggable.class) != null) {
            /* Iterate over all the fields of the class */
            for (Field field : clazz.getDeclaredFields()) {
                /* Get the field annotation. */
                Log fieldAnnot = field.getAnnotation(Log.class);

                /*
                 * If the field annotation is not null, the user wants us to log the value of
                 * this field.
                 */
                if (fieldAnnot != null && shouldLog(fieldAnnot)) {
                    /* If the field is private, this will allow us to retrieve it. */
                    field.setAccessible(true);

                    /* Get the name of the field as it is declared in code. */
                    String name = field.getName();

                    /* If the user provided a custom name for this field, use that instead. */
                    if (!fieldAnnot.as().trim().isBlank()) {
                        name = fieldAnnot.as().trim();
                    }

                    /* Get the field's value on the current object. */
                    Object fieldVal = field.get(obj);
                    LogEntry fieldEntry = recursiveLog(parent, name, fieldVal);

                    if (fieldVal == null) {
                        /* Do nothing, the assignment above appends a null object. */
                    } else if (fieldVal instanceof List<?>) {
                        /* If the field is a list, log each item in the list. */
                        List<LogEntry> itemEntries = new ArrayList<>();
                        for (Object item : (List<?>) fieldVal) {
                            itemEntries.add(recursiveLog(null, null, item));
                        }
                        fieldEntry.setValue(itemEntries);
                    } else if (fieldVal.getClass().isArray()) {
                        /* If the field is an array, log each item in the list. */
                        List<LogEntry> itemEntries = new ArrayList<>();
                        for (int i = 0; i < Array.getLength(fieldVal); i++) {
                            Object item = Array.get(fieldVal, i);
                            itemEntries.add(recursiveLog(null, null, item));
                        }
                        fieldEntry.setValue(itemEntries);
                    } else if (fieldVal instanceof Map) {
                        /* If the field is a map, log each item in the map. */
                        Map<?, ?> fieldMap = (Map<?, ?>) fieldVal;
                        for (Object k : fieldMap.keySet()) {
                            Object v = fieldMap.get(k);
                            recursiveLog(fieldEntry, k.toString(), v);
                        }
                    } else if (fieldVal instanceof Enum<?>) {
                        fieldEntry.setValue(fieldVal.toString());
                        fieldEntry.getChildren().clear();
                    }
                }
            }
        }
    }

    private LogEntry recursiveLog(LogEntry parent, String thisName, Object obj) throws IllegalAccessException {

        /* If the object is null, return a null entry. */
        if (obj == null) {
            return new LogEntry(parent, thisName, null);
        }

        /* Get this object's class. */
        Class<?> clazz = obj.getClass();
        String className = clazz.getSimpleName();

        LogEntry entry = new LogEntry(parent, thisName != null ? thisName : className, obj);

        /* Iterate over superclasses */
        Class<?> superClass = clazz.getSuperclass();
        while (superClass != null) {
            /* The conditions for including this superclass's fields. */
            boolean include = true;
            include &= !superClass.isInterface();
            include &= !superClass.equals(AbstractMap.class);
            include &= !superClass.equals(Number.class);
            include &= !superClass.equals(Object.class);
            if (include) {
                LogEntry superclassEntry = new LogEntry(entry, superClass.getSimpleName(), obj);
                extractFields(superclassEntry, superClass, obj);
            }
            superClass = superClass.getSuperclass();
        }

        extractFields(entry, clazz, obj);
        return entry;
    }

    /**
     * Calls {@link LogHandler#close()} on all handlers associated with this
     * logger.
     */
    @Override
    public void close() {
        if (level != LogLevel.NOTHING) {
            for (LogHandler output : to) {
                output.close();
            }
        }
    }
}
