/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Map;
import java.util.Properties;
import java.util.EnumMap;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj.DriverStation;
import edu.wpi.first.wpilibj.Filesystem;
import edu.wpi.first.math.kinematics.ChassisSpeeds;
import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.geometry.ChassisDimension;
import net.bancino.robotics.swerveio.geometry.ModuleVector;
import net.bancino.robotics.swerveio.geometry.SwerveVector;
import net.bancino.robotics.swerveio.geometry.Length.Unit;
import net.bancino.robotics.swerveio.gyro.Gyro;
import net.bancino.robotics.swerveio.gyro.WPILibGyro;
import net.bancino.robotics.swerveio.kinematics.SwerveKinematicsProvider;
import net.bancino.robotics.swerveio.kinematics.DefaultSwerveKinematics;
import net.bancino.robotics.swerveio.log.RobotLogger;
import net.bancino.robotics.swerveio.module.SwerveModule;
import net.bancino.robotics.swerveio.pid.PIDController;
import net.bancino.robotics.swerveio.pid.DefaultPIDController;

/**
 * A class that extends a WPILib subsystem to implement a swerve drive.
 *
 * <p>
 * This takes care of consolidating swerve modules so they can easily be driven
 * as a system. This class is compatible with the "new" command framework
 * provided by the {@code edu.wpi.first.wpilibj2.command} package. It is
 * <i>not</i> compatible with the original command framework, and therefore
 * requires WPILib 2020 or later.
 * </p>
 * <p>
 * The {@code SwerveDrive} is extremely flexible, offering multiple
 * {@code drive()} methods, as you'll see below. The different methods allow
 * driving the robot in different ways. For example,
 * {@link #drive(SwerveVector)} would be most appropriately used for driving
 * with a joystick, since it uses percentages of full output to drive the robot,
 * generally in the same units as whatever the joystick outputs. On the other
 * hand, {@link #drive(ChassisSpeeds)} takes an absolute speed reference, in
 * meters per second and/or radians per second. This makes it ideal for
 * autonomous, especially waypoint autonomous, where the state of the robot at
 * each loaded as an absolute position/velocity vector.
 * </p>
 * <p>
 * While there are many methods available in this class, the only methods that
 * the standard SwerveIO user will use should be the {@code drive()} methods, as
 * well as perhaps {@link #setReversed(DegreeOfFreedom, boolean)}, for reversing
 * axes. If the programmer utilizes the command package that SwerveIO provides,
 * they may not even have to touch {@code drive()} either, at least for
 * tele-operated period. Autonomous may require gyro resets or other state
 * setting, but SwerveIO provides an API for PathWeaver, and a command to
 * automatically drive the swerve with it.
 * </p>
 * <p>
 * If you need to debug your swerve drive, you can use the logging API, which is
 * defined in this class, as well as in the
 * {@link net.bancino.robotics.swerveio.log} package. The log package provides
 * the actual implementations of the loggers, which pull state information from
 * the API provided by this class and push data to the output destination. See
 * the log package documentation for more details. Every swerve drive logger is
 * passed a hierarchal collection of variables. If you are using an
 * already-existing logging implementation, see {@link #getLogger()} to add it.
 * </p>
 * <p>
 * This class automatically takes care of calling the loggers registered to it
 * by overriding {@code SubsystemBase.periodic()} (see {@link #periodic()}) to
 * interatively call the logging methods, so you do not have to manually add the
 * swerve drive logger to your robot's iterative methods, allowing for cleaner,
 * more efficient code. You can add your loggers via
 * {@link SwerveDrive#getLogger()} and then not worry about them anymore.
 * </p>
 *
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 1.0.0
 */
@Loggable
public class SwerveDrive extends SubsystemBase {

    /*
     * The GNU GPL recommends outputting a copyright message and states that you are
     * not allowed to remove this copyright message if it is present.
     */
    static {
        System.out.println("********************************************************************************");
        System.out.println("* SwerveIO - Copyright (C) Jordan Bancino <jordan@bancino.net>                 *");
        System.out.println("*                                                                              *");
        System.out.println("* This program comes with ABSOLUTELY NO WARRANTY. This is free software, and   *");
        System.out.println("* you are welcome to redistribute it under certain conditions. For details,    *");
        System.out.println("* see the license file that should have been distributed with this program. If *");
        System.out.println("* you did not receive a license with this program, visit                       *");
        System.out.println("* https://bancino.net/licenses/gpl-3.0.txt                                     *");
        System.out.println("********************************************************************************");
    }

    /**
     * Build a new swerve drive object. This class makes it easy to set up the
     * swerve drive with all the components it requires to run, and is required by
     * SwerveIO. In cases such as this one, SwerveIO will follow the "builder
     * pattern" of Java, because often each field used in the object's construction
     * requires a substantial amount of setup. The builder pattern is beyond the
     * scope of this documentation, but the details of this class and how to use it
     * are provided here.
     *
     * <p>
     * To construct a swerve drive, you must provide the following information:
     * </p>
     * <ul>
     * <li>The dimensions of your swerve drive</li>
     * <li>The CAN/PWM motors to use. These should implement WPILib's
     * {@code SpeedController}</li>
     * <li>The drive and pivot encoders attached to the motors. These will implement
     * {@link net.bancino.robotics.swerveio.encoder.Encoder}</li>
     * </ul>
     * <p>
     * Optionally, you can provide a gyro for field-centric navigation, or your own
     * custom swerve drive kinematics provider for controlling modules. At this
     * stage in the system code, you are not assigning joysticks or anything of that
     * sort (unless you use the initialize block, as documented below). This builder
     * is merely to bind the actual hardware of your swerve drive to the swerve
     * drive object.
     * </p>
     * <p>
     * Unless otherwise stated, all methods in this class overwrite the previous
     * value of the field they set, so if you call the same method twice, the latest
     * call will take precedence over the previous call.
     * </p>
     * <h2>Example</h2>
     * <p>
     * Here is a basic example (with many details omitted for the sake of brevity)
     * that shows how to construct a swerve drive using this builder class.
     * </p>
     *
     * <pre>
     * {@code
     * import net.bancino.robotics.swerveio.SwerveDrive;
     * import net.bancino.robotics.swerveio.geometry.ChassisDimension;
     * import net.bancino.robotics.swerveio.geometry.Length;
     * import net.bancino.robotics.swerveio.geometry.Unit;
     * import net.bancino.robotics.swerveio.SwerveModule;
     * import net.bancino.robotics.swerveio.DegreeOfFreedom;
     *
     * public class Drivetrain {
     *   // In your RobotContainer class, or wherever you declare your subsystems,
     *   // you an call Drivetrain.create() to construct your custom swerve drive.
     *   public static SwerveDrive create() {
     *     return new SwerveDrive.Builder()
     *       .useDefaultKinematics(new ChassisDimension(new Length(27.4, Unit.INCHES))) // This should be the size of your robot.
     *       .setGyro(gyro) // Create/add your gyro here.
     *       .setModuleMap((map) -> {
     *         map.put(SwerveModule.FRONT_RIGHT, frontRight); // Create/add your front right module here
     *         map.put(SwerveModule.FRONT_LEFT, frontLeft);   // Create/add your front left module here
     *         map.put(SwerveModule.REAR_LEFT, rearLeft);     // Create/add your rear left module here
     *         map.put(SwerveMOdule.REAR_RIGHT, rearRight);   //Create/add your rear right module here
     *       }, (module) -> {
     *         // This code is run on every one of the modules added to the map, above.
     *         var pid = module.getPivotPIDController();
     *         pid.setOutputRampRate(0.1);
     *         pid.setP(0.003);
     *         pid.setI(0.000015);
     *         pid.setD(0);
     *       }).build((swerve) -> {
     *         // Additional configuration on the swerve drive that isn't available in the builder
     *         swerve.setReversed(DegreeOfFreedom.FORWARD, true);
     *       });
     *   }
     * }
     * }
     * </pre>
     * 
     * <p>
     * Note that this example is not intended to explain functions like
     * {@link SwerveDrive#setReversed(DegreeOfFreedom, boolean)} or
     * {@link net.bancino.robotics.swerveio.pid.PIDController#setP(double)}. Nor is
     * it intended to compile. You must add your code where the comments indicate,
     * namely at {@code gyro}, {@code frontRight}, {@code frontLeft},
     * {@code readLeft}, and {@code rearRight}. It should also be noted that the
     * example PID values above are by no means an indication of what your values
     * should be. They are merely for the sake of completeness. You can create
     * swerve modules by implementing
     * {@link net.bancino.robotics.swerveio.module.SwerveModule}, instantiating
     * {@link net.bancino.robotics.swerveio.module.GenericSwerveModule}, or choosing
     * a hardware module that is aready created and tested. Each swerve mdoule
     * requires a minimum of one (1) encoder: the pivot encoder, as well as two
     * motors. See the module documentation (available at the link above) for
     * details on how to add encoders and motors to your modules.
     * </p>
     * <p>
     * Once you have constructed a swerve drive object, you can proceed to drive
     * your robot with it as documented in {@link SwerveDrive}.
     * </p>
     * 
     * @see SwerveDrive
     * @see net.bancino.robotics.swerveio.module.SwerveModule
     */
    public static class Builder {
        private SwerveKinematicsProvider kinematicsProvider = null;
        private Map<SwerveModule.Location, SwerveModule> moduleMap = null;

        private Gyro gyro = null;
        private PIDController gyroPID = null;
        private int gyroStandingSlot;
        private int gyroMovingSlot;

        private Runnable periodic;

        /**
         * Instantiate a new SwerveDrive builder. This accepts no parameters and does
         * nothing by default, because this class is a builder class: you call it's
         * methods to set it up. When you're done, call {@link #build()} or
         * {@link #build(Consumer)} to get the swerve drive object. Note that it still
         * may require extra initialization before it is ready to be used. For example,
         * you may set properties such as the "idle angle", and may want to reverse
         * axes. To do this before retrieving the {@link SwerveDrive} object from this
         * builder, use {@link #build(Consumer)}, providing a {@link SwerveDrive}
         * consumer.
         */
        public Builder() {
        }

        /**
         * Set the kinematics provider for this swerve drive. The kinematics provider is
         * responsible for supplying the dimensions of the swerve drive, as well as
         * performing all the mathematical computations for the swerve modules. If you
         * aren't sure what this does, or if you don't plan on implementing your own
         * kinematics provider, use {@link #useDefaultKinematics(ChassisDimension)}
         * instead. This function is for programmers that want to change the fundamental
         * way their swerve drive is driven. Users that just need a working swerve drive
         * can get by with the default kinematics without issue. The default kinematics
         * are of the highest quality and are used in competitions, so don't feel that
         * you have to implement the kinematics provider interface yourself.
         * 
         * @param kinematicsProvider The kinematics provider to use. This provides the
         *                           dimensions for your swerve drive.
         * @return The instance this method was called on so that methods can be
         *         chained.
         * @see net.bancino.robotics.swerveio.kinematics.SwerveKinematicsProvider
         */
        public Builder setKinematicsProvider(SwerveKinematicsProvider kinematicsProvider) {
            this.kinematicsProvider = kinematicsProvider;
            return this;
        }

        /**
         * Use the default swerve drive mode with your swerve drive's linear dimensions.
         * Under most circumstances, this is how you will set the dimensions of your
         * swerve drive. The dimensions are used in the vector calculations for each
         * module, as well as computing the angular velocity of your robot as it moves.
         * If you use a custom kinematics provider, see
         * {@link #setKinematicsProvider(SwerveKinematicsProvider)}. Your kinematics
         * provider is responsible for providing the chassis dimensions too.
         * 
         * <p>
         * If you don't know what a kinematics provider is, use this method, because you
         * can directly provide your chassis dimensions.
         * </p>
         * 
         * @param chassis Your swerve drive's dimensions. See
         *                {@link net.bancino.robotics.swerveio.geometry.ChassisDimension}.
         * @return The instance this method was called on so that methods can be
         *         chained.
         * @see net.bancino.robotics.swerveio.kinematics.SwerveKinematicsProvider
         */
        public Builder useDefaultKinematics(ChassisDimension chassis) {
            return setKinematicsProvider(new DefaultSwerveKinematics(chassis));
        }

        /**
         * Set the gyro that this swerve drive should use. If you provide a gyro,
         * field-centric navigation will be enabled, and you can optionally disable it
         * later. If you do not provide a gyro, field-centric navigation is disabled and
         * you cannot enable it after the creation of the swerve drive.
         * 
         * @param gyro The gyro implementation to use for field-centric navigation.
         * @return The instance this method was called on so that methods can be
         *         chained.
         * @see net.bancino.robotics.swerveio.gyro.Gyro
         */
        public Builder setGyro(Gyro gyro) {
            this.gyro = gyro;
            return this;
        }

        /**
         * Set the gyro that this swerve drive should use. If you provide a gyro,
         * field-centric navigation will be enabled, and you can optionally disable it
         * later. If you do not provide a gyro, field-centric navigation is disabled and
         * you cannot enable it after the creation of the swerve drive.
         * 
         * @param gyro The gyro implementation to use for field-centric navigation.
         * @return The instance this method was called on so that methods can be
         *         chained.
         * @see edu.wpi.first.wpilibj.interfaces.Gyro
         */
        public Builder setGyro(edu.wpi.first.wpilibj.interfaces.Gyro gyro) {
            return setGyro(new WPILibGyro(gyro));
        }

        /**
         * Tune the gyro PID controller to be used for angle positioning in
         * {@link #drive(SwerveVector,double)}. This controller is used to generate the
         * rotational parameter of a swerve drive vector. It recieves gyro angles and
         * produces a percentage output on the rotational component.
         * 
         * <p>
         * This method uses the default PID controller with
         * {@link #setAnglePID(PIDController, int, int, Consumer)}. Refer to the
         * documentation for that method for parameter descriptions.
         * </p>
         * 
         * @param standingSlot See
         *                     {@link #setAnglePID(PIDController, int, int, Consumer)}.
         * @param movingSlot   See
         *                     {@link #setAnglePID(PIDController, int, int, Consumer)}.
         * @param initialize   See
         *                     {@link #setAnglePID(PIDController, int, int, Consumer)}.
         * @return The instance this method was called on so that methods can be
         *         chained.
         */
        public Builder setAnglePID(int standingSlot, int movingSlot, Consumer<PIDController> initialize) {
            return setAnglePID(new DefaultPIDController(), standingSlot, movingSlot, initialize);
        }

        /**
         * Set a PID controller to be used for angle positioning in
         * {@link #drive(SwerveVector,double)}. This controller is used to generate the
         * rotational parameter of a swerve drive vector. It recieves gyro angles and
         * produces a percentage output on the rotational component.
         *
         * @param gyroPID      A PID controller that accepts gyro angles and outputs the
         *                     rotational component of a swerve vector.
         * @param standingSlot See
         *                     {@link #setAnglePID(PIDController, int, int, Consumer)}
         *                     for details.
         * @param movingSlot   See
         *                     {@link #setAnglePID(PIDController, int, int, Consumer)}
         *                     for details.
         *
         * @return The instance this method was called on so that methods can be
         *         chained.
         * 
         * @see net.bancino.robotics.swerveio.geometry.SwerveVector
         */
        public Builder setAnglePID(PIDController gyroPID, int standingSlot, int movingSlot) {
            return setAnglePID(gyroPID, standingSlot, movingSlot, null);
        }

        /**
         * Set a PID controller to be used for angle positioning in
         * {@link #drive(SwerveVector,double)}. This controller is used to generate the
         * rotational parameter of a swerve drive vector. It recieves gyro angles and
         * produces a percentage output on the rotational component.
         * 
         * <p>
         * Starting with SwerveIO 5.0.4, gyro PID controllers must have two active
         * profile slots configured. See
         * {@link net.bancino.robotics.swerveio.pid.PIDController} for information on
         * profile slots. Gyro PID controllers must have these two slots configured
         * because swerve drives behave differently depending on whether they're
         * stationary or in movement. Rotation while in movement is often easier and
         * requires lower PID gains than rotation while stationary.
         * </p>
         * 
         * @param <P>          The type of the PID controller you're using.
         * @param gyroPID      A PID controller that accepts gyro angles and outputs the
         *                     rotational component of a swerve vector.
         * @param standingSlot The PID profile slot to use when the robot is stationary.
         *                     Typically this has higher gains to break the friction of
         *                     the swerve drive.
         * @param movingSlot   The PID pofile slot to use when the robot is in motion.
         *                     Typically this has lower gains because there is a lot
         *                     less resitance to movement when the swerve drive is
         *                     already moving.
         * @param initialize   Initialize the passed PID controller. This can be used to
         *                     perform extra setup not provided by constructors. Note
         *                     that the setpoint range and the maximum output are
         *                     automatically set after this function is executed. You
         *                     can also configure the two required profile slots in this
         *                     function if you haven't done so before passing the PID
         *                     controller into this method.
         * @return The instance this method was called on so that methods can be
         *         chained.
         */
        public <P extends PIDController> Builder setAnglePID(P gyroPID, int standingSlot, int movingSlot,
                Consumer<P> initialize) {
            if (initialize != null) {
                initialize.accept(gyroPID);
            }
            this.gyroPID = gyroPID;
            this.gyroStandingSlot = standingSlot;
            this.gyroMovingSlot = movingSlot;
            return this;
        }

        /**
         * Set a PID controller to be used for angle positioning in
         * {@link #drive(SwerveVector,double)}. This controller is used to generate the
         * rotational parameter of a swerve drive vector. It recieves gyro angles and
         * produces a percentage output on the rotational component.
         *
         * @param gyroPID A PID controller that accepts gyro angles and outputs the
         *                rotational component of a swerve vector. This PID controller
         *                must have slots 0 and 1 configured for standing and moving
         *                gains respectively. See
         *                {@link #setAnglePID(PIDController, int, int)} for details.
         *
         * @return The instance this method was called on so that methods can be
         *         chained.
         * 
         * @see net.bancino.robotics.swerveio.geometry.SwerveVector
         */
        public Builder setAnglePID(PIDController gyroPID) {
            return setAnglePID(gyroPID, null);
        }

        /**
         * Set a PID controller to be used for angle positioning in
         * {@link #drive(SwerveVector,double)}. This controller is used to generate the
         * rotational parameter of a swerve drive vector. It recieves gyro angles and
         * produces a percentage output on the rotational component.
         *
         * @param gyroPID    A PID controller that accepts gyro angles and outputs the
         *                   rotational component of a swerve vector. This PID
         *                   controller must have slots 0 and 1 configured for standing
         *                   and moving gains respectively. See
         *                   {@link #setAnglePID(PIDController, int, int, Consumer)} for
         *                   details.
         * @param initialize Initialize the passed PID controller. This can be used to
         *                   perform extra setup not provided by constructors. Note that
         *                   the setpoint range and the maximum output are automatically
         *                   set after this function is executed.
         *
         * @return The instance this method was called on so that methods can be
         *         chained.
         * 
         * @see net.bancino.robotics.swerveio.geometry.SwerveVector
         * @see net.bancino.robotics.swerveio.pid.PIDController
         */
        public Builder setAnglePID(PIDController gyroPID, Consumer<PIDController> initialize) {
            return setAnglePID(gyroPID, 0, 1, initialize);
        }

        /**
         * Create and initialize a swerve module map. This method automatically creates
         * a map object for you before passing it to the map initializer. In the
         * initializer, it is your job to set up the swerve modules and add them to the
         * module map.
         * 
         * <p>
         * The "module map" is a Java map object consisting of exactly 4 elements: one
         * for each swerve module. You may provide no more and no less than 4 modules.
         * Each module in the map should be placed under it's corresponding enumeration
         * value. See {@link SwerveModule.Location} for the valid enumeration values.
         * </p>
         * <p>
         * While it would be standard practice to initialize your modules in the
         * initialization function, it is by no means necessary. You can instantiate and
         * initialize them in any way, but you will use the consumer function to consume
         * an already-existing, empty map, to which you will add your four swerve
         * modules using {@code Map.put()}.
         * </p>
         * 
         * @param mapInitializer The initializer function that adds the swerve modules
         *                       to the provided module map. This method does nothing if
         *                       this parameter is null.
         * @return The instance this method was called on so that methods can be
         *         chained.
         * @see net.bancino.robotics.swerveio.module.SwerveModule
         * 
         */
        public Builder setModuleMap(Consumer<Map<SwerveModule.Location, SwerveModule>> mapInitializer) {
            if (mapInitializer != null) {
                Map<SwerveModule.Location, SwerveModule> moduleMap = new EnumMap<>(SwerveModule.Location.class);
                mapInitializer.accept(moduleMap);
                this.moduleMap = moduleMap;
            }
            return this;
        }

        /**
         * Set the periodic function to be run on the swerve drive's event thread.
         * Because you cannot override {@link SwerveDrive#periodic()}, you can set this
         * runnable to be executed inside it.
         * 
         * @param periodic A function to execute periodically, or null to not run any
         *                 function.
         * 
         * @return The instance this method was called on so that methods can be
         *         chained.
         */
        public Builder setPeriodic(Runnable periodic) {
            this.periodic = periodic;
            return this;
        }

        /**
         * Build a new swerve drive object by calling the internal constructor
         * {@link SwerveDrive#SwerveDrive(Builder)} &mdash; which is inaccessible to the
         * public, see {@link SwerveDrive#SwerveDrive(Builder)} for details &mdash; with
         * this builder object, completing the configuration and checking it to ensure
         * that it is correct. All error-checking code happens in the constructor, not
         * in this builder class.
         * 
         * @param initialize Initialize the newly created swerve drive. Use this to set
         *                   up the swerve drive before this method returns it.
         * @return Your new swerve drive. At this point, it should be fully configured
         *         and ready to pass into a command.
         * @throws IllegalArgumentException If there is a configuration error. This is a
         *                                  runtime exception that you should <b>not</b>
         *                                  catch, because if there is a fatal error, it
         *                                  will stop your robot code, which is the
         *                                  intended behavior.
         */
        public SwerveDrive build(Consumer<SwerveDrive> initialize) throws IllegalArgumentException {
            SwerveDrive swerve = new SwerveDrive(this);
            if (initialize != null) {
                initialize.accept(swerve);
            }
            return swerve;
        }

        /**
         * See {@link #build(Consumer)}, this is the same thing, but it passes
         * {@code null} in for the consumer. Use this if you do not want to initialize
         * the swerve drive object before it is returned.
         * 
         * @return Your new swerve drive. Note that it may require additional
         *         configuration because by using this method instead of
         *         {@link #build(Consumer)}, you opt not to execute initialization code
         *         on it.
         * @throws IllegalArgumentException If there is a configuration error.
         */
        public SwerveDrive build() throws IllegalArgumentException {
            return build(null);
        }
    }

    /**
     * SwerveIO fully supports the three degrees of freedom. This enumeration lists
     * them for use in methods that may require a degree of freedom be represented.
     * Currently, the only method that uses this enumeration is
     * {@link SwerveDrive#setReversed(DegreeOfFreedom, boolean)}, which is used to
     * reverse the swerve drive input for a particular degree of freedom.
     * 
     * <p>
     * A "degree of freedom" is most simply an axis of motion. Swerve drive has
     * three axes of motion: forward/reverse (called forward), left/right (called
     * strafe), and angular rotation (called rotation). Sometimes these are given
     * labels as they would appear on a coordinate plane, where X represents the
     * strafe axis, Y represents the forward axis, and Z represents the rotation
     * axis.
     * </p>
     * 
     * @author Jordan Bancino
     * @version 3.1.0
     * @since 2.2.0
     * @see net.bancino.robotics.swerveio.geometry.SwerveVector
     */
    public static enum DegreeOfFreedom {
        /**
         * Forward/backward motion, commonly represented on the Y axis.
         */
        FORWARD,

        /**
         * Right/left motion, commonly represented on the X axis.
         */
        STRAFE,

        /**
         * Clockwise/counterclockwise motion, commonly represented on the Z axis.
         */
        ROTATION
    }

    /**
     * The default file that is written to and read from when using
     * {@link #saveAngleOffsets()} and {@link #loadAngleOffsets()}, respectively.
     */
    protected static final File DEFAULT_ANGLE_OFFSET_FILE = new File(Filesystem.getOperatingDirectory(),
            "SwerveIOAngleOffsets.properties");

    @Log(as = "modules", atLevel = LogLevel.IMPORTANT)
    private final Map<SwerveModule.Location, SwerveModule> moduleMap;

    @Log(as = "kinematics", atLevel = LogLevel.INFO)
    private final SwerveKinematicsProvider calc;

    @Log(atLevel = LogLevel.INFO)
    private boolean fieldCentric = true;

    @Log(atLevel = LogLevel.INFO)
    private final Gyro gyro;

    @Log(atLevel = LogLevel.DEBUG)
    private final PIDController gyroPID;
    private boolean gyroSlotChanged = false;

    @Log
    private final int gyroStandingSlot;

    @Log
    private final int gyroMovingSlot;

    /**
     * When not moving, the idle angle is the default position. This is set
     * initially to 135, to point the modules in towards the center of the bot.
     */
    @Log
    private double idleAngle = 135;

    @Log
    private boolean idleAngleEnabled = false;

    /**
     * Diagonal flip will alternate the idle angle on the modules, causing them to
     * point opposite on diagonal corners. For instance, if you have an idle angle
     * of 135, setting this to true will set two diagonal modules to 45, because it
     * will flip them all so they point toward the center.
     */
    @Log
    private boolean diagonalFlipIdle = true;

    @Log(as = "vector", atLevel = LogLevel.IMPORTANT)
    private SwerveVector lastSwerveVector = new SwerveVector(0, 0, 0);

    /* Used for idle-angle calculations. */
    private final EnumMap<SwerveModule.Location, Double> lastPivotAngle = new EnumMap<>(SwerveModule.Location.class);

    private final EnumMap<DegreeOfFreedom, Boolean> reversedInput = new EnumMap<>(DegreeOfFreedom.class);

    private final Runnable periodic;

    @Log(as = "logger")
    private final RobotLogger log = new RobotLogger();

    /**
     * Construct the actual swerve drive object. This constructor could very well be
     * {@code private}, but it has been made {@code protected} to allow subclassing.
     * The normal way of creating a swerve drive object would be to use the
     * {@link Builder} class, but if you are extending this class, you wouldn't be
     * able to call this constructor. By making it protected, you can manually
     * invoke it with a builder object that you create in your class constructor.
     * 
     * @param builder The builder object used to construct the swerve drive class.
     *                This object acts as a wrapper for all the required values.
     * @see SwerveDrive.Builder
     */
    protected SwerveDrive(Builder builder) {
        if (builder == null) {
            throw new IllegalArgumentException("SwerveDrive Builder cannot be null.");
        }
        if (builder.moduleMap != null) {
            if (builder.moduleMap.size() == SwerveModule.Location.values().length) {
                this.moduleMap = builder.moduleMap;
                forEachModule((SwerveModule.Location location, SwerveModule module) -> {
                    if (module == null) {
                        throw new IllegalArgumentException("The " + location + " swerve module is null.");
                    }
                });
            } else {
                throw new IllegalArgumentException("Expected map of size " + SwerveModule.Location.values().length
                        + ". Got size " + builder.moduleMap.size());
            }
        } else {
            throw new IllegalArgumentException("Expected a module map, but the swerve builder didn't provide one.");
        }
        if (builder.kinematicsProvider != null) {
            this.calc = builder.kinematicsProvider;
        } else {
            throw new IllegalArgumentException(
                    "Expected a kinematics provider, but the swerve builder didn't provide one.");
        }

        /* Gyro configuration */
        this.gyro = builder.gyro;
        this.gyroPID = builder.gyroPID;

        /*
         * Whether or not these will be used doesn't matter, they just need to be set
         * because they're final.
         */
        this.gyroStandingSlot = builder.gyroStandingSlot;
        this.gyroMovingSlot = builder.gyroMovingSlot;

        /*
         * Configure the gyro PID controller if one was provided.
         */
        if (this.gyroPID != null) {
            /* Gyro goes from 0 -> 360 */
            this.gyroPID.setSetpointRange(360);

            /* Acceptable error is half a degree; a quarter degree on each side. */
            this.gyroPID.setAcceptableError(0.25);

            /* Each slot's output limit is 1. */
            this.gyroPID.setOutputLimits(this.gyroStandingSlot, 1);
            this.gyroPID.setOutputLimits(this.gyroMovingSlot, 1);
        }

        setFieldCentric(this.gyro != null);

        /* Set a user-defined periodic function, if any. */
        this.periodic = builder.periodic;

        /* Initial values in case these aren't changed. */
        setReversed(DegreeOfFreedom.FORWARD, false);
        setReversed(DegreeOfFreedom.STRAFE, false);
        setReversed(DegreeOfFreedom.ROTATION, false);
    }

    /**
     * Execute a task on every module associated with this swerve drive. This
     * provides a safe, easy way to access the module map without actually giving
     * access to the module map. This is safer than just returning the module map
     * because it prevents unauthorized additions and removals.
     * 
     * @param execute A {@code BiConsumer} that accepts a a swerve module map value
     *                and the actual module.
     * @throws NullPointerException If the passed BiConsumer is null.
     */
    public void forEachModule(BiConsumer<SwerveModule.Location, SwerveModule> execute) {
        moduleMap.forEach(execute);
    }

    /**
     * Get a swerve module from the module map. This is generally safer than
     * returning the swerve module map, because this prevents unauthorized additions
     * and subtractions to and from it.
     * 
     * @param module The swerve module to get. See the {@link SwerveModule.Location}
     *               enumeration for valid values.
     * @return A swerve module implementation for the given module.
     */
    public SwerveModule getModule(SwerveModule.Location module) {
        return moduleMap.get(module);
    }

    /**
     * Reverse the input of an axis of motion. If for some reason something is
     * running backwards, it can be flipped here. This will reverse the input that
     * is provided to the kinematics provider to produce calcuations that are
     * exactly opposite. While ideally this method isn't needed at all, it can help
     * quickly fix things if the gyro is out of phase or something, or your joystick
     * is reading backwards
     * 
     * @param direction The degree of motion to select. See {@link DegreeOfFreedom}
     *                  for the degrees of freedom that SwerveIO supports.
     * @param reversed  Whether or not to reverse the given degree. False indicates
     *                  that the axis should not be fipped, and true indicates that
     *                  it should.
     */
    public void setReversed(DegreeOfFreedom direction, boolean reversed) {
        reversedInput.put(direction, reversed);
    }

    /**
     * Get whether or not a degree of freedom is reversed. By default, all axes are
     * not reversed. This will only ever return true if an axis was explicitly
     * reversed by the user with {@link #setReversed(DegreeOfFreedom, boolean)}.
     * 
     * @param direction The degree of freedom to check.
     * @return Whether the given direction is being reversed before being passed
     *         into the kinematics provider.
     */
    public boolean getReversed(DegreeOfFreedom direction) {
        return reversedInput.get(direction);
    }

    /**
     * Set the field centric mode of the swerve drive. This requires a gyro
     * implementation. Field-centric operation is described in the overview of this
     * documentation. This function merely enables or disables it.
     * 
     * @param fieldCentric Whether or not field-centric drive is enabled or
     *                     disabled.
     * @throws UnsupportedOperationException If no gyro is available to the swerve
     *                                       drive object, because a gyro is
     *                                       required for field-centric navigation.
     */
    public void setFieldCentric(boolean fieldCentric) {
        if (fieldCentric && gyro == null) {
            throw new UnsupportedOperationException("Cannot enable field-centric mode without a gyro.");
        } else {
            this.fieldCentric = fieldCentric;
        }
    }

    /**
     * Whether or not this swerve drive is field centric. By default, this is true
     * if a gyro is provided, but may be false if it is explicitly disabled. If no
     * gyro is provided when the swerve drive is created, this will always return
     * false.
     * 
     * @return Whether or not the gyro is being used to offset the kinematic
     *         calculations to provide field-centric navigation.
     */
    public boolean isFieldCentric() {
        return fieldCentric;
    }

    /**
     * Set the angle at which all swerve modules will idle at. The "idle angle" is
     * defined as the default angle of each individual module when no motion is
     * occuring. This feature is disabled by default, which means that the modules
     * will stay in the position that they were last told to go to, which is more
     * energy-efficient. However, the idle angle feature is provided with the
     * intention that it can be used to "lock" the swerve modules in place, which
     * can prove crucial to a defensive strategy. If the modules are locked to point
     * toward the center of the robot, the robot should theoretically be immovable,
     * until the driver makes an effort to move the robot.
     * 
     * @param angle        The angle (in degrees) at which to idle the modules at
     *                     when no motion is occuring. This can be any arbitrary
     *                     value, as it will automatically be converted to the
     *                     correct angle.
     * @param diagonalFlip Whether or not to flip two diagonal angles to the
     *                     opposite of the provided angle. This is useful for
     *                     "locking" the swerve drive, ensuring it doesn't move. For
     *                     example, providing an angle of 45 and setting this
     *                     parameter to true, you can lock the swerve in place to
     *                     prevent it from being pushed around. If this is true, the
     *                     <b>front right</b> and <b>rear left</b> modules are
     *                     flipped from the other two.
     */
    public void setIdleAngle(double angle, boolean diagonalFlip) {
        if (angle < 0) {
            this.idleAngle = (angle <= -360) ? 360 - Math.abs(angle) % 360 : 360 - Math.abs(angle);
        } else {
            this.idleAngle = (angle >= 360) ? angle % 360 : angle;
        }
        this.diagonalFlipIdle = diagonalFlip;
        enableIdleAngle(true);
    }

    /**
     * The idle angle functionality tells the wheels to snap to a certain angle when
     * the drivetrain is idling. See {@link #setIdleAngle(double, boolean)}. This is
     * disabled by default, but can be toggled with this function. Note that this
     * method is implied if you call {@link #setIdleAngle(double, boolean)}, so
     * there is no need to call it explicitly if you are not using the default
     * idle-angle.
     *
     * @param enabled Whether or not to enable the idle angle functionality. This be
     *                called at any point in the lifetime of the robot code, even
     *                while it is being driven. This allows it to function as a
     *                convenient locking mechanism that can be employed at the press
     *                of a button.
     */
    public void enableIdleAngle(boolean enabled) {
        this.idleAngleEnabled = enabled;
    }

    /**
     * The fastest velocity that this base can achieve, based on its slowest swerve
     * module. This will iterate over the module map and calculate the max velocity
     * of each module, then return the slowest velocity.
     * 
     * <p>
     * <b>Note to module developers:</b> It is important to make sure your module's
     * implementation reports an accurate maximum velocity, because it is necessary
     * in the process of converting relative velocity vectors into absolute velocity
     * vectors. {@link net.bancino.robotics.swerveio.geometry.SwerveVector} is
     * considered a <i>percentage vector</i>, which means it provides a percentage
     * of the maximum velocity. WPILib's {@code ChassisSpeeds} vector is absolute;
     * it provides exact measurements. So, to convert from one to the other, it is
     * necessary to have accurate maximum velocities, even if they are theoretical.
     * </p>
     *
     * @return The velocity of the slowest swerve module, which will be all the
     *         faster this swerve drive can go, in <b>meters per second</b>.
     * 
     * @see net.bancino.robotics.swerveio.module.SwerveModule#getMaxSpeed()
     */
    public double getChassisMaxSpeed() {
        double maxSpeed = 0;
        for (SwerveModule module : moduleMap.values()) {
            double speed = module.getMaxSpeed();
            if (speed > maxSpeed) {
                maxSpeed = speed;
            }
        }
        return maxSpeed;
    }

    /**
     * The fastest angular velocity that this base can achieve, based on
     * {@link #getChassisMaxSpeed()} and the dimensions of the chassis itself.
     *
     * @return The maximum angular velocity in radians per second that this swerve
     *         drive can achieve.
     */
    public double getChassisMaxOmega() {
        double maxRevsPerSecond = (calc.getChassisDimensions().getRadius().get(Unit.METERS) * Math.PI)
                / getChassisMaxSpeed();
        return maxRevsPerSecond * (2 * Math.PI);
    }

    /**
     * Drive the entire chassis with a WPILib ChassisSpeeds object. This converts
     * the chassis speed object into a raw swerve vector using
     * {@link net.bancino.robotics.swerveio.geometry.SwerveVector#fromChassisSpeeds(ChassisSpeeds,SwerveDrive)}.
     * 
     * <p>
     * This method would be the most useful for autonomous mode, as it provides an
     * easy way to drive the swerve with vectors generated by WPILib utilities.
     * </p>
     *
     * @param v A WPILib ChassisSpeeds vector that represents the velocity to apply
     *          to this swerve base. See
     *          {@link net.bancino.robotics.swerveio.command.PathweaverSwerveDrive}.
     */
    public void drive(ChassisSpeeds v) {
        SwerveVector swerveVector = SwerveVector.fromChassisSpeeds(v, this);
        drive(swerveVector);
    }

    /**
     * Drive the entire chassis with a
     * {@link net.bancino.robotics.swerveio.geometry.SwerveVector} and a gyro angle
     * setpoint. This method requires that a functional gyro and a tuned PID
     * controller are passed via the builder.
     *
     * <p>
     * This method works by passing the current gyro angle and the specified target
     * angle into the gyro PID controller and generating a new rotational parameter
     * for the current scan. The rotational parameter provided by the swerve vector
     * is discarded and replaced with the output of the gyro angle controller before
     * being passed down to {@link #drive(SwerveVector)}.
     * </p>
     *
     * <p>
     * Many possible use cases for this exist; it can be useful in creating an
     * angle-based joystick control loop that can correct for unwanted drift, or it
     * can be used in an autonomous mode to ensure trajectory compass headings are
     * reached.
     * </p>
     *
     * @param v           The swerve vector to drive by. This used for the forward
     *                    (Y) and strafe (X) parameters only. The rotational
     *                    parameter (Z) is overwritten by a PID controller's
     *                    calculations regarding the desired angle of the chassis.
     * @param targetAngle The desired angle to reach on the gyro. This is passed
     *                    into a pre-tuned gyro PID controller and is used to
     *                    produce a new rotational parameter. It must be a value on
     *                    the interval [0,360).
     *
     * @throws IllegalStateException    If the gyro or gyro PID controller are null.
     * @throws IllegalArgumentException If the target angle is out of the acceptable
     *                                  bounds.
     */
    public void drive(SwerveVector v, double targetAngle) {
        if (gyro == null) {
            throw new IllegalStateException("A gyro is required to use this method. Add one with Builder.setGyro().");
        }

        if (gyroPID == null) {
            throw new IllegalStateException(
                    "A gyro position controller is required to use this method. Add one with Builder.setAnglePID().");
        }

        if (targetAngle < 0 || targetAngle >= 360) {
            throw new IllegalArgumentException("Angle setpoint out of bounds: " + targetAngle);
        }

        /*
         * If the swerve drive is moving, use the moving profile slot. Otherwise, use
         * the standing profile slot.
         */
        // double speedThreshold = 0.05;
        // if (v.getFwd() > speedThreshold || v.getFwd() < -speedThreshold || v.getStr()
        // > speedThreshold
        // || v.getStr() < -speedThreshold || v.getRcw() > speedThreshold || v.getRcw()
        // < -speedThreshold) {
        // gyroPID.setActiveProfile(gyroMovingSlot);
        // } else {
        // gyroPID.setActiveProfile(gyroStandingSlot);
        // }

        /*
         * correctPivotFeedback() was originally intended only for the swerve module PID
         * controller, but the gyro angle is the same problem, so we can use the same
         * solution.
         */
        double adjustedGyroAngle = SwerveModule.correctPivotFeedback(gyro.getAngle(), targetAngle, 360);

        double error = Math.abs(adjustedGyroAngle - targetAngle);
        double errorThreshhold = 7; // Degrees
        if (error > errorThreshhold) {
            gyroSlotChanged = gyroPID.getActiveProfile() != gyroStandingSlot;
            gyroPID.setActiveProfile(gyroStandingSlot);
        } else {
            gyroSlotChanged = gyroPID.getActiveProfile() != gyroMovingSlot;
            gyroPID.setActiveProfile(gyroMovingSlot);
        }

        /* Reset the PID controller if we switched slots */
        if (gyroSlotChanged) {
            gyroPID.reset();
        }

        double rcw = gyroPID.getOutput(adjustedGyroAngle, targetAngle);

        double outputThreshhold = getModule(SwerveModule.Location.FRONT_RIGHT).getOutputThreshhold();
        /*
         * If there is minimal to no forward or strafe movement and the PID output is
         * less than the minimum allowed by the motor controllers, don't use the PID
         * output.
         * 
         * The first condition is applied because the problem this logic solves only
         * occurs if the robot is performing a still rotation. If the robot is in
         * motion, this isn't a problem.
         */
        if ((Math.abs(v.getFwd()) < outputThreshhold || Math.abs(v.getStr()) < outputThreshhold)
                && Math.abs(rcw) < outputThreshhold) {
            v.setRcw(0);
        } else {
            v.setRcw(rcw);
        }

        /* Pass the modified vector down to the actual drive method. */
        drive(v);
    }

    /**
     * Drive the entire chassis with a ChassisSpeeds object and a gyro angle
     * setpoint. This method requires that a functional gyro and a tuned PID
     * controller are passed via the builder.
     *
     * <p>
     * This method works by converting the ChassisSpeds into a SwerveVector and
     * passing the current gyro angle and the specified target angle into the gyro
     * PID controller and generating a new rotational parameter for the current
     * scan. The rotational parameter provided by the swerve vector is discarded and
     * replaced with the output of the gyro angle controller before being passed
     * down to {@link #drive(SwerveVector,double)}.
     * </p>
     *
     * <p>
     * Many possible use cases for this exist; it can be useful in creating an
     * angle-based joystick control loop that can correct for unwanted drift, or it
     * can be used in an autonomous mode to ensure trajectory compass headings are
     * reached.
     * </p>
     *
     * @param v           The chassis vector to drive by. This used for the forward
     *                    (Y) and strafe (X) parameters only. The rotational
     *                    parameter (Z) is overwritten by a PID controller's
     *                    calculations regarding the desired angle of the chassis.
     * @param targetAngle The desired angle to reach on the gyro. This is passed
     *                    into a pre-tuned gyro PID controller and is used to
     *                    produce a new rotational parameter.
     *
     * @throws IllegalStateException if the gyro or gyro PID controller are null.
     */
    public void drive(ChassisSpeeds v, double targetAngle) {
        SwerveVector swerveVector = SwerveVector.fromChassisSpeeds(v, this);
        drive(swerveVector, targetAngle);
    }

    /**
     * Drive the entire chassis with a
     * {@link net.bancino.robotics.swerveio.geometry.SwerveVector}, a relative
     * vector designed to be generated from a joystick. The references in this
     * vector are passed directly down to the kinematics provider for conversion to
     * module vectors.
     * 
     * @param v The vector (X, Y, Z). See
     *          {@link net.bancino.robotics.swerveio.geometry.SwerveVector}, and
     *          {@link net.bancino.robotics.swerveio.command.SwerveDriveTeleop},
     *          which generates these vectors.
     */
    public void drive(SwerveVector v) {
        /* Check for reversed input on the parameters and adjust accordingly. */
        if (reversedInput.get(DegreeOfFreedom.FORWARD)) {
            v.setFwd(v.getFwd() * -1);
        }
        if (reversedInput.get(DegreeOfFreedom.STRAFE)) {
            v.setStr(v.getStr() * -1);
        }
        if (reversedInput.get(DegreeOfFreedom.ROTATION)) {
            v.setRcw(v.getRcw() * -1);
        }

        /* Get the gyro reading, or zero if we aren't using one. */
        double gyroAngle = ((gyro != null) && fieldCentric) ? gyro.getAngle() : 0;

        forEachModule((location, module) -> {
            /* Use the swerve drive calculator to calculate target speeds and angles. */
            ModuleVector moduleVector = calc.getModuleVector(location, v, gyroAngle);
            double speed = moduleVector.getSpeed();
            double targetAngle = moduleVector.getAngle();

            /* Get a reference to the module to get feedback from it. */
            SwerveModule swerveModule = getModule(location);
            double currentAngle = swerveModule.getAngle();

            if (!idleAngleEnabled) {
                /*
                 * Set the target angle of the pivot motor to what it was before if there's no
                 * new output.
                 */
                if (speed < swerveModule.getOutputThreshhold()) {
                    if (lastPivotAngle.containsKey(location)) {
                        targetAngle = lastPivotAngle.get(location);
                    }
                } else {
                    lastPivotAngle.put(location, currentAngle);
                }
            } else { /* Idle-Angle mode */
                /*
                 * If the calculated drive speed for this module is less than that module's
                 * minimum output threshhold then snap to the idle angle, not the calculated
                 * angle. Otherwise, drive.
                 */
                if (speed < swerveModule.getOutputThreshhold()) {
                    switch (location) {
                        case FRONT_LEFT:
                        case REAR_RIGHT:
                            targetAngle = idleAngle;
                            break;
                        case FRONT_RIGHT:
                        case REAR_LEFT:
                            targetAngle = diagonalFlipIdle ? 180 - idleAngle : idleAngle;
                            break;
                        default:
                            targetAngle = 0;
                            break;
                    }
                }
            }

            /* Compute the opposite angle before the optimization check. */
            double flippedAngle;
            if (targetAngle >= 180) {
                flippedAngle = targetAngle - 180;
            } else if (targetAngle < 180) {
                flippedAngle = targetAngle + 180;
            } else {
                flippedAngle = targetAngle;
            }

            /* This is a small optimization to reverse the angle if it is more efficient. */
            double diff = Math.abs(targetAngle - currentAngle);
            double flipDiff = Math.abs(flippedAngle - currentAngle);
            if (diff >= 90 && flipDiff <= 90) {
                targetAngle = flippedAngle;
                speed *= -1;
            }

            /* Pass the modified module vector down to the module. */
            moduleVector.setSpeed(speed).setAngle(targetAngle);
            swerveModule.drive(moduleVector);

            /* Record inputs for logging purposes. */
            lastSwerveVector = v;
        });
    }

    /**
     * Stop this swerve drive. This will immediately stop this swerve drive by
     * invoking {@link net.bancino.robotics.swerveio.module.SwerveModule#stop()} on
     * all swerve modules. This takes effect until {@link #drive(SwerveVector)} is
     * called again. At that point, the modules will begin movement again.
     */
    public void stop() {
        forEachModule((SwerveModule.Location location, SwerveModule module) -> {
            module.stop();
        });
    }

    private static String makePropKey(SwerveModule.Location location) {
        String asString = location.toString();
        int i;
        while ((i = asString.indexOf('_')) >= 0) {
            asString = asString.replace("_.", String.valueOf(Character.toUpperCase(asString.charAt(i + 1))));
        }
        asString += "AngleOffset";
        return asString;
    }

    /**
     * Dump the current raw angle of each swerve module to the given properties
     * object. This also updates the offset angle on each swerve module to be the
     * dumped angle. This method is helpful when calibrating your swerve drive's
     * encoders. Instead of having to manually program the offsets into your code,
     * or manually copying them into a config file, you can simply call this method.
     * 
     * @param props The properties object to dump the angle offsets to. This is
     *              helpful if you want to store the offsets in with the rest of
     *              your robot tuning parameters.
     */
    public void saveAngleOffsets(Properties props) {
        if (DriverStation.isDisabled()) {
            forEachModule((SwerveModule.Location location, SwerveModule module) -> {
                module.setAngleOffset(0);
                double angle = module.getAngle();
                props.setProperty(makePropKey(location), Double.toString(angle));
                module.setAngleOffset(angle);
            });
        } else {
            DriverStation.reportError("Angle offsets cannot be saved while the robot is enabled.", false);
        }
    }

    /**
     * Load raw angle offsets for each swerve module from the given properties
     * object. This immediately updates the angle offset on each swerve module. This
     * is helpful when initializing your swerve drive because you can directly load
     * the angle offsets previously saved by {@link #saveAngleOffsets(Properties)}.
     * 
     * @param props The properties object to load angle offsets from. This is
     *              helpful if you to store the offsets in with the rest of your
     *              robot tuning parameters.
     * 
     * @throws NullPointerException  If any angle offset property does not exist in
     *                               the given properties object.
     * @throws NumberFormatException If the value of any required offset property is
     *                               not a valid double.
     */
    public void loadAngleOffsets(Properties props) throws NullPointerException, NumberFormatException {
        for (SwerveModule.Location location : SwerveModule.Location.values()) {
            String rawOffset = props.getProperty(makePropKey(location));
            if (rawOffset != null) {
                double offset = Double.parseDouble(rawOffset);
                moduleMap.get(location).setAngleOffset(offset);
            } else {
                throw new NullPointerException(
                        "Property [" + makePropKey(location) + "] was not found in the given properties.");
            }
        }
    }

    /**
     * Dump the current raw angle of each swerve module to the given file. This also
     * updates the offset angle on each swerve module to be the dumped angle. This
     * method is helpful when calibrating your swerve drive's encoders. Instead of
     * having to manually program the offsets into your code, or manually copying
     * them into a config file, you can simply call this method.
     * 
     * @param propFile The file to dump the angle offsets to. This file, if it
     *                 exists, will be overwritten. The format of the file is the
     *                 standard Java properties format.
     * 
     * @throws IOException If there is an error opening the file for writing.
     */
    public void saveAngleOffsets(File propFile) throws IOException {
        Properties props = new Properties();
        saveAngleOffsets(props);
        props.store(new FileOutputStream(propFile), "SwerveIO Angle Offsets");
    }

    /**
     * Load raw angle offsets for each swerve module from the given file. This
     * immediately updates the angle offset on each swerve module. This is helpful
     * when initializing your swerve drive because you can directly load the angle
     * offsets previously saved by {@link #saveAngleOffsets(File)}.
     * 
     * @param propFile The file to load the angle offsets from.
     * 
     * @throws NullPointerException  If any angle offset property does not exist in
     *                               the given properties object.
     * @throws NumberFormatException If the value of any required offset property is
     *                               not a valid double.
     * @throws IOException           If there is an error opening the file for
     *                               reading.
     */
    public void loadAngleOffsets(File propFile) throws IOException {
        Properties props = new Properties();
        props.load(new FileInputStream(propFile));
        loadAngleOffsets(props);
    }

    /**
     * Dump the current raw angle of each swerve module to the default properties
     * file, as defined by {@link #DEFAULT_ANGLE_OFFSET_FILE}. This also updates the
     * offset angle on each swerve module to be the dumped angle. This method is
     * helpful when calibrating your swerve drive's encoders. Instead of having to
     * manually program the offsets into your code, or manually copying them into a
     * config file, you can simply call this method.
     * 
     * @throws IOException If there is an error opening the file for writing.
     */
    public void saveAngleOffsets() throws IOException {
        saveAngleOffsets(DEFAULT_ANGLE_OFFSET_FILE);
    }

    /**
     * Load raw angle offsets for each swerve module from the default properties
     * file, as defined by {@link #DEFAULT_ANGLE_OFFSET_FILE}. This immediately
     * updates the angle offset on each swerve module. This is helpful when
     * initializing your swerve drive because you can directly load the angle
     * offsets previously saved by {@link #saveAngleOffsets()}.
     * 
     * @throws NullPointerException  If any angle offset property does not exist in
     *                               the given properties object.
     * @throws NumberFormatException If the value of any required offset property is
     *                               not a valid double.
     * @throws IOException           If there is an error opening the file for
     *                               reading.
     */
    public void loadAngleOffsets() throws IOException {
        loadAngleOffsets(DEFAULT_ANGLE_OFFSET_FILE);
    }

    /**
     * The SubsystemBase periodic method is run once every scheduler routine, or
     * about every 20 milliseconds.
     */
    @Override
    public final void periodic() {
        /* Run a user-defined periodic function. */
        if (periodic != null) {
            periodic.run();
        }

        /* Logging */
        try {
            log.log(this);
        } catch (UnsupportedOperationException e) {
            DriverStation.reportWarning("Logging failed.", e.getStackTrace());
        }
    }

    /**
     * Get the logger associated with this class. This logger monitors this swerve
     * drive and recursively logs all values, depending on the log level set.
     * 
     * @return A logging instance.
     */
    public RobotLogger getLogger() {
        return log;
    }

    /**
     * Get the last point that was fed into this swerve drive. Contained in the
     * point are the forward, strafe, and rotation variables of this vector. This
     * will be exactly one scan behind what the actual received vector is is, unless
     * {@link #drive(SwerveVector)} is not being called iteratively. In any case,
     * this returns the last vector that was recorded by
     * {@link #drive(SwerveVector)}.
     * 
     * @return A Swerve vector.
     */
    public SwerveVector getLastSwerveVector() {
        return lastSwerveVector;
    }

    /**
     * Get the gyro object that this swerve drive was created with, if any. This can
     * be used to read values off the gyro being used in this swerve drive, as well
     * as to reset it if necessary, such as when the robot code starts.
     * 
     * @return The gyro object that this swerve drive is reading from. Null if none.
     */
    public Gyro getGyro() {
        return gyro;
    }

    /**
     * Get the PID Controller used to generate the rotational parameter in
     * {@link #drive(SwerveVector,double)}. This is
     *
     * @return The PID controller passed into this swerve drive from the builder, if
     *         any. It is very possible for this to be null.
     */
    public PIDController getAnglePID() {
        return gyroPID;
    }

    /**
     * Get the calculator that this swerve drive is using to compute the values it
     * needs to drive the swerve modules. To get this swerve drive's chassis
     * dimensions, use
     * {@link net.bancino.robotics.swerveio.kinematics.SwerveKinematicsProvider#getChassisDimensions()}.
     * 
     * @return A swerve drive calculator that is actively being used by the swerve
     *         drive.
     */
    public SwerveKinematicsProvider getKinematicsProvider() {
        return calc;
    }
}
