/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.command;

import edu.wpi.first.wpilibj.Filesystem;
import edu.wpi.first.math.trajectory.Trajectory;
import edu.wpi.first.math.trajectory.TrajectoryUtil;
import edu.wpi.first.math.kinematics.ChassisSpeeds;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import net.bancino.robotics.swerveio.SwerveDrive;
import net.bancino.robotics.swerveio.module.SwerveModule;

/**
 * Execute a PathWeaver trajectory on a SwerveIO Swerve Drive. This is intended
 * to greatly simplify autonomous development, because you can use the standard
 * WPILib PathWeaver to generate your trajectory and then just pass it here.
 * 
 * <p>
 * PathWeaver trajectories are fully supported as of version 5.0.0.
 * </p>
 * 
 * <p>
 * This command will execute until it is interrupted, or until the path has been
 * completed.
 * </p>
 *
 * @author Jordan Bancino
 * @version 6.0.0
 * @version 2.1.0
 */
public class PathweaverSwerveDrive extends SwerveDriveCommand {

    /**
     * Convert a Pathweaver state object into a ChassisSpeeds vector that can be
     * passed into a swerve drive.
     * 
     * @param pathweaverState The trajectory state to convert to the chassis speed
     *                        object.
     * @param angularVelocity The angular velocity to apply to the resultant vector.
     * @return A ChassisSpeeds object that has the velocity components of the
     *         current state object, and the constant angular velocity component.
     */
    protected static ChassisSpeeds convertToChassisSpeeds(Trajectory.State pathweaverState, double angularVelocity) {
        double angle = pathweaverState.poseMeters.getRotation().getRadians();
        double velocity = pathweaverState.velocityMetersPerSecond;
        double velocityX = velocity * Math.cos(angle);
        double velocityY = velocity * Math.sin(angle);

        return new ChassisSpeeds(velocityX, velocityY, angularVelocity);
    }

    /**
     * Compute the average angular velocity between two trajectory states.
     * Theoretically, we could and should take the derivative of each state to get
     * the angular velocity, but it simply isn't worth the computing power. If the
     * states passed to this function are close enough, you'll get an extremely
     * close approximation of the angular velocity at the given point.
     * 
     * @param start The starting trajectory state to read the rotation vector from.
     * @param end   The ending trajectory state to read the rotation vector from.
     * 
     * @return The average angular velocity in radians per second that will
     *         accomplish the given rotation between the two states.
     */
    protected static double getAngularVelocity(Trajectory.State start, Trajectory.State end) {
        double r1 = end.poseMeters.getRotation().getRadians();
        double r2 = start.poseMeters.getRotation().getRadians();

        double t1 = end.timeSeconds;
        double t2 = start.timeSeconds;

        return (r1 - r2) / (t1 - t2);
    }

    /**
     * PathweaverSwerveDrive supports multiple path execution modes. These modes
     * allow you to alter the way the Pathweaver trajectory is executed, based on
     * the orientation of the robot, or the side of the field you are on.
     */
    public static enum PathExecutionMode {
        /**
         * Run the path without modification. This will execute the raw Pathweaver
         * trajectory without any altering of the {@code ChassisSpeeds} objects that it
         * generates from {@link #convertToChassisSpeeds(Trajectory.State, double)}.
         */
        NORMAL,
        /**
         * Run the path on the assumption that the robot is oriented backwards, or
         * rotated 180 degrees from the front position. This will invert the forward
         * parameter on the chassis speed vector.
         */
        ROBOT_BACKWARDS,
        /**
         * Run the entire path backwards. This will invert both the forward and strafe
         * parameters on the chassis speed vector.
         */
        FULLY_BACKWARDS
    }

    private final Trajectory trajectory;
    private final PathExecutionMode executionMode;

    private Trajectory.State previousState;
    private long startTime;
    private double calTime = 0.75;

    private final boolean useGyro;
    private final boolean doRotation;

    /**
     * Create a Pathweaver swerve drive command that will execute the given
     * pathweaver path on the given swerve drive.
     *
     * @param swerve         The swerve drive to drive.
     * @param pathweaverJson The pathweaver file to load, relative to the deploy
     *                       directory.
     * @param executionMode  The mode to execute the path in.
     * @throws IOException If there is an IO error when reading the pathweaver json
     *                     file.
     */
    public PathweaverSwerveDrive(SwerveDrive swerve, String pathweaverJson, PathExecutionMode executionMode)
            throws IOException {
        this(swerve, pathweaverJson, executionMode, true);
    }

    /**
     * Create a Pathweaver swerve drive command that will execute the given
     * pathweaver path on the given swerve drive.
     *
     * @param swerve         The swerve drive to drive.
     * @param pathweaverJson The pathweaver file to load, relative to the deploy
     *                       directory.
     * @param executionMode  The mode to execute the path in.
     * @param doRotation     Whether or not to enable the rotation component of the
     *                       path. Setting this to false will prohibit the swerve
     *                       drive from rotating while executing the path.
     * @throws IOException If there is an IO error when reading the pathweaver json
     *                     file.
     */
    public PathweaverSwerveDrive(SwerveDrive swerve, File pathweaverJson, PathExecutionMode executionMode,
            boolean doRotation) throws IOException {
        super(swerve);
        if (pathweaverJson != null) {
            Path jsonPath = pathweaverJson.toPath();
            trajectory = TrajectoryUtil.fromPathweaverJson(jsonPath);
        } else {
            throw new IllegalArgumentException("Pathweaver JSON file name cannot be null.");
        }

        if (executionMode != null) {
            this.executionMode = executionMode;
        } else {
            throw new IllegalArgumentException("Execution mode cannot be null.");
        }

        this.doRotation = doRotation;
        this.useGyro = swerve.getGyro() != null && swerve.getAnglePID() != null;
    }

    /**
     * Create a Pathweaver swerve drive command that will execute the given
     * pathweaver path on the given swerve drive.
     *
     * @param swerve         The swerve drive to drive.
     * @param pathweaverJson The pathweaver file to load, relative to the deploy
     *                       directory.
     * @param executionMode  The mode to execute the path in.
     * @param doRotation     Whether or not to enable the rotation component of the
     *                       path. Setting this to false will prohibit the swerve
     *                       drive from rotating while executing the path.
     * @throws IOException If there is an IO error when reading the pathweaver json
     *                     file.
     */
    public PathweaverSwerveDrive(SwerveDrive swerve, String pathweaverJson, PathExecutionMode executionMode,
            boolean doRotation) throws IOException {
        this(swerve, new File(Filesystem.getDeployDirectory(), pathweaverJson), executionMode, doRotation);
    }

    /**
     * Create a Pathweaver swerve drive command that will execute the given
     * pathweaver path on the given swerve drive.
     *
     * @param swerve         The swerve drive to drive.
     * @param pathweaverJson The pathweaver file to load, relative to the deploy
     *                       directory.
     * @param doRotate       Whether or not to enable the rotation component of the
     *                       path. Setting this to false will prohibit the swerve
     *                       drive from rotating while executing the path.
     * @throws IOException If there is an IO error when reading the pathweaver json
     *                     file.
     */
    public PathweaverSwerveDrive(SwerveDrive swerve, String pathweaverJson, boolean doRotate) throws IOException {
        this(swerve, pathweaverJson, PathExecutionMode.NORMAL, doRotate);
    }

    /**
     * Create a Pathweaver swerve drive that uses the normal path execution mode.
     *
     * @param swerve         The swerve drive object to drive.
     * @param pathweaverJson The pathweaver file to load, relative to the deploy
     *                       directory.
     * @throws IOException If there is an IO error when reading the pathweaver json
     *                     file.
     */
    public PathweaverSwerveDrive(SwerveDrive swerve, String pathweaverJson) throws IOException {
        this(swerve, pathweaverJson, true);
    }

    /**
     * Set the calibration time on this path. This parameter controls how long this
     * command will instruct the swerve drive modules to pivot to the forward
     * position before the path is run. This is helpful for autonomous because it
     * ensures the swerve modules are aligned (or roughly aligned) before attempting
     * to drive. This will result in drastically more accurate paths.
     * 
     * @param calTime The calibration time, <b>in seconds</b>. The swerve drive will
     *                not move during this time, instead the control loops will be
     *                used to drive the pivot motors to the starting position. After
     *                this time is up, the full path is run. The default calibration
     *                time is 0.75 seconds, which is quite generous for a well-tuned
     *                pivot PID loop.
     */
    public void setCalibrationTime(double calTime) {
        this.calTime = Math.abs(calTime);
    }

    @Override
    protected void initialize(SwerveDrive swerve) {
        startTime = System.currentTimeMillis();
    }

    private double getTimeSeconds() {
        long currentTime = System.currentTimeMillis();
        double timeSeconds = (double) (currentTime - startTime) / 1000d;
        return timeSeconds - this.calTime;
    }

    @Override
    protected void execute(SwerveDrive swerve) {
        double timeSeconds = getTimeSeconds();

        /*
         * If our current time is less than zero, we are in the calibration period. To
         * calibrate, we simply set all the module angles to zero. The calibration time
         * is important because it will take a few scans to actually achieve the angle.
         */
        if (timeSeconds < 0) {
            swerve.forEachModule((SwerveModule.Location location, SwerveModule module) -> {
                module.setAngle(0);
            });
            return;
        }

        /* Get the current trajectory state from the trajectory. */
        Trajectory.State currentState = trajectory.sample(timeSeconds);

        /*
         * Approximate the angular velocity by using the previous state we passed. It
         * would not be efficient to compute the derivative to get the actual
         * instantanious velocity, but this gets us close enough.
         * 
         * This code is only used if the gyro is null, because if a gyro is present, we
         * will be using a PID controller to ensure the angle is actually set right.
         */
        double angularVelocity;
        if (doRotation && !useGyro && previousState != null) {
            angularVelocity = getAngularVelocity(previousState, currentState);
        } else {
            angularVelocity = 0;
        }

        ChassisSpeeds swerveVector = convertToChassisSpeeds(currentState, angularVelocity);

        /* Flip the chassis speed axes based on the direction the path should go. */
        switch (executionMode) {
        case NORMAL:
            swerveVector.vyMetersPerSecond *= -1;
            break;
        case ROBOT_BACKWARDS:
            swerveVector.vyMetersPerSecond *= -1;
            swerveVector.vxMetersPerSecond *= -1;
            break;
        case FULLY_BACKWARDS:
            swerveVector.vxMetersPerSecond *= -1;
            break;
        default:
            throw new IllegalStateException("Unknown execution mode.");
        }

        if (useGyro) {
            double targetAngle = 0;
            if (doRotation) {
                targetAngle = currentState.poseMeters.getRotation().getDegrees();
            }
            swerve.drive(swerveVector, targetAngle);
        } else {
            swerve.drive(swerveVector);
        }

        previousState = currentState;
    }

    @Override
    public boolean isFinished() {
        return trajectory.getTotalTimeSeconds() <= getTimeSeconds();
    }
}
