/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */

package net.bancino.robotics.swerveio.command;

import java.io.IOException;

import edu.wpi.first.wpilibj.DriverStation;
import edu.wpi.first.wpilibj2.command.InstantCommand;
import net.bancino.robotics.swerveio.SwerveDrive;

/**
 * Execute {@link net.bancino.robotics.swerveio.SwerveDrive#saveAngleOffsets()}
 * on a swerve drive. This simple command is set up to properly save the angle
 * offsets while the robot is disabled. You can add this command to the smart
 * dashboard, or even a joystick button (not recommended) to easily calibrate
 * your swerve drive.
 * 
 * <p>
 * The general usage of this command is simple: align your swerve modules so
 * that they are all facing perfectly straight in the forward position, then
 * invoke this command to save their positions. You can then call
 * {@link net.bancino.robotics.swerveio.SwerveDrive#loadAngleOffsets()} in your
 * swerve drive's initialization code to load the saved offsets.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class SaveSwerveAngleOffsets extends InstantCommand {

    /**
     * Construct a new offset-saving command.
     * 
     * @param swerve The given swerve drive to save the angle offsets for.
     */
    public SaveSwerveAngleOffsets(SwerveDrive swerve) {
        super(() -> {
            try {
                swerve.saveAngleOffsets();
                DriverStation.reportWarning("Saved swerve drive angle offsets.", false);

            } catch (IOException e) {
                DriverStation.reportError("Unable to save swerve angle offsets", e.getStackTrace());
            }
        }, swerve);
    }

    /**
     * This signals to the command scheduler that this command is allowed to run
     * when the robot is disabled.
     */
    @Override
    public boolean runsWhenDisabled() {
        return true;
    }
}
