/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.command;

import edu.wpi.first.wpilibj2.command.CommandBase;
import net.bancino.robotics.swerveio.SwerveDrive;

/**
 * A simple wrapper for the WPILib {@code CommandBase} that runs a command on a
 * SwerveIO swerve drive. This class provides minimal functionality. It simply
 * accepts a swerve drive in a constructor and then iteratively delivers that
 * swerve drive into the abstract methods for convenience.
 * 
 * <p>
 * This class was written with the intention of easing development of commands
 * that operate on a swerve drive by removing boilerplate swerve drive code from
 * those commands. Wherever possible, it is recommended that all commands that
 * operate on a SwerveIO swerve drive extend this class. This class provides the
 * following functions:
 * </p>
 * <ul>
 * <li>It automatically stores the swerve drive object so you don't have to and
 * passes it into the command functions you need.</li>
 * <li>It automatically adds the swerve drive as a subsystem requirement so you
 * don't have to.</li>
 * </ul>
 * 
 * @author Jordan Bancino
 * @version 5.0.4
 * @since 5.0.4
 */
public abstract class SwerveDriveCommand extends CommandBase {

    private final SwerveDrive swerve;

    /**
     * Construct a new swerve drive command with the given swerve drive. This
     * command will operate on this swerve drive.
     * 
     * @param swerve The swerve drive object that this command will use for
     *               performing swerve-related functions.
     */
    protected SwerveDriveCommand(SwerveDrive swerve) {
        if (swerve != null) {
            this.swerve = swerve;
            addRequirements(swerve);
        } else {
            throw new NullPointerException("Swerve drive cannot be null.");
        }
    }

    @Override
    public final void initialize() {
        super.initialize();
        initialize(swerve);
    }

    @Override
    public final void execute() {
        super.execute();
        execute(swerve);
    }

    @Override
    public final void end(boolean interrupted) {
        super.end(interrupted);
        end(swerve, interrupted);
    }

    /**
     * Initialize this command. This method is called once when the command is
     * initially scheduled. Use this for any state setting you need to do, such as
     * disabling field-centric drive if it is enabled.
     * 
     * @param swerve The swerve drive to initialize this command with. This is the
     *               same swerve drive passed into the constructor.
     */
    protected abstract void initialize(SwerveDrive swerve);

    /**
     * The main body of the command. This method is called iteratively while this
     * command is scheduled. You should put your swerve driving code here. The idea
     * is that eventually you'll call a {@code drive()} method on the passed swerve
     * drive, but swerve drive commands can also be state-setting commands.
     * 
     * @param swerve The swerve drive to execute this command with. This is the same
     *               swerve drive passed into the constructor.
     */
    protected abstract void execute(SwerveDrive swerve);

    /**
     * The action to take when the command ends. Called when either the command
     * finishes normally, or when it is interrupted/canceled.
     * 
     * @param swerve      The swerve drive that was passed into the constructor.
     * @param interrupted Whether the command was interrupted/canceled
     */
    protected void end(SwerveDrive swerve, boolean interrupted) {
        /*
         * It isn't common to implement this, so provide a default implementation that
         * does nothing.
         */
    }

    /*
     * Overriding this method to make it abstract requires subclasses to implement
     * it, which will help prevent undesired behavior.
     */
    @Override
    public abstract boolean isFinished();
}
