/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.command;

import edu.wpi.first.wpilibj.GenericHID;
import edu.wpi.first.wpilibj.Joystick;
import edu.wpi.first.wpilibj.XboxController;

import net.bancino.robotics.swerveio.SwerveDrive;
import net.bancino.robotics.swerveio.geometry.SwerveVector;

/**
 * A simple command for driving a swerve drive with any WPILib-supported
 * joystick.
 *
 * <p>
 * This command also demonstrates how simple it is to drive a SwerveIO swerve
 * drive with a joystick. It simply extends {@link SwerveDriveTeleopCommand},
 * which provides support for many useful features.
 * </p>
 *
 * <p>
 * This command runs until it is interrupted; there is no condition under which
 * this command will stop on it's own. The command scheduler should resume it
 * after autonomous commands are executed.
 * </p>
 *
 * <p>
 * As of SwerveIO 5.0.2, this joystick command uses the gyro for angular
 * movement if it is available and an angle PID controller is set up. This will
 * prevent any drifting that may occur because a position loop will be running
 * as the rotational parameter. See
 * {@link net.bancino.robotics.swerveio.SwerveDrive#drive(SwerveVector, double)}
 * and
 * {@link net.bancino.robotics.swerveio.SwerveDrive.Builder#setAnglePID(PIDController,int,int,Consumer)}
 * for more information on this.
 * </p>
 *
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 2.0.0
 */
public final class SwerveDriveTeleop extends SwerveDriveTeleopCommand {

    private final boolean useGyro;
    private double angleSetpoint;
    private double degreeIncrement = 8;

    /**
     * Create a new teleop command using a generic joystick and the default
     * {@link edu.wpi.first.wpilibj.Joystick.AxisType} axes. The defaults are
     * {@code kY} for forward, {@code kX} for strafe, and {@code kZ} for rotation.
     *
     * <p>
     * See
     * {@link SwerveDriveTeleop#SwerveDriveTeleop(SwerveDrive, Joystick, Joystick.AxisType, Joystick.AxisType, Joystick.AxisType)}
     * if you want to use custom axes.
     * </p>
     *
     * @param swerve   The swerve drive to drive.
     * @param joystick The joystick to read values from.
     */
    public SwerveDriveTeleop(SwerveDrive swerve, Joystick joystick) {
        this(swerve, joystick, Joystick.AxisType.kY, Joystick.AxisType.kX, Joystick.AxisType.kZ);
    }

    /**
     * Create a new teleop command using an Xbox controller and the default
     * {@link edu.wpi.first.wpilibj.XboxController.Axis} axes. The defaults are
     * {@code kLeftY} for forward, {@code kLeftX} for strafe, and {@code kRightX}
     * for rotation.
     *
     * <p>
     * See
     * {@link SwerveDriveTeleop#SwerveDriveTeleop(SwerveDrive, XboxController, XboxController.Axis, XboxController.Axis, XboxController.Axis)}
     * if you want to use custom axes.
     * </p>
     *
     * @param swerve   The swerve drive to drive.
     * @param joystick The Xbox controller to read values from.
     */
    public SwerveDriveTeleop(SwerveDrive swerve, XboxController joystick) {
        this(swerve, joystick, XboxController.Axis.kLeftY, XboxController.Axis.kLeftX, XboxController.Axis.kRightX);
    }

    /**
     * Create a new teleop command using an Xbox controller and custom axes.
     *
     * @param swerve   The swerve drive to drive.
     * @param joystick The Xbox controller to read from.
     * @param fwdAxis  The axis that controls the foward parameter.
     * @param strAxis  The axis that controls the strafe parameter.
     * @param rcwAxis  The axis that controls the rotation parameter.
     */
    public SwerveDriveTeleop(SwerveDrive swerve, XboxController joystick, XboxController.Axis fwdAxis,
            XboxController.Axis strAxis, XboxController.Axis rcwAxis) {
        this(swerve, joystick, fwdAxis.value, strAxis.value, rcwAxis.value);
    }

    /**
     * Create a new teleop command using a generic joystick and custom axes.
     *
     * @param swerve   The swerve drive to drive.
     * @param joystick The joystick to read from.
     * @param fwdAxis  The axis that controls the foward parameter.
     * @param strAxis  The axis that controls the strafe parameter.
     * @param rcwAxis  The axis that controls the rotation parameter.
     */
    public SwerveDriveTeleop(SwerveDrive swerve, Joystick joystick, Joystick.AxisType fwdAxis,
            Joystick.AxisType strAxis, Joystick.AxisType rcwAxis) {
        this(swerve, joystick, fwdAxis.value, strAxis.value, rcwAxis.value);
    }

    /**
     * Create a new teleop command using any generic human interaction device and
     * custom axes.
     *
     * @param swerve   The swerve drive to drive.
     * @param joystick The Xbox controller to read from.
     * @param fwdAxis  The axis that controls the foward parameter.
     * @param strAxis  The axis that controls the strafe parameter.
     * @param rcwAxis  The axis that controls the rotation parameter.
     */
    public SwerveDriveTeleop(SwerveDrive swerve, GenericHID joystick, int fwdAxis, int strAxis, int rcwAxis) {
        super(swerve, joystick, fwdAxis, strAxis, rcwAxis);
        this.useGyro = swerve.getGyro() != null && swerve.getAnglePID() != null;
    }

    /**
     * Set the degree increment that the joystick modifies. This is the maximum
     * number of degrees that the angle setpoint will change per scan if the
     * joystick is at 100% rotation. Note that this is applied after the joystick
     * ramp rate.
     * 
     * <p>
     * <b>Note:</b> This is only applicable if the swerve drive angle PID controller
     * is enabled. This increments the angle setpoint, but if there is no PID
     * controller, the angle setpoint is not used.
     * </p>
     * 
     * @param degreeIncrement The number of degrees, per scan, to change by when the
     *                        robot joystick's rotational axis is at 100%. If the
     *                        rotation is less than this, the increment actual will
     *                        be proportional. Lower values mean the swerve drive
     *                        will be less responsive to rotation, whereas higher
     *                        values will result in a touchier robot. The default
     *                        value is 8 degrees.
     */
    public void setAngleIncrement(double degreeIncrement) {
        this.degreeIncrement = Math.abs(degreeIncrement);
    }

    @Override
    protected void initialize(SwerveDrive swerve) {
        /*
         * Set the angle setpoint to the current gyro angle so this command doesn't move
         * the robot upon starting.
         */
        if (useGyro) {
            angleSetpoint = swerve.getGyro().getAngle();
        }
    }

    @Override
    protected void execute(SwerveDrive swerve, SwerveVector joystickVector) {
        if (useGyro) {
            /* Modify the angle setpoint based on the rotational joystick parameter. */
            angleSetpoint += joystickVector.getRcw() * this.degreeIncrement;
            angleSetpoint %= 360;
            if (angleSetpoint < 0) {
                angleSetpoint += 360;
            }

            /* Drive the swerve drive with the angle setpoint. */
            swerve.drive(joystickVector, angleSetpoint);
        } else {
            /* Simply drive the swerve drive with the joystick vector */
            swerve.drive(joystickVector);
        }
    }

    @Override
    public boolean isFinished() {
        return false;
    }
}
