/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.command;

import edu.wpi.first.wpilibj.GenericHID;

import net.bancino.robotics.swerveio.SwerveDrive;
import net.bancino.robotics.swerveio.geometry.SwerveVector;

/**
 * A simple abstract class for writing swerve drive teleop commands. This class
 * eliminates all the boilerplate code involved with writing a joystick command
 * but still allows you to utilize and modify joysick input as needed before
 * passing it to the swerve drive. This class is useful for writing commands
 * that drive a swerve drive based on input from a joystick. It automatically
 * takes care of throttling and deadbanding the joystick, as well as generating
 * a swerve vector, so you can directly pass the vector down to the swerve
 * drive, or perform modifications based on sensor readings, such as a gyro or
 * or Limelight.
 * 
 * <p>
 * The options for setting the deadband around zero, as well as the throttle,
 * can be great for training users.
 * </p>
 * 
 * <p>
 * When writing commands for a SwerveIO swerve drive that involve user joystick
 * input, it is highly recommended to use this class, because it automatically
 * handles many useful features and allows you to inherit the axes, throttle,
 * and deadband used in another joystick command, allowing for a seemless
 * transition across commands that the joystick operator won't even notice.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.4
 * @since 5.0.4
 */
public abstract class SwerveDriveTeleopCommand extends SwerveDriveCommand {

    private final GenericHID joystick;
    private final int fwdAxis, strAxis, rcwAxis;

    private double deadband, throttle, rampRate;

    private SwerveVector lastVector = new SwerveVector(0, 0, 0);

    /**
     * Construt a new swerve drive teleop command that inherits all the properties
     * of an already-existing joystick command. This allows you to configure your
     * joystick once with one command, then have all the other ones automatically
     * inherit the settings set on that "master" command. Most commonly, you'll use
     * this constructor to inherit a {@link SwerveDriveTeleop} command.
     * 
     * @param swerve  The swerve drive to pass to the superclass constructor and the
     *                abstract execute method. This is the swerve drive your
     *                joystick command will drive.
     * @param inherit The swerve drive teleop command to inherit the joystick,
     *                joystick axes, deadband, and throttle from. The joystick is
     *                inherited by-reference, but the axes, deadband, and throttle
     *                are inherited by-value. This means that modifying the joystick
     *                in another command will cause it to be modified in this
     *                command too, but changing the deadband or throttle on another
     *                command will not affect this command.
     */
    protected SwerveDriveTeleopCommand(SwerveDrive swerve, SwerveDriveTeleopCommand inherit) {
        super(swerve);

        this.joystick = inherit.joystick;
        this.fwdAxis = inherit.fwdAxis;
        this.strAxis = inherit.strAxis;
        this.rcwAxis = inherit.rcwAxis;

        this.deadband = inherit.deadband;
        this.throttle = inherit.throttle;
        this.rampRate = inherit.rampRate;
    }

    /**
     * Create a new teleop command using any generic human interaction device and
     * custom axes. This sets a default throttle of 90%, a default deadband of 20%,
     * and a default ramp rate of 0.25% per scan. See the relevant documentation for
     * throttle and deadband to override these defaults.
     *
     * @param swerve   The swerve drive to pass to the superclass constructor and
     *                 the abstract execute method. This is the swerve drive your
     *                 joystick command will drive.
     * @param joystick The joystick to read from.
     * @param fwdAxis  The axis to read from that will provide values for the Y
     *                 movement of the swerve drive.
     * @param strAxis  The axis to read from that will provide values for the X
     *                 movement of the swerve drive.
     * @param rcwAxis  The axis to read from that will provide values for the
     *                 angular movement of the swerve drive.
     */
    protected SwerveDriveTeleopCommand(SwerveDrive swerve, GenericHID joystick, int fwdAxis, int strAxis, int rcwAxis) {
        super(swerve);
        if (joystick != null) {
            this.joystick = joystick;
            this.fwdAxis = fwdAxis;
            this.strAxis = strAxis;
            this.rcwAxis = rcwAxis;
        } else {
            throw new IllegalArgumentException("Xbox controller cannot be null.");
        }

        setDeadband(0.1);
        setThrottle(0.9);
        setRampRate(0.025);
    }

    @Override
    public final void execute(SwerveDrive swerve) {
        /* Grab the joystick parameters */
        double fwd = -throttle(deadband(joystick.getRawAxis(fwdAxis)));
        double str = throttle(deadband(joystick.getRawAxis(strAxis)));
        double rcw = throttle(deadband(joystick.getRawAxis(rcwAxis)));

        /* Create a swerve vector and ramp it. */
        SwerveVector joystickVector = ramp(new SwerveVector(fwd, str, rcw));

        /* We have to duplicate the vector because it may be modified down the road. */
        this.lastVector = new SwerveVector(joystickVector);

        /* Pass the vector to the abstract execute method. */
        execute(swerve, joystickVector);
    }

    /**
     * Set a deadband on the joystick. This is a little bit of range around the zero
     * mark that does absolutely nothing. This is helpful for joysticks that are
     * overly sensitive or don't always read zero in the neutral position.
     * 
     * @param deadband The deadband to set, between 0 and 1.
     */
    public final void setDeadband(double deadband) {
        this.deadband = deadband;
    }

    /**
     * Set a throttle on the joystick. This is helpful for limiting the top speed of
     * the swerve drive, which can be useful for training.
     * 
     * @param throttle The throttle to set. This is the maximum speed the joystick
     *                 should output. So, to throttle this command at 50% power,
     *                 you'd put in 0.5 for the throttle.
     */
    public final void setThrottle(double throttle) {
        this.throttle = throttle;
    }

    /**
     * Set a ramp rate on the joystick. This is helpful for preventing sharp changes
     * in speed or direction.
     * 
     * @param rampRate The ramp rate to set. This is the maximum amount that the
     *                 joystick vector is allowed to change by per scan. A value of
     *                 zero or less disables the ramp rate.
     */
    public final void setRampRate(double rampRate) {
        this.rampRate = rampRate;
    }

    /**
     * Throttle a raw value based on the currently set throttle.
     * 
     * @param raw The raw joystick value to scale.
     * @return A scaled joystick value.
     */
    private final double throttle(double raw) {
        return raw * throttle;
    }

    /**
     * Calculate a deadband
     * 
     * @param raw The input on the joystick to mod
     * @return The result of the mod.
     */
    private final double deadband(double raw) {
        /* Compute the deadband mod */
        if (raw < 0.0d) {
            if (raw <= -deadband) {
                return (raw + deadband) / (1 - deadband);
            } else {
                return 0.0d;
            }
        } else {
            if (raw >= deadband) {
                return (raw - deadband) / (1 - deadband);
            } else {
                return 0.0d;
            }
        }
    }

    /**
     * Calculate a ramp rate, using the history vector.
     * 
     * @param v The vector to ramp. All parameters are ramped the same.
     * @return The same vector, modified with the ramp rate.
     */
    private final SwerveVector ramp(SwerveVector v) {
        if (this.rampRate > 0) {
            double maxFwd = this.lastVector.getFwd() + this.rampRate;
            double minFwd = this.lastVector.getFwd() - this.rampRate;

            if (v.getFwd() > maxFwd) {
                v.setFwd(maxFwd);
            }

            if (v.getFwd() < minFwd) {
                v.setFwd(minFwd);
            }

            double maxStr = this.lastVector.getStr() + this.rampRate;
            double minStr = this.lastVector.getStr() - this.rampRate;

            if (v.getStr() > maxStr) {
                v.setStr(maxStr);
            }

            if (v.getStr() < minStr) {
                v.setStr(minStr);
            }

            double maxRcw = this.lastVector.getRcw() + this.rampRate;
            double minRcw = this.lastVector.getRcw() - this.rampRate;

            if (v.getRcw() > maxRcw) {
                v.setRcw(maxRcw);
            }

            if (v.getRcw() < minRcw) {
                v.setRcw(minRcw);
            }
        }
        return v;
    }

    /**
     * The main body of the command. This method is called iteratively while this
     * command is scheduled. You should put your swerve driving code here. The idea
     * is that you'll perform any modifications necessary on the vector, then
     * eventually call a {@code drive()} method on the pasted swerve drive with the
     * modified vector.
     * 
     * @param swerve         The swerve drive to execute this command with. This is
     *                       the same swerve drive passed into the constructor.
     * @param joystickVector A swerve drive vector from the joystick. This vector
     *                       has been throttled and deadbanded according to the
     *                       settings before it is passed here.
     */
    protected abstract void execute(SwerveDrive swerve, SwerveVector joystickVector);
}
