/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

import edu.wpi.first.wpilibj.AnalogInput;
import edu.wpi.first.wpilibj.RobotController;
import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * An analog encoder that connects to the RoboRIO. This class can be used to
 * access 5-volt encoders that connect to the analog IO ports on the RoboRIO.
 * 
 * <p>
 * This is a wrapper for WPILib's {@code AnalogInput}, which reads a 12-bit
 * number representing the voltage (that is, 0-5 volts).
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.0
 * @since 1.0.0
 */
@Loggable
public class AnalogEncoder implements Encoder {

    private AnalogInput encoder;

    @Log(as = "rawValue", atLevel = LogLevel.IMPORTANT)
    private double lastValue;

    /**
     * Create an analog encoder.
     * 
     * @param port The analog port on the roboRIO that this encoder is connected to.
     */
    public AnalogEncoder(int port) {
        encoder = new AnalogInput(port);
    }

    @Override
    public double get() {
        lastValue = encoder.getVoltage();
        return lastValue;
    }

    @Override
    public double countsPerRevolution() {
        /*
         * The "full" voltage may not always be the same due to load on the battery and
         * other factors, so instead of hard-coding "5", we pull the actual voltage on
         * the 5v rail.
         */
        return RobotController.getVoltage5V();
    }
}