/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

/**
 * The encoder representation used by SwerveIO. This is a complete definition of
 * the encoder API. Every encoder that is used with SwerveIO must implement this
 * interface.
 * 
 * <p>
 * The interface itself is fairly simplistic because there are only a few
 * methods and they can usually be implemented in a just a few lines of code.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.0
 * @since 1.0.0
 */
public interface Encoder {

    /**
     * Get the encoder's current position. Note that this may not always have double
     * precision. It could be an integer as well. There is no limit on the value of
     * the encoder, as long as it fits within the {@code double} data type.
     * 
     * @return The encoder's current position, directly from the hardware. This
     *         value should not be modified at all; it is a raw reading specified in
     *         arbitrary units representing an arbitrary position.
     */
    public double get();

    /**
     * Get the number of counts it takes this encoder to do one full revolution.
     * This is a number of arbitrary units that represents how many units it takes
     * to travel one full revolution of the encoder shaft.
     * 
     * @return How many counts it takes to perform one full revolution of the
     *         encoder shaft, <b>in the units of {@link #get()}</b>. This should
     *         always come from the encoder documentation; don't use an experimental
     *         value. Also remember that this revolutions of the encoder, not
     *         revolutions of the pivot mechanism, though these will be the same if
     *         the encoder is 1:1.
     */
    public double countsPerRevolution();
}