/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

/**
 * An abstract representation of a physical swerve drive chassis. Objects of this class
 * store the dimensions of the SwerveIO swerve base and provide methods for calculating
 * the diagonal ({@link #getDiagonal()}), as well as the radius ({@link #getRadius()}).
 * See the {@link #getRadius()} documentation for what the radius is.
 * 
 * <p>
 * Note that the chassis dimensions are measured from <b>module to module</b>, not
 * necessarily from physical end to physical end. A chassis may have physical dimensions
 * of 30 inches, but this class should represent the measurement of center of the module
 * wheel in one corner to the center of the module wheel in each adjacent corner, which
 * will most likely be less; say, 24.3 inches, for example.
 * </p>
 * 
 * <p>
 * The chassis dimensions should be accurate because they are used in the computation of
 * the swerve module vectors, as well as in the computation of the angular velocity of
 * the swerve drive itself. Innacurate chassis dimensions will result in undefined and 
 * most likely undesirable behavior.
 * </p>
 *
 * @author Jordan Bancino
 * @version 5.0.0
 * @since 2.0.0
 */
@Loggable
public class ChassisDimension {

    @Log(atLevel = LogLevel.IMPORTANT)
    private final Length trackWidth, trackLength;

    /**
     * Construct a chassis dimension with two lengths. This should be used for rectangular
     * chassis that have different side lengths, even if only slightly.
     *
     * @param trackWidth  The track width of the chassis, measured from left to
     *                    right.
     * @param trackLength The track length of the chassis, measured from front to
     *                    back.
     */
    public ChassisDimension(Length trackWidth, Length trackLength) {
        if (trackWidth != null) {
            this.trackWidth = trackWidth;
        } else {
            throw new IllegalArgumentException("A chassis must have a track width.");
        }
        if (trackLength != null) {
            this.trackLength = trackLength;
        } else {
            throw new IllegalArgumentException("A chassis must have a track length.");
        }
    }

    /**
     * Construct a square chassis. This can be used for a chassis that is confirmed to
     * be a perfect square.
     * 
     * @param length The length of both the track width and length.
     */
    public ChassisDimension(Length length) {
        this(length, length);
    }

    /**
     * Get the track width of this chassis.
     * 
     * @return The width, measured from left to right.
     */
    public Length getWidth() {
        return trackWidth;
    }

    /**
     * Get the track length of this chassis.
     *
     * @return the length, measured from back to front.
     */
    public Length getLength() {
        return trackLength;
    }

    /**
     * Get the radius of the circle that this base will produce when rotating.
     * This value is used in calculating the maximum angular velocity of a swerve drive.
     * This is the same as {@code {@link #getDiagonal()} / 2}.
     *
     * @return The radius, that is, the length from the center of the chassis to one
     *         of the corners. This will work even if the base isn't a square.
     */
    public Length getRadius() {
        double halfLength = trackLength.get(Unit.BASE_UNIT) / 2;
        double halfWidth = trackWidth.get(Unit.BASE_UNIT) / 2;
        double hypotenuse = Math.sqrt(Math.pow(halfLength, 2) + Math.pow(halfWidth, 2));
        return new Length(hypotenuse, Unit.BASE_UNIT);
    }

    /**
     * Get the diagonal length that this base has, measured from opposite corners
     * of the base. This value is used in the kinematics calculations.
     * 
     * @return The diagonal length, that is, the length from one corner of the
     *         chassis to the diagonal corner. This will work even if the base isn't
     *         a square.
     */
    public Length getDiagonal() {
        double length = trackLength.get(Unit.BASE_UNIT);
        double width = trackWidth.get(Unit.BASE_UNIT);
        double hypotenuse = Math.sqrt(Math.pow(length, 2) + Math.pow(width, 2));
        return new Length(hypotenuse, Unit.BASE_UNIT);
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof ChassisDimension) {
            ChassisDimension cd = (ChassisDimension) o;
            return cd.getLength().equals(getLength()) && cd.getWidth().equals(getWidth());
        } else {
            return false;
        }
    }

    @Override
    public String toString() {
        return getLength() + " x " + getWidth();
    }
}