/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * An abstract length that can be retrieved in any units. This class makes use
 * of {@link Unit}, which was designed specifically to provide units of length
 * for use with this class. {@code Length} exists to provide an easy mechanism
 * for allowing users to specify physical lengths in any units they want,
 * because it allows the length to be retrieved in any units required, no matter
 * what units it ws instantiated with. For example, if swerve module
 * implementation specifies that the diameter of a wheel is 4 inches, the module
 * method that calculates the linear velocity in meters per second can easily
 * get the wheel diameter in meters without having to do any extra work.
 * 
 * <p>
 * Length objects are immutable, but can be operated in very clean an efficient
 * ways.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 2.1.0
 * @since 1.3.0
 */
@Loggable
public class Length {

    /**
     * A representation of a unit with conversion factors. This primarily intended
     * for use with {@link Length}, which represents arbitrary-unit lengths. This
     * class really makes no sense outside of {@link Length}. Units represented here
     * are sensible; this class is not intended to represent all the possible units.
     * Units are also based in terms of centimeters, so the conversion factor is
     * always in terms of centimeters. This makes conversions between units simple.
     * 
     * @author Jordan Bancino
     * @version 1.3.0
     * @since 1.3.0
     */
    public static enum Unit {
        /**
         * US Inches &mdash; Approximately 2.54 Centimeters.
         */
        INCHES(2.54),

        /**
         * US Feet &mdash; Approximately 30.48 Centimeters
         */
        FEET(30.48),

        /**
         * Metric Centimeters, the base unit of this class.
         */
        CENTIMETERS(1),

        /**
         * Metric Meters &mdash; Exactly 100 Centimeters.
         */
        METERS(100);

        /**
         * An alias to the base unit, which is centimeters. Programs should use this
         * instead of directly referencing centimeters to preserve compatibility with
         * new versions. I don't know that this will ever change, but if it does,
         * programs that use this will still work as intended.
         */
        public static final Unit BASE_UNIT = CENTIMETERS;

        private double conversionFactor;

        /**
         * Construct a new unit with the given conversion factor.
         * 
         * @param conversionFactor How many centimeters make up this unit.
         */
        private Unit(double conversionFactor) {
            this.conversionFactor = conversionFactor;
        }

        /**
         * Get the conversion factor of this unit. This is how many centimeters are
         * equivalent to this unit.
         *
         * @return The number of centimeters in this unit.
         */
        public double conversionFactor() {
            return conversionFactor;
        }
    }

    @Log(as = "unit", atLevel = LogLevel.IMPORTANT)
    private final Unit baseUnit;

    @Log(as = "value", atLevel = LogLevel.IMPORTANT)
    private final double baseValue;

    /**
     * Create an immutable length object with a given length and unit measure. 1
     * foot could be specified as {@code new Length(1, Unit.FEET)}, or as
     * {@code new Length(12, Unit.INCHES)}. Both represent the same value, and
     * demonstrate proper usage of this constructor.
     *
     * @param value The numerical value of this length in terms of the second
     *              parameter, {@code unit}.
     * @param unit  The unit that the value is in.
     * 
     * @throws IllegalArgumentException If the length is negative. Lengths of zero
     *                                  are allowed.
     */
    public Length(double value, Unit unit) {
        if (value >= 0) {
            this.baseValue = value;
            this.baseUnit = unit;
        } else {
            throw new IllegalArgumentException("Lengths cannot be negative.");
        }
    }

    /**
     * Get this length in the specifed units.
     * 
     * @param unit The unit to retrieve the length in.
     * @return The value of the length in the specified units.
     */
    public double get(Unit unit) {
        return (baseValue * baseUnit.conversionFactor()) / unit.conversionFactor();
    }

    /**
     * Add two lengths.
     * 
     * @param length The length to add.
     * @return The resulting length from the operation.
     */
    public Length plus(Length length) {
        return new Length(baseValue + length.get(baseUnit), baseUnit);
    }

    /**
     * Subtract two lengths.
     * 
     * @param length The length to subract.
     * @return The resulting length from the operation.
     */
    public Length minus(Length length) {
        return new Length(baseValue - length.get(baseUnit), baseUnit);
    }

    /**
     * Multiply two lengths.
     * 
     * @param length The length to multiply.
     * @return The resulting length from the operation.
     */
    public Length times(Length length) {
        return new Length(baseValue * length.get(baseUnit), baseUnit);
    }

    /**
     * Divide two lengths.
     * 
     * @param length The length to divide.
     * @return The resulting length from the operation.
     */
    public Length dividedBy(Length length) {
        return new Length(baseValue / length.get(baseUnit), baseUnit);
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof Length) {
            return ((Length) o).get(Unit.BASE_UNIT) == get(Unit.BASE_UNIT);
        } else {
            return false;
        }
    }

    @Override
    public String toString() {
        return String.format("%.2d %s", baseValue, baseUnit.toString().toLowerCase());
    }
}