/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * A mutable container for a module vector that holds the module angle and speed
 * for a single iteration. This is an important core class for driving a
 * SwerveIO swerve drive. it is required to drive any swerve drive because,
 * through the kinematics API, all {@link SwerveVector}s are converted to
 * individual module vectors. This class represents a <b>module</b> vector; in
 * other words, it describes the motion of a single swerve module only.
 * 
 * <p>
 * The module vector is intended to be used only by SwerveIO, because SwerveIO
 * aims to define a complete API that abstracts everything behind just a few
 * WPILib commands and a subsystem. There should be no case in which you are
 * manually feeding module vectors to a swerve module.
 * </p>
 * 
 * <p>
 * When using this vector, keep in mind that you are assigning a speed in units
 * of percentage output, from -1 to 1, 0 being 0% output, 1 being 100% output in
 * the "normal" direction, and -1 being 100% output in the "backwards"
 * direction. The angle component should be in <b>degrees</b>, ranging from -360
 * to 360, the direction specifying the direction of rotation.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.0
 * @since 1.0.0
 */
@Loggable
public class ModuleVector {

    /**
     * Check that a given speed value is within the acceptable bounds, as described
     * in this documentation.
     * 
     * @param speed The speed to check.
     * @return Whether or not {@code speed} should be accepted as motor controller
     *         input or not.
     */
    public static boolean checkSpeed(double speed) {
        return speed <= 1 && speed >= -1;
    }

    /**
     * Check that a given angle value is within the acceptable bounds, as described
     * in this documentation.
     * 
     * @param angle The angle to check.
     * @return Whether or not {@code angle} should be accepted as a swerve module
     *         angle reference or not.
     */
    public static boolean checkAngle(double angle) {
        return angle <= 360 && angle >= -360;
    }

    @Log(atLevel = LogLevel.IMPORTANT)
    private double speed, angle;

    /**
     * Construct a module vector. This is a convenience constructor for the setter
     * methods, and ensures that a module vector is initialized.
     * 
     * @param angle The desired angle of the swerve module, in degrees from 0 to
     *              360.
     * @param speed The desired speed of the swerve module, in percentage output
     *              from 0 to 1.
     * @throws IllegalArgumentException If the speed or the angle is out of bounds.
     */
    public ModuleVector(double angle, double speed) {
        setAngle(angle);
        setSpeed(speed);
    }

    /**
     * Duplicate a module vector using {@link #ModuleVector(double, double)}. This
     * can be useful when you need to store a vector before passing it down a logic
     * chain where it could possibly be modified.
     * 
     * @param v The vector to duplicate.
     */
    public ModuleVector(ModuleVector v) {
        this(v.getAngle(), v.getSpeed());
    }

    /**
     * Set the speed parameter of this module vector.
     *
     * @param speed The speed parameter, as described above.
     *
     * @return This instance, for chaining purposes.
     */
    public ModuleVector setSpeed(double speed) {
        if (checkSpeed(speed)) {
            this.speed = speed;
        } else {
            throw new IllegalArgumentException("Speed out of bounds: " + speed);
        }
        return this;
    }

    /**
     * Set the angle parameter of this module vecctor.
     *
     * @param angle The angle parameter, as described above.
     *
     * @return This instance, for chaining purposes.
     */
    public ModuleVector setAngle(double angle) {
        if (checkAngle(angle)) {
            this.angle = angle;
        } else {
            throw new IllegalArgumentException("Angle out of bounds: " + angle);
        }
        return this;
    }

    /**
     * Get the speed parameter of this module vector that was set with
     * {@link #setSpeed(double)}.
     * 
     * @return The speed value for this vector.
     */
    public double getSpeed() {
        return speed;
    }

    /**
     * Get the angle parameter of this module vector that was set with
     * {@link #setAngle(double)}.
     * 
     * @return The angle value for this vector.
     */
    public double getAngle() {
        return angle;
    }

    /**
     * Add a swerve vector to this one by summing each component.
     * 
     * @param v The vector to add to this one.
     * @return The resulting vector from adding all the components together.
     */
    public ModuleVector plus(ModuleVector v) {
        return new ModuleVector(getAngle() + v.getAngle(), getSpeed() + v.getSpeed());
    }

    /**
     * Subtract a swerve vector from this one by finding the difference between each
     * component.
     * 
     * @param v The vector to subtract.
     * @return The resulting vector from subtracting all the components.
     */
    public ModuleVector minus(ModuleVector v) {
        return new ModuleVector(getAngle() - v.getAngle(), getSpeed() - v.getSpeed());
    }

    /**
     * Multiply a swerve vector by this one by finding the product of each
     * component.
     * 
     * @param v The vector to multiply by.
     * @return The resulting vector from multiplying all the components.
     */
    public ModuleVector times(ModuleVector v) {
        return new ModuleVector(getAngle() * v.getAngle(), getSpeed() * v.getSpeed());
    }

    /**
     * Divide this swerve vector by another one by dividing each component.
     * 
     * @param v The vector to divide by.
     * @return The resulting vector from dividing all the components.
     */
    public ModuleVector dividedBy(ModuleVector v) {
        return new ModuleVector(getAngle() / v.getAngle(), getSpeed() / v.getSpeed());
    }

    @Override
    public String toString() {
        return String.format("ModuleVector(angle: %.2f, speed: %.2f)", angle, speed);
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof ModuleVector) {
            ModuleVector v = (ModuleVector) o;
            return getSpeed() == v.getSpeed() && getAngle() == getAngle();
        } else {
            return false;
        }
    }
}
