/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.SwerveDrive;
import edu.wpi.first.math.kinematics.ChassisSpeeds;

/**
 * A mutable container for a swerve drive vector that holds the forward, strafe,
 * and rotation values for a single iteration. This is the core class for
 * driving a SwerveIO swerve drive. it is required to drive any swerve drive,
 * even autonomously, because WPILib {@code ChassisSpeeds} objects are converted
 * to swerve vectors internally. This class represents the <b>chassis</b>
 * vector; in other words, it describes the motion of the entire swerve drive as
 * one system.
 * 
 * <p>
 * The swerve vector is intended to be used only by SwerveIO, because SwerveIO
 * aims to define a complete API that abstracts everything behind just a few
 * WPILib commands and a subsystem. However, there may be cases, such as in
 * basic autonomous commands, that you may wish to generate your own swerve
 * vectors to pass into the swerve drive.
 * </p>
 * 
 * <p>
 * The is a <b>vector</b>, which means it provides velocity, not position.
 * SwerveIO has basic positioning support, but nothing beyond what is trivial.
 * When using this vector, keep in mind that you are assigning a speed in units
 * of percentage output, from -1 to 1, -1 being 100% output in reverse, 0 being
 * 0% output, and 1 being 100% output in the positive direction.
 * </p>
 * 
 * <p>
 * The directions are defined as follows:
 * </p>
 * 
 * <ul>
 * <li><b>FWD:</b> A positive value moves the swerve drive <b>forward</b></li>
 * <li><b>STR:</b> A positive value moves the swerve drive <b>right</b></li>
 * <li><b>RCW:</b> A positive value rotates the swerve drive
 * <b>clockwise</b></li>
 * </ul>
 * 
 * <p>
 * Negative values do the opposite. So, for example, if you wish to move your
 * swerve drive to the left and downward, you would provide negative numbers for
 * the forward and strafe parameters.
 * </p>
 * <p>
 * All three axes of this vector are fully independent, which means they can be
 * combined in any combination. You should be able to apply all three axes at
 * the same time, and the kinematics provider should handle them. See
 * {@link net.bancino.robotics.swerveio.kinematics} for how swerve vector values
 * are used. Ultimately, while these vectors are passed into the
 * {@link net.bancino.robotics.swerveio.SwerveDrive}, they will eventually make
 * their way down to a kinematics provider, which is responsible for computing
 * individuial module vectors from this overall swerve vector.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 4.2.0
 * @since 1.0.0
 */
@Loggable
public class SwerveVector {

    /**
     * Convert a chassis speed object into a raw swerve vector based on the maximum
     * velocity that a swerve drive can achieve. If the passed vector exceeds the
     * limits of this swerve drive, driving it with this vector will result in a
     * full-throttle.
     *
     * @param v      A WPILib ChassisSpeeds vector that represents the velocity to
     *               apply to this swerve base. See
     *               {@link net.bancino.robotics.swerveio.command.PathweaverSwerveDrive}.
     * @param swerve A swerve drive. This is required to obtain the maximum speed
     *               angular velocity, so that the ChassisSpeeds can be scaled to
     *               the specific swerve.
     * @return A swerve vector designed to drive the specified swerve drive at the
     *         speeds given by the ChassisSpeeds object.
     */
    public static SwerveVector fromChassisSpeeds(ChassisSpeeds v, SwerveDrive swerve) {
        double maxMPS = swerve.getChassisMaxSpeed();
        SwerveVector swerveVector = new SwerveVector(v.vxMetersPerSecond / maxMPS, v.vyMetersPerSecond / maxMPS,
                v.omegaRadiansPerSecond / swerve.getChassisMaxOmega());
        return swerveVector;
    }

    /**
     * A convenience method that simply calls
     * {@link #convertToChassisSpeeds(SwerveDrive)} on the passed vector. This
     * method exists solely to remain consistent with
     * {@link #fromChassisSpeeds(ChassisSpeeds, SwerveDrive)}. It might be useful
     * for querying the absolute velocity in meters per second or radians per second
     * that a particular swerve vector will cause the swerve drive to go.
     * 
     * @param v      The swerve vector to convert to a ChassisSpeeds object for the
     *               given swerve drive.
     * @param swerve The swerve drive object; used to properly scale the resulting
     *               speed object.
     * @return See {@link #convertToChassisSpeeds(SwerveDrive)}
     */
    public static ChassisSpeeds toChassisSpeeds(SwerveVector v, SwerveDrive swerve) {
        return v.convertToChassisSpeeds(swerve);
    }

    @Log(atLevel = LogLevel.IMPORTANT)
    private double fwd, str, rcw;

    /**
     * Construct a swerve vector. No checks are done on the passed parameters, so it
     * is possible to overrun the boundaries of the swerve drive. However, WPILib's
     * motor controllers&mdash;where these values will ultimately end up&mdash; will
     * not overrun the motors, so any value that is out of bounds will effectively
     * be treated as 100%.
     *
     * <p>
     * This constructor is just a convenience method for calling all the setter
     * methods, and prevent uninitialized swerve vectors.
     * </p>
     * 
     * @param fwd The forward (Y) velocity for this vector.
     * @param str The strafe (X) velocity for this vector.
     * @param rcw The rotation (Z) velocity for this vector.
     */
    public SwerveVector(double fwd, double str, double rcw) {
        setFwd(fwd);
        setStr(str);
        setRcw(rcw);
    }

    /**
     * Duplicate an existing swerve vector using
     * {@link #SwerveVector(double, double, double)}. This can be useful when you
     * need to store a vector before passing it down a logic chain where it could
     * possibly be modified.
     * 
     * @param v The vector to duplicate.
     */
    public SwerveVector(SwerveVector v) {
        this(v.getFwd(), v.getStr(), v.getRcw());
    }

    /**
     * Set the forward parameter of this swerve vector.
     *
     * @param fwd The forward parameter, as described above.
     *
     * @return This instance, for chaining purposes.
     */
    public SwerveVector setFwd(double fwd) {
        this.fwd = fwd;
        return this;
    }

    /**
     * Set the strafe parameter of this swerve vector.
     *
     * @param str The strafe parameter, as described above.
     *
     * @return This instance, for chaining purposes.
     */
    public SwerveVector setStr(double str) {
        this.str = str;
        return this;
    }

    /**
     * Set the rotational parameter of this swerve vector.
     *
     * @param rcw The rotational parameter, as described above.
     *
     * @return This instance, for chaining purposes.
     */
    public SwerveVector setRcw(double rcw) {
        this.rcw = rcw;
        return this;
    }

    /**
     * Get the forward parameter of this swerve vector that was set with
     * {@link #setFwd(double)}.
     * 
     * @return The fwd value for this vector.
     */
    public double getFwd() {
        return fwd;
    }

    /**
     * Get the strafe parameter of this swerve vector that set with
     * {@link #setStr(double)}.
     * 
     * @return The str value for this vector.
     */
    public double getStr() {
        return str;
    }

    /**
     * Get the angular parameter of this swerve vector that set with
     * {@link #setRcw(double)}.
     * 
     * @return The rcw value for this vector.
     */
    public double getRcw() {
        return rcw;
    }

    /**
     * Convert this vector into a chassis speed vector using a specific swerve
     * drive, because the absolute speed of the chassis depends on the maximum
     * linear and angular velocity of the swerve drive, since this class merely
     * represents percentage of full output.
     * 
     * @param drive The swerve drive to use when converting. This is required
     *              because the maximum speed and angular speed are used in the
     *              calculation.
     * @return A WPILib chassis speed object that represents the swerve vector
     *         provided.
     */
    public ChassisSpeeds convertToChassisSpeeds(SwerveDrive drive) {
        double maxOutputMPS = drive.getChassisMaxSpeed();
        double maxChassisRadiansPerSecond = drive.getChassisMaxOmega();
        return new ChassisSpeeds(getFwd() * maxOutputMPS, getStr() * maxOutputMPS,
                getRcw() * maxChassisRadiansPerSecond);
    }

    /**
     * Add a swerve vector to this one by summing each component.
     * 
     * @param v The vector to add to this one.
     * @return The resulting vector from adding all the components together.
     */
    public SwerveVector plus(SwerveVector v) {
        return new SwerveVector(fwd + v.getFwd(), str + v.getStr(), rcw + v.getRcw());
    }

    /**
     * Subtract a swerve vector from this one by finding the difference between each
     * component.
     * 
     * @param v The vector to subtract.
     * @return The resulting vector from subtracting all the components.
     */
    public SwerveVector minus(SwerveVector v) {
        return new SwerveVector(fwd - v.getFwd(), str - v.getStr(), rcw - v.getRcw());
    }

    /**
     * Multiply a swerve vector by this one by finding the product of each
     * component.
     * 
     * @param v The vector to multiply by.
     * @return The resulting vector from multiplying all the components.
     */
    public SwerveVector times(SwerveVector v) {
        return new SwerveVector(fwd * v.getFwd(), str * v.getStr(), rcw * v.getRcw());
    }

    /**
     * Divide this swerve vector by another one by dividing each component.
     * 
     * @param v The vector to divide by.
     * @return The resulting vector from dividing all the components.
     */
    public SwerveVector dividedBy(SwerveVector v) {
        return new SwerveVector(fwd / v.getFwd(), str / v.getStr(), rcw / v.getRcw());
    }

    @Override
    public String toString() {
        return String.format("SwerveVector(fwd: %.3f, str: %.3f, rcw: %.3f)", getFwd(), getStr(), getRcw());
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof SwerveVector) {
            SwerveVector v = (SwerveVector) o;
            return getFwd() == v.getFwd() && getStr() == v.getStr() && getRcw() == v.getRcw();
        } else {
            return false;
        }
    }
}
