/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.gyro;

/**
 * The gyro representation used by SwerveIO. This is a complete definition of
 * the gyro API. Every gyro that is used with SwerveIO must implement this
 * interface.
 * 
 * <p>
 * The interface itself is fairly simplistic because there are only a few
 * methods and some of them can usually be implemented in a just a few lines of
 * code. However, the output of the methods documented below has strict rules.
 * Pay careful attention to the documentation for each method and be sure to
 * follow the specifications, otherwise things may misbehave in unexpected ways.
 * </p>
 * 
 * @author Jordan Bancino
 * @since 6.0.2
 * @version 2.0.0
 */
public interface Gyro {

    /**
     * Get the gyro's current yaw. Note that this may not always be double
     * precision. It could be an integer as well. In most cases, this should be a
     * raw value straight from the hardware, however in some obscure cases,
     * scaling may need to occur to convert a JNI reading into a degree measure in
     * terms of 360.
     * 
     * <p>
     * This should go from 0 to 360. If the sensor does not return the angle in
     * terms of 0 to 360, then this function should convert it. For instance, if the
     * sensor returns 0 to -180 and 0 to 180, it should be scaled to 360 here. If
     * the gyro angle is continuous&mdash;that is, it doesn't reset at all and just
     * continues in the positive or negative direction&mdash;it should be converted
     * to a 0-360 reading.
     * </p>
     * 
     * @return The yaw angle, in degrees.
     */
    public double getAngle();

    /**
     * Zero the gyro. If the gyro hardware does not support this, a software offset
     * should be performed in this method so that subsequent calls to
     * {@link #getAngle()} are adjusted. This method is not allowed to fail.
     */
    public void zero();
}