/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.gyro;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * Convert a WPILib gyro to a SwerveIO Gyro. This class wraps a WPILib gyro
 * object in an implementation of the SwerveIO gyro class so that it can be used
 * with SwerveIO. This can be used as an easy way to add gyro support for any
 * FRC gyro that implements the WPILib interface. The reason that SwerveIO does
 * not use the WPILib gyro interface is because SwerveIO has strict gyro
 * requirements that the WPILib interface does not specify. For example, the
 * WPILib interface specifies a method that reports an angle measure without any
 * discontinuities. This is fine for some uses, but SwerveIO does not operate on
 * angles less than zero or greater than 360, so SwerveIO must scale continuous
 * values outside of that range down. That's what this class does. It wraps the
 * WPILib continuous angle in a SwerveIO discontinuous angle. It also renames
 * the reset() method specified by WPILib Gyro to {@link #zero()}, which more
 * accurately describes the purpose of the method.
 * 
 * @author Jordan Bancino
 * @version 6.0.2
 * @since 6.0.2
 * 
 */
@Loggable
public class WPILibGyro implements Gyro {

    private final edu.wpi.first.wpilibj.interfaces.Gyro gyro;

    @Log(as = "angle", atLevel = LogLevel.IMPORTANT)
    private double lastAngle;

    /**
     * Construct a new SwerveIO gyro with a WPILib gyro.
     * 
     * @param gyro The WPILib to wrap.
     */
    public WPILibGyro(edu.wpi.first.wpilibj.interfaces.Gyro gyro) {
        if (gyro != null) {
            this.gyro = gyro;
        } else {
            throw new IllegalArgumentException("Gyro cannot be null.");
        }
    }

    @Override
    public double getAngle() {
        double angle = gyro.getAngle() % 360;
        if (angle < 0) {
            angle += 360;
        }
        lastAngle = angle;
        return angle;
    }

    @Override
    public void zero() {
        gyro.reset();
    }
}
