/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.kinematics;

import net.bancino.robotics.swerveio.module.SwerveModule;
import net.bancino.robotics.swerveio.geometry.SwerveVector;
import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.geometry.ChassisDimension;
import net.bancino.robotics.swerveio.geometry.ModuleVector;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

import static java.lang.Math.toDegrees;
import static java.lang.Math.toRadians;
import static java.lang.Math.atan2;
import static java.lang.Math.sin;
import static java.lang.Math.cos;
import static java.lang.Math.sqrt;
import static java.lang.Math.pow;

import java.util.EnumMap;
import java.util.Map;

/**
 * The default kinematics provider for SwerveIO. These kinematics will drive the
 * swerve drive in what's often called "crab mode". This means that all four
 * modules are driven independently by four different vectors.
 * 
 * <p>
 * The exact implementation of this kinematics calculator is beyond the scope of
 * this documentation, but the implementation generally follows the pseudocode
 * contained in popular, web-searchable swerve drive documents.
 * 
 * @author Jordan Bancino
 * @version 4.2.0
 * @since 2.1.0
 */
@Loggable
public class DefaultSwerveKinematics implements SwerveKinematicsProvider {

    @Log(as = "chassisDimensions")
    private ChassisDimension chassis;

    @Log(as = "calculatedVectors", atLevel = LogLevel.IMPORTANT)
    private final Map<SwerveModule.Location, ModuleVector> lastVectors = new EnumMap<>(SwerveModule.Location.class);

    /**
     * Create a default swerve drive kinematics calculator. This will drive the
     * swerve drive using default, sensible behavior.
     * 
     * @param chassis The chassis dimensions to use for these calculations. This is
     *                required to ensure that the vectors are accurate. This class
     *                uses these dimensions solely for the ratio of the width to the
     *                height, but please ensure that the units are correct as well,
     *                because this dimension object is returned by
     *                {@link #getChassisDimensions()}, which is used throughout
     *                SwerveIO.
     */
    public DefaultSwerveKinematics(ChassisDimension chassis) {
        if (chassis != null) {
            this.chassis = chassis;
        } else {
            throw new IllegalArgumentException("ChassisDimensions cannot be null.");
        }
    }

    @Override
    public ChassisDimension getChassisDimensions() {
        return chassis;
    }

    @Override
    public ModuleVector getModuleVector(SwerveModule.Location module, SwerveVector vector, double gyroAngle) {
        double cosAngle = cos(toRadians(gyroAngle));
        double sinAngle = sin(toRadians(gyroAngle));

        double fwd = vector.getFwd();
        double str = vector.getStr();
        double rcw = vector.getRcw();

        // Field-centric drive adjustments
        double modFwd = (fwd * cosAngle) + (str * sinAngle);
        double modStr = (-fwd * sinAngle) + (str * cosAngle);
        fwd = modFwd;
        str = modStr;

        double vX, vY;
        switch (module) {
        case FRONT_RIGHT:
            vX = str + rcw;
            vY = fwd - rcw;
            break;
        case FRONT_LEFT:
            vX = str + rcw;
            vY = fwd + rcw;
            break;
        case REAR_LEFT:
            vX = str - rcw;
            vY = fwd + rcw;
            break;
        case REAR_RIGHT:
            vX = str - rcw;
            vY = fwd - rcw;
            break;
        default: // Should never happen.
            throw new IllegalArgumentException("Unknown swerve module location: " + module);
        }
        double baseLength = chassis.getLength().get(Unit.BASE_UNIT);
        double baseWidth = chassis.getWidth().get(Unit.BASE_UNIT);
        double baseDiagonal = chassis.getDiagonal().get(Unit.BASE_UNIT);

        vX *= baseLength / baseDiagonal;
        vY *= baseWidth / baseDiagonal;

        double angle = toDegrees(atan2(vX, vY));
        double speed = sqrt(pow(vX, 2) + pow(vY, 2));

        ModuleVector moduleVector = new ModuleVector(angle, (speed > 1) ? 1 : speed);
        lastVectors.put(module, moduleVector);
        return moduleVector;
    }
}
