/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.kinematics;

import net.bancino.robotics.swerveio.module.SwerveModule;
import net.bancino.robotics.swerveio.geometry.SwerveVector;
import net.bancino.robotics.swerveio.geometry.ChassisDimension;
import net.bancino.robotics.swerveio.geometry.ModuleVector;

/**
 * The complete definition of the SwerveIO kinematics API. This interface must
 * be implemented by all calculator classes that wish to supply a SwerveIO
 * swerve drive with wheel angles and speeds. As described in the package
 * documentation, the kinematics API is only responsible for performing the
 * theoretical calculations; leave the execution of those calculations to the
 * main swerve drive class.
 * 
 * <p>
 * The method implementations required by this interface should be stateless;
 * they should have no memory of past or future data. Each method should merely
 * calculate the target angle for the given function call.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 2.1.0
 * @since 2.1.0
 */
public interface SwerveKinematicsProvider {

    /**
     * Provide the chassis dimensions of the swerve drive. These dimensions are
     * usually required to effectively drive a swerve drive. Most implementations
     * require the dimensions be provided in the constructor; if that is the case,
     * the dimensions from the constructor can be returned here.
     * 
     * <p>
     * If the kinematics provider does not require the chassis dimensions to
     * operate, this method still <b>must</b> provide accurate chassis dimensions,
     * because the result of this function is used in numerous places throughout
     * SwerveIO.
     * </p>
     * 
     * @return The chassis dimensions that are being used to compute the inverse
     *         kinematics, or can be used in the future to compute angular velocity.
     */
    public ChassisDimension getChassisDimensions();

    /**
     * Compute the module vector for a given swerve module and swerve vector. This
     * method is also field-centric aware; it should adjust the vector based on the
     * provided gyro angle.
     * 
     * <p>
     * See the {@link net.bancino.robotics.swerveio.geometry.ModuleVector}
     * documentation for the module vector requirements.
     * </p>
     * 
     * @param module       The swerve module to perform the computation for.
     * @param swerveVector The swerve drive vector used to calculate the module
     *                     vector.
     * @param gyroAngle    The gyro angle to modify the vector calculations with.
     *                     This value will always come from
     *                     {@link net.bancino.robotics.swerveio.gyro.Gyro#getAngle()},
     *                     so be sure to check the documentation for that method for
     *                     the exact standards that this method should adhere to.
     *                     You can reasonably assume that the output of that
     *                     function is correct, and if it isn't, it isn't your
     *                     problem here, it's a gyro implementation problem. It is
     *                     not the job of the kinematics provider to correct invalid
     *                     gyro input.
     * @return A valid module vector for driving the provided swerve module.
     */
    public ModuleVector getModuleVector(SwerveModule.Location module, SwerveVector swerveVector, double gyroAngle);

    /**
     * Compute the module vector for a given swerve module and swerve vector. The
     * default implementation of this function calls
     * {@link #getModuleVector(Location, SwerveVector, double)} with a gyro
     * angle of zero. For most cases, this should behavior should not need to be
     * overridden.
     * 
     * <p>
     * See the {@link net.bancino.robotics.swerveio.geometry.ModuleVector}
     * documentation for the module vector requirements.
     * </p>
     * 
     * @param module       The swerve module to perform the computation for.
     * @param swerveVector The swerve drive vector used to calculate the module
     *                     vector.
     * @return A valid module vector for driving the provided swerve module.
     */
    public default ModuleVector getModuleVector(SwerveModule.Location module, SwerveVector swerveVector) {
        return getModuleVector(module, swerveVector, 0);
    }
}