/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.log;

import java.util.List;

import edu.wpi.first.wpilibj.smartdashboard.SmartDashboard;
import net.bancino.log.LogEntry;
import net.bancino.log.LogHandler;

/**
 * Log to the smart dashboard. This class is used with {@link RobotLogger} to
 * log the state of the swerve drive to the dashboard in real time.
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class DashboardLog implements LogHandler {

    private final String rootKey;

    /**
     * Construct a new dashboard log with the default root key, "SwerveIO".
     */
    public DashboardLog() {
        this("SwerveIO");
    }

    /**
     * Construct a new dashboard log with a custom root key.
     * @param rootKey The root key that all dashboard entries are placed under.
     */
    public DashboardLog(String rootKey) {
        this.rootKey = rootKey;
    }

    @Override
    public void accept(LogEntry entry) {
        recursiveLog(rootKey, entry);
        SmartDashboard.updateValues();
    }

    private void recursiveLog(String parentKey, LogEntry entry) {
        if (entry != null) {
            parentKey += "/" + entry.getName();
            switch (entry.getType()) {
            case LIST:
                List<LogEntry> list = entry.getAsList();
                Object testVal = list.get(0).getValue();
                if (testVal instanceof Number) {
                    double[] arr = new double[list.size()];
                    for (int i = 0; i < list.size(); i++) {
                        arr[i] = ((Number) list.get(i).getValue()).doubleValue();
                    }
                    SmartDashboard.putNumberArray(parentKey, arr);
                } else if (testVal instanceof Boolean) {
                    boolean[] arr = new boolean[list.size()];
                    for (int i = 0; i < list.size(); i++) {
                        arr[i] = (boolean) list.get(i).getValue();
                    }
                    SmartDashboard.putBooleanArray(parentKey, arr);
                } else {
                    String[] arr = new String[list.size()];
                    for (int i = 0; i < list.size(); i++) {
                        arr[i] = list.get(i).getValue().toString();
                    }
                    SmartDashboard.putStringArray(parentKey, arr);
                }
                break;
            case NODE:
                for (LogEntry child : entry.getChildren()) {
                    recursiveLog(parentKey, child);
                }
                break;
            case VALUE:
                Object val = entry.getValue();
                if (val != null) {
                    if (val instanceof Boolean) {
                        SmartDashboard.putBoolean(parentKey, (Boolean) val);
                    } else if (val instanceof Number) {
                        SmartDashboard.putNumber(parentKey, ((Number) val).doubleValue());
                    } else {
                        SmartDashboard.putString(parentKey, val.toString());
                    }
                } else {
                    SmartDashboard.putString(parentKey, "(null)");
                }
                break;
            default:
                break;
            }
        }
    }
}
