/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.log;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.log.Logger;

/**
 * A simple logger designed for iterative logging. As robot code runs
 * iteratively, so does the logging method. This class provides a simple wrapper
 * around the logger class to ensure that logging occurs at a set interval.
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
@Loggable
public class RobotLogger extends Logger {
    @Log(atLevel = LogLevel.INFO)
    private long interval = 0;

    @Log
    private long lastTime = 0;

    /**
     * Set the interval to log at. Note that this is a fuzzy interval; the actual
     * period depends on how often {@link #log(Object)} is called.
     * 
     * @param interval The minimum amount of time, in milliseconds, to place between
     *                 each log. Set this to zero to log as fast as possible.
     * @return This logger, for method chaining.
     * 
     * @throws IllegalArgumentException If {@code interval} is less than zero.
     */
    public RobotLogger setInterval(long interval) {
        if (interval >= 0) {
            this.interval = interval;
        } else {
            throw new IllegalArgumentException("Interval " + interval + " is less than zero ms.");
        }
        return this;
    }

    /**
     * Get the logging interval that this logger is using.
     * 
     * @return An interval, in milliseconds, or zero for as fast as possible.
     */
    public long getInterval() {
        return interval;
    }

    @Override
    public void log(Object o) throws UnsupportedOperationException {
        long time = System.currentTimeMillis();
        if (time - lastTime >= interval) {
            super.log(o);
            lastTime = time;
        }
    }
}
