/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.module;

import java.util.function.Consumer;
import java.util.function.BiConsumer;

import edu.wpi.first.wpilibj.motorcontrol.MotorController;
import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;
import net.bancino.robotics.swerveio.encoder.Encoder;
import net.bancino.robotics.swerveio.pid.PIDController;
import net.bancino.robotics.swerveio.pid.DefaultPIDController;
import net.bancino.robotics.swerveio.geometry.Length;

/**
 * A generic swerve module class that provides boilerplate module code. This
 * class is designed to greatly simplify module implementation development
 * because it implements the methods required by {@link SwerveModule} in a way
 * that is reusable.
 * 
 * <p>
 * This class can be instantiated with {@link GenericSwerveModule.Builder}, but
 * this should be for prototyping purposes only. While GenericSwerveModule is
 * complete and usable on its own, the recommended way to use it is to extend
 * it, providing your PID constants and physical property data within the class
 * itself, and declaring a constructor that takes CAN IDs, encoder information,
 * or other fields that aren't hard-coded.
 * </p>
 * <p>
 * Swerve module hardware should have dedicated classes that abstract all the
 * necessary information from the user. A module kit class provides all the
 * constant information for that kit hard-coded into it. Its constructor
 * requires only the CAN IDs and encoder IDs to use.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 6.0.0
 * @version 1.0.0
 */
@Loggable
public class GenericSwerveModule implements SwerveModule {

    /**
     * Build a generic swerve module object. This can be used on its own, but as
     * described in {@link GenericSwerveModule}, it is primarily intended to be
     * instantiated anonymously and passed into the superclass constructor of an
     * extending class.
     * 
     * <p>
     * Because instances of this class are primarily intended to be passed into a
     * superclass constructor, it is designed with complete encapsulation in mind.
     * Various functions support consumers so that lambda expressions can be used
     * for initialization.
     * </p>
     * 
     */
    public static class Builder {
        private MotorController driveMotor = null, pivotMotor = null;
        private Encoder pivotEncoder = null, driveEncoder = null;
        private double driveGearRatio = 0, driveMaxRPM = 0;
        private Length wheelDiameter = null;
        private PIDController drivePid, pivotPid;

        private double pivotGearRatio = 1;
        private double angleOffset = 0;

        /**
         * Set the drive motor for this swerve drive module, optionally providing an
         * initializer function that can initialize the motor before it is added. The
         * initializer function takes the builder instance that this method was called
         * on, as well as the motor that is passed as the first parameter.
         * 
         * @param <S>        The WPILib speed controller type. This is automatically
         *                   inferred from the first parameter provided to this function
         *                   and will be the type that is provided to the
         *                   {@code BiConsumer}, so casting is not necessary.
         * @param driveMotor The the drive motor add to this builder.
         * @param initialize A consumer function that takes this builder, as well as the
         *                   motor that was passed as the first parameter. This is
         *                   primarily intended to be used as the motor initializer.
         *                   There may be times when all initialization needs to occur
         *                   inside the builder, such as when passing an anonymous
         *                   builder into a superclass constructor.
         * @return This builder, for command chaining.
         */
        public <S extends MotorController> Builder setDriveMotor(S driveMotor, BiConsumer<Builder, S> initialize) {
            if (initialize != null) {
                initialize.accept(this, driveMotor);
            }
            this.driveMotor = driveMotor;
            return this;
        }

        /**
         * Set the drive motor for this builder. This does not provide an initialization
         * interface, it merely assigns the drive motor to the builder. For advanced
         * usage, such as performing initialization in the superclass constructor, see
         * {@link #setDriveMotor(MotorController, BiConsumer)}.
         * 
         * @param driveMotor The drive motor to assign to this swerve module builder.
         * @return This builder, for command chaining.
         */
        public Builder setDriveMotor(MotorController driveMotor) {
            return setDriveMotor(driveMotor, null);
        }

        /**
         * Set the pivot motor for this swerve drive module, optionally providing an
         * initializer function that can initialize the motor before it is added. The
         * initializer function takes the builder instance that this method was called
         * on, as well as the motor that is passed as the first parameter.
         * 
         * @param <S>        The WPILib speed controller type. This is automatically
         *                   inferred from the first parameter provided to this function
         *                   and will be the type that is provided to the
         *                   {@code BiConsumer}, so casting is not necessary.
         * @param pivotMotor The the pivot motor add to this builder.
         * @param initialize A consumer function that takes this builder, as well as the
         *                   motor that was passed as the first parameter. This is
         *                   primarily intended to be used as the motor initializer.
         *                   There may be times when all initialization needs to occur
         *                   inside the builder, such as when passing an anonymous
         *                   builder into a superclass constructor.
         * @return This builder, for command chaining.
         */
        public <S extends MotorController> Builder setPivotMotor(S pivotMotor, BiConsumer<Builder, S> initialize) {
            if (initialize != null) {
                initialize.accept(this, pivotMotor);
            }
            this.pivotMotor = pivotMotor;
            return this;
        }

        /**
         * Set the pivot motor for this builder. This does not provide an initialization
         * interface, it merely assigns the drive motor to the builder. For advanced
         * usage, such as performing initialization in the superclass constructor, see
         * {@link #setDriveMotor(MotorController, BiConsumer)}.
         * 
         * @param pivotMotor The drive motor to assign to this swerve module builder.
         * @return This builder, for command chaining.
         */
        public Builder setPivotMotor(MotorController pivotMotor) {
            return setPivotMotor(pivotMotor, null);
        }

        /**
         * Provide the encoder connected to the drive motor.
         * 
         * @param driveEncoder The encoder that monitors the drive motor.
         * @return This builder, for command chaining.
         */
        public Builder setDriveEncoder(Encoder driveEncoder) {
            return setDriveEncoder(driveEncoder, null);
        }

        /**
         * Provide the encoder connected to the drive motor and initialize it.
         *
         * @param <E>          The encoder type. This is automatically inferred from the
         *                     first parameter and allows you to use the initializer
         *                     function without having to cast the object to your
         *                     encoder type.
         * @param driveEncoder The encoder that monitors the drive motor.
         * @param initialize   The function to run on the encoder. It also provides this
         *                     builder function.
         * @return This instance for chaining.
         */
        public <E extends Encoder> Builder setDriveEncoder(E driveEncoder, BiConsumer<Builder, E> initialize) {
            if (initialize != null) {
                initialize.accept(this, driveEncoder);
            }
            this.driveEncoder = driveEncoder;
            return this;
        }

        /**
         * Provide the encoder connected to the pivot motor.
         * 
         * @param pivotEncoder The encoder that monitors the pivot motor.
         * @return This builder, for command chaining.
         */
        public Builder setPivotEncoder(Encoder pivotEncoder) {
            return setPivotEncoder(pivotEncoder, null);
        }

        /**
         * Provide the encoder connected to the pivot motor and initialize it.
         *
         * @param <E>          The encoder type. This is automatically inferred from the
         *                     first parameter and allows you to use the initializer
         *                     function without having to cast the object to your
         *                     encoder type.
         * @param pivotEncoder The encoder that monitors the pivot motor.
         * @param initialize   The function to run on the encoder. It also provides this
         *                     builder function.
         * @return This instance for chaining.
         */
        public <E extends Encoder> Builder setPivotEncoder(E pivotEncoder, BiConsumer<Builder, E> initialize) {
            if (initialize != null) {
                initialize.accept(this, pivotEncoder);
            }
            this.pivotEncoder = pivotEncoder;
            return this;
        }

        /**
         * Set a custom PID controller that is connected to the drive motor. This should
         * be used for hardware-level PID controllers, or PID controllers implemented in
         * hardware vendor libraries. Calls to
         * {@link GenericSwerveModule#getDrivePIDController()} will return this
         * controller, so it can be beneficial to properly set this up if you want your
         * robot code to be able to modify the PID controller properties.
         * 
         * @param drivePid The PID controller attached to the drive motor.
         * @return This builder, for chaining.
         */
        public Builder setDrivePIDController(PIDController drivePid) {
            this.drivePid = drivePid;
            return this;
        }

        /**
         * Set a custom PID controller that is connected to the pivot motor. This should
         * be used for hardware-level PID controllers, or PID controllers implemented in
         * hardware vendor libraries. Calls to
         * {@link GenericSwerveModule#getPivotPIDController()} will return this
         * controller, so it can be beneficial to properly set this up if you want your
         * robot code to be able to modify the PID controller properties.
         * 
         * @param pivotPid The PID controller attached to the pivot motor.
         * @return This builder, for chaining.
         */
        public Builder setPivotPIDController(PIDController pivotPid) {
            this.pivotPid = pivotPid;
            return this;
        }

        /**
         * Set the input/output ratio of the drive motor gearbox. This is used for
         * determining the maximum velocity of this module.
         * 
         * @param driveGearRatio The decimal value of the gear ratio of the gearbox that
         *                       connects the drive motor to the wheel.
         * @return This builder, for command chaining.
         */
        public Builder setDriveGearRatio(double driveGearRatio) {
            this.driveGearRatio = driveGearRatio;
            return this;
        }

        /**
         * Set the input/output ratio of the pivot motor gearbox. This is used for
         * determining how many encoder counts there are per one module revolution, so
         * you can ignore this method if you are using a 1:1 encoder.
         * 
         * @param pivotGearRatio The decimal value of the gear ratio of the gearbox that
         *                       connects the pivot motor to the output rotating
         *                       mechanism that controls the swerve module's rotation.
         * @return This builder, for command chaining.
         */
        public Builder setPivotGearRatio(double pivotGearRatio) {
            this.pivotGearRatio = pivotGearRatio;
            return this;
        }

        /**
         * Set the maximum drive motor rotations per minute. This is used for
         * determining the maximum velocity of this module.
         * 
         * @param driveMaxRPM The decimal value of the maximum drive rotations per
         *                    minute. This is usually discovered by experimentation
         * @return This builder, for command chaining.
         */
        public Builder setDriveMaxRPM(double driveMaxRPM) {
            this.driveMaxRPM = driveMaxRPM;
            return this;
        }

        /**
         * Set the diameter of the drive wheel. This is used for determining the maximum
         * velocity of this module.
         * 
         * @param wheelDiameter The diameter of the wheel, given as a length object so
         *                      you can choose your units.
         * @return This builder, for command chaining.
         */
        public Builder setWheelDiameter(Length wheelDiameter) {
            this.wheelDiameter = wheelDiameter;
            return this;
        }

        /**
         * Get the angle offset that this module should use. This is useful for setting
         * the "forward" position of the swerve module. When using absolute encoders,
         * you're going to get an angle measure that is aligned with the values those
         * encoders output. If your encoder is not zero at the forward position, which
         * is nearly impossible, then this should be the angle value at which the swerve
         * module is considered to be in the "forward" position.
         * 
         * @param angleOffset The angle measure, in degrees, that represents the
         *                    "forward" position of the swerve module. This is used in
         *                    the calculation performed by
         *                    {@link SwerveModule#getAngle()}, so be sure that it is
         *                    accurate for every swerve module, because
         *                    {@link SwerveModule#getAngle()} is adjusted according to
         *                    this value so that it reports a correct reading.
         * @return This builder, for command chaining.
         */
        public Builder setAngleOffset(double angleOffset) {
            this.angleOffset = angleOffset;
            return this;
        }

        /**
         * Build the generic swerve module using an initialization function.
         * 
         * @param initializer A function that accepts the newly-constructed swerve
         *                    module and initializes it or modifies it before the build
         *                    returns. This is useful for applying a custom setup when
         *                    bulding the module.
         * @return An object that represents the swerve module using the provided
         *         information.
         * @throws IllegalArgumentException If there is a configuration error.
         */
        public GenericSwerveModule build(Consumer<GenericSwerveModule> initializer) throws IllegalArgumentException {
            GenericSwerveModule module = new GenericSwerveModule(this);
            if (initializer != null) {
                initializer.accept(module);
            }
            return module;
        }

        /**
         * Build the generic swerve module without using an initializer function.
         * 
         * @return An object that represents the swerve module using the provided
         *         information.
         * @throws IllegalArgumentException If there is a configuration error.
         */
        public GenericSwerveModule build() throws IllegalArgumentException {
            return build(null);
        }
    }

    private MotorController driveMotor, pivotMotor;

    @Log(atLevel = LogLevel.INFO)
    private Encoder pivotEncoder, driveEncoder;

    @Log
    private double driveGearRatio, driveMaxRPM;

    @Log
    private Length wheelDiameter;

    @Log(atLevel = LogLevel.DEBUG)
    private PIDController pivotPid = new DefaultPIDController(), drivePid = new DefaultPIDController();

    @Log
    private double pivotGearRatio;

    @Log
    private double angleOffset;

    /**
     * Construct a generic swerve module. This can only be used in subclasses. The
     * public way to build a generic swerve module is with the builder class.
     * 
     * @param builder The builder to use to construct this swerve module. This will
     *                most likely be anonymous because the call to this superclass
     *                constructor must be the first line in the constructor, unless
     *                you pass a static builder in.
     */
    protected GenericSwerveModule(Builder builder) {
        if (builder.driveMotor != null) {
            this.driveMotor = builder.driveMotor;
        } else {
            throw new IllegalArgumentException("Drive motor cannot be null.");
        }

        if (builder.pivotMotor != null) {
            this.pivotMotor = builder.pivotMotor;
        } else {
            throw new IllegalArgumentException("Pivot motor cannot be null.");
        }

        if (builder.driveEncoder != null) {
            this.driveEncoder = builder.driveEncoder;
        } else {
            throw new IllegalArgumentException("Drive encoder cannot be null");
        }

        if (builder.pivotEncoder != null) {
            this.pivotEncoder = builder.pivotEncoder;
        } else {
            throw new IllegalArgumentException("Pivot encoder cannot be null.");
        }

        if (builder.driveGearRatio > 0) {
            this.driveGearRatio = builder.driveGearRatio;
        } else {
            throw new IllegalArgumentException("Gear ratio cannot be less than or equal to zero.");
        }

        if (builder.driveMaxRPM > 0) {
            this.driveMaxRPM = builder.driveMaxRPM;
        } else {
            throw new IllegalArgumentException("Maximum RPMs cannot be less than or equal to zero.");
        }

        if (builder.wheelDiameter != null) {
            this.wheelDiameter = builder.wheelDiameter;
        } else {
            throw new IllegalArgumentException("Wheel diameter cannot be null.");
        }

        if (builder.drivePid != null) {
            this.drivePid = builder.drivePid;
        }

        if (builder.pivotPid != null) {
            this.pivotPid = builder.pivotPid;
        } else {
            /*
             * Set sensible defaults for this module, but only if we're using the default
             * PID controller.
             */
            this.pivotPid.setOutputLimits(1);
            this.pivotPid.setP(0.0052);
            this.pivotPid.setI(0.000069);
            this.pivotPid.setD(0);
        }

        if (builder.pivotGearRatio > 0) {
            this.pivotGearRatio = builder.pivotGearRatio;
        } else {
            throw new IllegalArgumentException("Pivot gear ratio must be greater than zero.");
        }

        setAngleOffset(builder.angleOffset);

        /*
         * Don't let the PID loop output anything greater than the limits of the motor.
         */
        pivotPid.setOutputLimits(1);
        drivePid.setOutputLimits(1);

        /* Tell the PID loop what the maximum range of the encoder is. */
        pivotPid.setSetpointRange(pivotEncoder.countsPerRevolution());
    }

    @Override
    public MotorController getDriveMotor() {
        return driveMotor;
    }

    @Override
    public MotorController getPivotMotor() {
        return pivotMotor;
    }

    @Override
    public Encoder getDriveEncoder() {
        return driveEncoder;
    }

    @Override
    public Encoder getPivotEncoder() {
        return pivotEncoder;
    }

    @Override
    public PIDController getDrivePIDController() {
        return drivePid;
    }

    @Override
    public PIDController getPivotPIDController() {
        return pivotPid;
    }

    @Override
    public double getDriveGearRatio() {
        return driveGearRatio;
    }

    @Override
    public double getPivotGearRatio() {
        return pivotGearRatio;
    }

    @Override
    public double getDriveMaxRPM() {
        return driveMaxRPM;
    }

    @Override
    public Length getWheelDiameter() {
        return wheelDiameter;
    }

    @Override
    public double getAngleOffset() {
        return angleOffset;
    }

    @Override
    public void setAngleOffset(double offset) {
        this.angleOffset = offset;
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof SwerveModule) {
            SwerveModule asm = (SwerveModule) o;
            return asm.getDriveMaxRPM() == getDriveMaxRPM() && asm.getDriveGearRatio() == getDriveGearRatio()
                    && asm.getOutputThreshhold() == getOutputThreshhold()
                    && asm.getWheelDiameter() == getWheelDiameter();
        } else {
            return false;
        }
    }

    @Override
    public String toString() {
        return "SwerveModule(Drive: " + driveMotor.getClass().getSimpleName() + ", Pivot: "
                + pivotMotor.getClass().getSimpleName() + ")";
    }
}