/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.module;

import net.bancino.robotics.swerveio.pid.PIDController;
import edu.wpi.first.wpilibj.motorcontrol.MotorController;
import net.bancino.robotics.swerveio.encoder.Encoder;
import net.bancino.robotics.swerveio.geometry.Length;
import net.bancino.robotics.swerveio.geometry.ModuleVector;
import net.bancino.robotics.swerveio.geometry.Length.Unit;

/**
 * The complete definition of a SwerveIO swerve module. Every swerve module
 * implementation must implement this interface. A SwerveIO swerve drive
 * contains a module map of modules that can be driven using the methods
 * described in this interface.
 * 
 * <p>
 * This interface binds all the components of the a single swerve module
 * together. A swerve module consists of the following components, all available
 * through this interface:
 * </p>
 * <ul>
 * <li>Pivot Motor</li>
 * <li>Pivot Encoder</li>
 * <li>Pivot PID controller</li>
 * <li>Drive Motor</li>
 * <li>Drive Encoder</li>
 * <li>Drive PID Controller</li>
 * </ul>
 * <p>
 * This interface is also responsible for providing certain metadata for the
 * swerve module. The following information is required for proper operation of
 * the swerve drive, because it is used in the kinematic calculations:
 * </p>
 * <ul>
 * <li>Drive gear ratio (input/output)</li>
 * <li>Drive output threshhold</li>
 * <li>Wheel diameter</li>
 * <li>Drive motor max speed in rotations per minute (RPM)</li>
 * </ul>
 * <p>
 * These are all used by the default implementation of {@link #getMaxSpeed()}
 * and should be reasonably accurate. For information on how to implement the
 * relevant getter methods, see their individual documentation.
 * </p>
 * <p>
 * This interface is not designed to be implemented directly. See
 * {@link GenericSwerveModule} for a basic implementation that provides
 * boilerplate code.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 6.0.0
 * @since 1.0.0
 */
public interface SwerveModule {

    /**
     * When manually driving the PID loop, we must modify our feedback to output the
     * correct the distance and orientation of PID output.
     * 
     * <p>
     * Normally, one would not modify the feedback, however this is necessary to
     * ensure the PID loop behaves properly, providing the correct outputs in the
     * correct direction, since we have to deal with crossing over 180 and of course
     * 0.
     * </p>
     *
     * @param currentPos               The current position, that is, the feedback.
     * @param targetPos                The setpoint, the desired position.
     * @param countsPerPivotRevolution The number of encoder counts in a single
     *                                 pivot revolution (360 degrees)
     * @return A corrected "feedback" value that should be put into the pivot pid
     *         loop.
     */
    public static double correctPivotFeedback(double currentPos, double targetPos, double countsPerPivotRevolution) {
        /* Set up some convenience variables for processing. */
        double quarterPivot = countsPerPivotRevolution / 4.0;
        double halfPivot = countsPerPivotRevolution / 2.0;

        /*
         * Normally, one would not modify the feedback, however this is necessary to
         * ensure the PID loop behaves properly, providing the correct outputs in the
         * correct direction, since we have to deal with crossing over 180 and of course
         * 0.
         */
        double feedbackMod = currentPos;
        /* Calculate the error between the setpoint and the feedback. */
        double positionDiff = targetPos - currentPos;

        /*
         * Modify the feedback to produce the correct PID output.
         */
        if (targetPos < currentPos) { /* If setpoint is less than feedback. */
            if (positionDiff > -halfPivot) { /* If error is greater than -180 degrees. */
                feedbackMod = currentPos; /* Do not modify the feedback. */
            } else { /* Setpoint is greater than feedback. */
                /*
                 * If setpoint is greater than 90 degrees and the position difference is not
                 * equal to -180
                 */
                if (targetPos > quarterPivot && positionDiff != -halfPivot) {
                    feedbackMod = -1 * (countsPerPivotRevolution + positionDiff);
                } else {
                    feedbackMod = currentPos - countsPerPivotRevolution;
                }
            }
        }

        /* Setpoint is greater than feedback. */
        if (targetPos > currentPos) {
            if (positionDiff > halfPivot) { /* error is greater than 180 */
                feedbackMod = currentPos + countsPerPivotRevolution;
            } else {
                feedbackMod = currentPos;
            }
        }
        return feedbackMod;
    }

    /**
     * An enumeration that represents the four swerve modules. This is used in
     * {@link net.bancino.robotics.swerveio.SwerveDrive}'s module map to map swerve
     * drive module implementations to a human-readable and easily-accessible name.
     * 
     * <p>
     * Because of the nature of swerve drive, it technically doesn't matter which
     * side of the robot is the "front" and which is the "back", but all robots have
     * fronts and backs, so just make your swerve drive matches. The physical
     * locations are important because the kinematics providers generate per-module
     * vectors, so we need some way to differentiate the modules. Referring to
     * modules by their location is the easiest way to refer to them. Some swerve
     * drive libraries literally refer to modules as 1, 2, 3, and 4, but that is
     * unintuitive.
     * </p>
     * 
     * <p>
     * This enumeration can be used by methods that operate on one or more swerve
     * modules, such as
     * {@link net.bancino.robotics.swerveio.SwerveDrive#drive(net.bancino.robotics.swerveio.geometry.SwerveVector)}
     * for example. The constant value represents the module that is being operated
     * on.
     * </p>
     * 
     * @see net.bancino.robotics.swerveio.SwerveDrive#getModule(Location)
     */
    public static enum Location {
        /**
         * The swerve module located in the front left corner of the robot.
         */
        FRONT_LEFT,

        /**
         * The swerve module located in the front right corner of the robot.
         */
        FRONT_RIGHT,

        /**
         * The swerve module located in the rear left corner of the robot.
         */
        REAR_LEFT,

        /**
         * The swerve module located in the rear right corner of the robot.
         */
        REAR_RIGHT
    }

    /**
     * Drive the swerve module using the provide module vector. The module vector
     * provides bounded and checked speed and angle values. See the documentation
     * for it to know how to implement this method, though the default
     * implementation should be fine for the overwhelming majority of cases.
     * 
     * <p>
     * This function can directly pass the speed component of the module vector into
     * the drive motor, but it will most likely need to pass the angle into a PID
     * controller, and then pass the output of the PID controller into the pivot
     * motor.
     * </p>
     * 
     * @param vector The vector to be used to set the motor speeds in this module.
     */
    public default void drive(ModuleVector vector) {
        setSpeed(vector.getSpeed());
        setAngle(vector.getAngle());
    }

    /**
     * Read the pivot encoder and report the angle that this module is currently set
     * to. The default implementation uses the following methods:
     *
     * <ul>
     * <li>{@link #getPivotGearRatio()}</li>
     * <li>{@link #getPivotEncoder()}</li>
     * <li>{@link #getAngleOffset()}</li>
     * </ul>
     *
     * There is no need to override this default implementation because it
     * automatically takes care of scaling the raw encoder reading based on the
     * number of counts per pivot motor shaft revolution, as well as the pivot
     * <i>encoder</i> gear ratio.
     *
     * @return The angle measure, in degrees, that this swerve module is currently
     *         facing. This will be within the interval [0,360).
     */
    public default double getAngle() {
        /* Calculate the actual angle based on the encoder reading. */
        double countsPerPivotRevolution = getPivotEncoder().countsPerRevolution() * getPivotGearRatio();
        double angle = ((360 / countsPerPivotRevolution) * getPivotEncoder().get()) % 360;

        /*
         * If the encoder reading is continuous, the angle will go below zero. This will
         * fix the angle within the interval [0,360)
         */
        if (angle < 0) {
            angle += 360;
        }

        /* Apply the angle offset. */
        double offset = getAngleOffset();
        if (angle - offset < 0) {
            angle += 360;
        }
        angle -= offset;

        return angle;
    }

    /**
     * Set the pivot angle of this swerve module. The default implementation uses
     * the following methods:
     * 
     * <ul>
     * <li>{@link #getAngle()}</li>
     * <li>{@link #getPivotPIDController()}</li>
     * <li>{@link #getPivotMotor()}</li>
     * </ul>
     * 
     * There usually is no need to override this default implementation because it
     * automatically takes care of adjusting the angle reference as needed before
     * passing it into the PID controller, the output of which is passed into the
     * pivot motor.
     * 
     * @param angle The angle measure, in degrees, that this swerve module should be
     *              facing facing. This should be within the interval (-360,360), or
     *              [0,360). Both behave the same.
     * 
     * @throws IllegalArgumentException If {@code angle} is out of bounds.
     */
    public default void setAngle(double angle) {
        if (ModuleVector.checkAngle(angle)) {
            /*
             * If the kinematics provider gives a negative angle, we should convert it to a
             * positive angle that is equivalent, because that's what correctPivotFeedback()
             * expects.
             */
            if (angle < 0) {
                angle += 360;
            }

            /* Calculate the angle feedback and setpoint */
            double setpoint = angle;
            double feedback = getAngle();
            double feedbackMod = correctPivotFeedback(feedback, setpoint, 360);

            /* Pass the feedback and setpoint into the PID controller */
            double outRef = getPivotPIDController().getOutput(feedbackMod, setpoint);

            /* Pass the PID controller output into the pivot motor. */
            getPivotMotor().set(outRef);
        } else {
            throw new IllegalArgumentException("Angle out of bounds: " + angle);
        }
    }

    /**
     * Get the angle offset that this module should use. This is useful for setting
     * the "forward" position of the swerve module. When using absolute encoders,
     * you're going to get an angle measure that is aligned with the values those
     * encoders output. If your encoder is not zero at the forward position, which
     * is nearly impossible, then this should be the angle value at which the swerve
     * module is considered to be in the "forward" position.
     * 
     * @return The angle measure, in degrees, that represents the "forward" position
     *         of the swerve module. This is used in the calculation performed by
     *         {@link #getAngle()}, so be sure that it is accurate for every swerve
     *         module, because {@link #getAngle()} is adjusted according to this
     *         value so that it reports a correct reading.
     */
    public double getAngleOffset();

    /**
     * Set the angle offset that this angle should use. This is useful for setting
     * the "forward" position of the swerve module. When using absolute encoders,
     * you're going to get an angle measure that is aligned with the values those
     * encoders output. If your encoder is not zero at the forward position, which
     * is nearly impossible, then this should be the angle value at which the swerve
     * module is considered to be in the "forward" position.
     * 
     * <p>
     * This is used for adjusting the offset "on the fly" in the sense that the
     * robot program can be running for any duration of time before an offset change
     * is requested. Most likely, this will be for re-calibrating the encoders. It
     * would not be wise to call this method more than once per robot code lifetime.
     * </p>
     * 
     * @param offset The angle measure, in degrees, that represents the "forward"
     *               position of the swerve module. This is used in the calculation
     *               performed by {@link #getAngle()}, so be sure that it is
     *               accurate for every swerve module, because {@link #getAngle()}
     *               is adjusted according to this value so that it reports a
     *               correct reading.
     */
    public void setAngleOffset(double offset);

    /**
     * Get the speed of the drive motor for this swerve module.
     * 
     * @return The speed of the drive motor in percent output. This will be within
     *         the bounds [-1,1].
     */
    public default double getSpeed() {
        return getDriveMotor().get();
    }

    /**
     * Set the speed of the drive motor for this swerve module.
     * 
     * @param speed The speed of the drive motor in percent output. This should be
     *              within the bounds [-1, 1].
     * 
     * @throws IllegalArgumentException If {@code speed} is out of bounds.s
     */
    public default void setSpeed(double speed) {
        if (ModuleVector.checkSpeed(speed)) {
            getDriveMotor().set(speed);
        } else {
            throw new IllegalArgumentException("Speed out of bounds: " + speed);
        }

    }

    /**
     * Get the drive motor associated with this swerve module. This allows you to
     * collect information on it, but you should never modify it's state, as doing
     * so could mess up the operation of the swerve drive.
     * 
     * @return A WPILib SpeedController that represents the drive motor of this
     *         module, never null.
     */
    public MotorController getDriveMotor();

    /**
     * Get the pivot motor associated with this swerve module. This allows you to
     * collect information on it, but you should never modify it's state, as doing
     * so could mess up the operation of the swerve drive.
     * 
     * @return A WPILib SpeedController that represents the pivot motor of this
     *         module, never null.
     */
    public MotorController getPivotMotor();

    /**
     * Retrieve the encoder attached to the pivot motor.
     * 
     * @return The encoder that is monitoring the pivot motor output, never null,
     *         because pivot encoders are required.
     * @see net.bancino.robotics.swerveio.encoder.Encoder
     */
    public Encoder getPivotEncoder();

    /**
     * Retrieve the encoder attached to the drive motor. Since this is not used by
     * SwerveIO directly, module implementations may return null if no encoder is
     * present, but it is common practice to provide an interface for the integrated
     * encoder, if the drive motor is equipped with one.
     * 
     * @return The encoder that is monitoring the drive motor output, or null if
     *         there isn't one.
     * @see net.bancino.robotics.swerveio.encoder.Encoder
     */
    public Encoder getDriveEncoder();

    /**
     * Get the PID controller driving the drive motor. Note that this may not always
     * be in use if open-loop control is in effect.
     * 
     * @return The PID controller associated with this swerve module's drive motor,
     *         or null if there isn't one.
     */
    public PIDController getDrivePIDController();

    /**
     * Get the PID controller driving the pivot motor. Note that this may not always
     * be in use if open-loop control is in effect.
     * 
     * @return The PID controller driving the pivot motor, never null, because a
     *         position loop is used to set the pivot angle.
     */
    public PIDController getPivotPIDController();

    /**
     * The output threshold is the absolute minimum current percentage that a motor
     * controller can run on that produces motion. You'll only ever want to override
     * this if your module misbehaves at extremely low speeds. The output of this
     * method is used to compute the "idle angle" of the pivot motors. A swerve
     * module is considered to be "idling" if it's percent output is below this
     * threshhold.
     * 
     * @return The absolute minimum current percentage. This value is usually
     *         discovered by experimentation. The default is 1% current, as this
     *         seems to be a safe average for FRC motor controllers and motors.
     * @see net.bancino.robotics.swerveio.SwerveDrive#setIdleAngle(double, boolean)
     */
    public default double getOutputThreshhold() {
        return 0.01;
    }

    /**
     * The gear ratio of the <b>drive</b> motor. This is the ratio of the input
     * shaft rotations to the output shaft rotations. For an example, if your drive
     * gearbox outputs 1 revolution of the shaft for every 8 motor revolutions, the
     * gear ratio is 8.
     * 
     * <p>
     * This value is used to compute the maximum velocity of this swerve module.
     * </p>
     *
     * @return The gear ratio used to drive this module, (input / output)
     */
    public double getDriveGearRatio();

    /**
     * The gear ratio of the <b>pivot</b> motor. This is the ratio of the input
     * shaft rotations to the output steering rotations. For an example, if your
     * pivot gearbox outputs 1 complete swerve module revolution for every 6 motor
     * revolutions, the gear ratio is 6.
     * 
     * <p>
     * This value is used to compute the encoder counts per revolution if the
     * encoder is on the input shaft. You can safely use the default value (1) if
     * you are using a 1:1 encoder, even if the actual gear ratio isn't 1:1.
     * </p>
     * 
     * @return The gear ratio of the pivot motor and encoder, (input / output).
     */
    public default double getPivotGearRatio() {
        return 1;
    }

    /**
     * Get the maximum realistic output in rotations per minute that this module's
     * drive motor can accomplish. Note that this is the rotations of the drive
     * motor itself, <b>not</b> the output wheel. In other words, this is rotations
     * per minute of the input shaft to the drive gearbox. This can usually be found
     * on your motor's spec page, or observed with a tachometer.
     * 
     * <p>
     * This value is used to compute the maximum velocity of this swerve module.
     * </p>
     *
     * @return The maximum rotations per minute of the drive motor's shaft.
     */
    public double getDriveMaxRPM();

    /**
     * Get the diameter of the wheel attached to this module. This is used for
     * computing the maximum velocity of this swerve module, and should be found
     * from a specifications sheet, or measured accurately.
     * 
     * @return The diameter of the wheel attached to this swerve module.
     */
    public Length getWheelDiameter();

    /**
     * Get the maximum speed drive speed of this module based on the maximum drive
     * RPMs, the wheel diameter, and the gear ratio. The following functions are
     * used in the default implementation, and should thus be accurate:
     * 
     * <ul>
     * <li>{@link #getDriveMaxRPM()}</li>
     * <li>{@link #getDriveGearRatio()}</li>
     * <li>{@link #getWheelDiameter()}</li>
     * </ul>
     *
     * @return The maximum linear velocity that this module can achieve in meters
     *         per second, using the provided module specifications. There is no
     *         reason to override this method, because as long as the values
     *         provided by the listed functions are accurate, this will accurately
     *         compute the maximum velocity.
     */
    public default double getMaxSpeed() {
        /* Rotations per minute -> rotations per second. */
        double maxInputRPS = getDriveMaxRPM() / 60;
        /* Convert input revolutions to output revolutions. */
        double maxOutputRPS = maxInputRPS / getDriveGearRatio();
        /* Get the circumference of the wheel. */
        double wheelCircumference = getWheelDiameter().get(Unit.METERS) * Math.PI;
        return maxOutputRPS * wheelCircumference;
    }

    /**
     * Stop the entire module instantly. This just calls {@code stopMotor()} on the
     * pivot motor and the drive motor.There is generally no need to override this
     * default implementation.
     */
    public default void stop() {
        getPivotMotor().stopMotor();
        getDriveMotor().stopMotor();
    }
}
