/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.pid;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * The default PID controller implementation for SwerveIO. This controller
 * implements all the PID controller interface functionality and provides some
 * useful features that prevent integral windup and overshoot. It provides sane
 * default behavior that should be easy to tune.
 *
 * @author Jordan Bancino
 * @version 6.0.0
 * @since 1.0.0
 */
@Loggable
public class DefaultPIDController implements PIDController {

	/*
	 * This is the absolute minimum number of slots a SwerveIO PID controller is
	 * required to implement.
	 */
	private static final int SLOT_COUNT = 4;

	@Log
	private double[] p = new double[SLOT_COUNT];

	@Log
	private double[] i = new double[SLOT_COUNT];

	@Log
	private double[] d = new double[SLOT_COUNT];

	@Log
	private double[] f = new double[SLOT_COUNT];

	@Log(atLevel = LogLevel.DEBUG)
	private double[] maxIOutput = new double[SLOT_COUNT];

	@Log(atLevel = LogLevel.DEBUG)
	private double[] maxError = new double[SLOT_COUNT];

	@Log(atLevel = LogLevel.INFO)
	private int slot = 0;

	@Log(atLevel = LogLevel.INFO)
	private double errorSum[] = new double[SLOT_COUNT];

	@Log(atLevel = LogLevel.IMPORTANT)
	private double acceptableError = -1;

	@Log
	private double maxOutput = 0;

	@Log
	private double minOutput = 0;

	@Log
	private double lastActual = 0;

	@Log
	private boolean firstRun = true;

	@Log(atLevel = LogLevel.DEBUG)
	private boolean reversed = false;

	@Log(atLevel = LogLevel.INFO)
	private double outputRampRate = 0;

	@Log(atLevel = LogLevel.DEBUG)
	private double lastOutput = 0;

	@Log
	private double outputFilter = 0;

	@Log
	private double setpointRange = 0;

	/**
	 * Create a new PID controller. This initializes all the controller's gains to
	 * zero.
	 */
	public DefaultPIDController() {
		this(0, 0, 0);
	}

	/**
	 * Create a PID controller. See setP, setI, setD methods for more detailed
	 * parameters.
	 *
	 * @param p Proportional gain. Large if large difference between setpoint and
	 *          target.
	 * @param i Integral gain. Becomes large if setpoint cannot reach target
	 *          quickly.
	 * @param d Derivative gain. Responds quickly to large changes in error. Small
	 *          values prevents P and I terms from causing overshoot.
	 */
	public DefaultPIDController(double p, double i, double d) {
		this(p, i, d, 0);
	}

	/**
	 * Create a DefaultPIDConroller class object. See setP, setI, setD, setF methods
	 * for more detailed parameters.
	 *
	 * @param p Proportional gain. Large if large difference between setpoint and
	 *          target.
	 * @param i Integral gain. Becomes large if setpoint cannot reach target
	 *          quickly.
	 * @param d Derivative gain. Responds quickly to large changes in error. Small
	 *          values prevents P and I terms from causing overshoot.
	 * @param f Feed-forward gain. Open loop "best guess" for the output should be.
	 *          Only useful if setpoint represents a rate.
	 */
	public DefaultPIDController(double p, double i, double d, double f) {
		this.p[slot] = p;
		this.i[slot] = i;
		this.d[slot] = d;
		this.f[slot] = f;
		checkSigns(slot);
	}

	@Override
	public void setP(int slot, double p) {
		this.p[slot] = p;
		checkSigns(slot);
	}

	@Override
	public void setI(int slot, double i) {
		/*
		 * This scales the accumulated error to avoid output errors and ensure a
		 * consistent transition.
		 */
		if (this.i[slot] != 0) {
			errorSum[slot] = errorSum[slot] * this.i[slot] / i;
		}
		if (maxIOutput[slot] != 0) {
			maxError[slot] = maxIOutput[slot] / i;
		}
		this.i[slot] = i;
		checkSigns(slot);
	}

	@Override
	public void setD(int slot, double d) {
		this.d[slot] = d;
		checkSigns(slot);
	}

	@Override
	public void setF(int slot, double f) {
		this.f[slot] = f;
		checkSigns(slot);
	}

	@Override
	public void setPID(int slot, double p, double i, double d) {
		this.p[slot] = p;
		this.d[slot] = d;
		/*
		 * the I term has additional calculations, so we need to use it's specific
		 * method for setting it.
		 */
		setI(slot, i);
		checkSigns(slot);
	}

	@Override
	public void setPID(int slot, double p, double i, double d, double f) {
		setPID(p, i, d);
		this.f[slot] = f;
	}

	@Override
	public void setMaxIOutput(int slot, double maximum) {
		/*
		 * maxError and Izone are similar, but scaled for different purposes. The
		 * maxError is generated for simplifying math, since calculations against the
		 * max error are far more common than changing the I term or Izone.
		 */
		maxIOutput[slot] = maximum;
		if (this.i[slot] != 0) {
			maxError[slot] = maxIOutput[slot] / this.i[slot];
		}
	}

	@Override
	public void setOutputLimits(int slot, double minimum, double maximum) {
		if (maximum < minimum)
			return;
		maxOutput = maximum;
		minOutput = minimum;

		/*
		 * Ensure the bounds of the I term are within the bounds of the allowable output
		 * swing
		 */
		if (maxIOutput[slot] == 0 || maxIOutput[slot] > (maximum - minimum)) {
			setMaxIOutput(slot, maximum - minimum);
		}
	}

	@Override
	public void setAcceptableError(double acceptableError) {
		this.acceptableError = acceptableError;
	}

	@Override
	public double getAcceptableError() {
		return this.acceptableError;
	}

	@Override
	public void setDirection(boolean reversed) {
		this.reversed = reversed;
		/* Apply to all slots. */
		for (int slot = 0; slot < SLOT_COUNT; slot++) {
			checkSigns(slot);
		}
	}

	@Override
	public double getOutput(double actual, double setpoint) {
		double output;
		double Poutput;
		double Ioutput;
		double Doutput;
		double Foutput;

		/* Ramp the setpoint used for calculations if user has opted to do so */
		if (setpointRange != 0) {
			setpoint = constrain(setpoint, actual - setpointRange, actual + setpointRange);
		}

		/* Do the simple parts of the calculations */
		double error = setpoint - actual;

		/*
		 * If the error is within an acceptable amount, act as if there is no error.
		 */
		if (acceptableError > 0 && Math.abs(error) < acceptableError) {
			error = 0;
		}

		/*
		 * Calculate F output. Notice, this depends only on the setpoint, and not the
		 * error.
		 */
		Foutput = f[slot] * setpoint;

		/* Calculate P term */
		Poutput = p[slot] * error;

		/*
		 * If this is our first time running this, we don't actually _have_ a previous
		 * input or output. For sensor, sanely assume it was exactly where it is now.
		 * For last output, we can assume it's the current time-independent outputs.
		 */
		if (firstRun) {
			lastActual = actual;
			lastOutput = Poutput + Foutput;
			firstRun = false;
		}

		/*
		 * Calculate D Term Note, this is negative. This actually "slows" the system if
		 * it's doing the correct thing, and small values helps prevent output spikes
		 * and overshoot
		 */
		Doutput = -d[slot] * (actual - lastActual);
		lastActual = actual;

		/*
		 * The Iterm is more complex. There's several things to factor in to make it
		 * easier to deal with.
		 * 
		 * 1. maxIoutput restricts the amount of output contributed by the Iterm.
		 * 
		 * 2. prevent windup by not increasing errorSum if we're already running against
		 * our max Ioutput
		 * 
		 * 3. prevent windup by not increasing errorSum if output is output=maxOutput
		 */
		Ioutput = i[slot] * errorSum[slot];
		if (maxIOutput[slot] != 0) {
			Ioutput = constrain(Ioutput, -maxIOutput[slot], maxIOutput[slot]);
		}

		/* And, finally, we can just add the terms up */
		output = Foutput + Poutput + Ioutput + Doutput;

		/* Figure out what we're doing with the error. */
		if (minOutput != maxOutput && !bounded(output, minOutput, maxOutput)) {
			errorSum[slot] = error;
			/*
			 * reset the error sum to a sane level Setting to current error ensures a smooth
			 * transition when the P term decreases enough for the I term to start acting
			 * upon the controller From that point the I term will build up as would be
			 * expected
			 */
		} else if (outputRampRate != 0 && !bounded(output, lastOutput - outputRampRate, lastOutput + outputRampRate)) {
			errorSum[slot] = error;
		} else if (maxIOutput[slot] != 0) {
			errorSum[slot] = constrain(errorSum[slot] + error, -maxError[slot], maxError[slot]);
			/*
			 * In addition to output limiting directly, we also want to prevent I term
			 * buildup, so restrict the error directly
			 */
		} else {
			errorSum[slot] += error;
		}

		/* Restrict output to our specified output and ramp limits */
		if (outputRampRate != 0) {
			output = constrain(output, lastOutput - outputRampRate, lastOutput + outputRampRate);
		}
		if (minOutput != maxOutput) {
			output = constrain(output, minOutput, maxOutput);
		}
		if (outputFilter != 0) {
			output = lastOutput * outputFilter + output * (1 - outputFilter);
		}

		lastOutput = output;
		return output;
	}

	@Override
	public void reset() {
		firstRun = true;
		for (int i = 0; i < SLOT_COUNT; i++) {
			errorSum[i] = 0;
		}
	}

	@Override
	public void setOutputRampRate(double rate) {
		outputRampRate = rate;
	}

	@Override
	public void setSetpointRange(double range) {
		setpointRange = range;
	}

	@Override
	public void setOutputFilter(double strength) {
		if (strength == 0 || bounded(strength, 0, 1)) {
			outputFilter = strength;
		}
	}

	/**
	 * Forces a value into a specific range
	 *
	 * @param value input value
	 * @param min   maximum returned value
	 * @param max   minimum value in range
	 * @return Value if it's within provided range, min or max otherwise
	 */
	private double constrain(double value, double min, double max) {
		if (value > max) {
			return max;
		}
		if (value < min) {
			return min;
		}
		return value;
	}

	/**
	 * Test if the value is within the min and max, inclusive
	 *
	 * @param value to test
	 * @param min   Minimum value of range
	 * @param max   Maximum value of range
	 * @return true if value is within range, false otherwise
	 */
	private boolean bounded(double value, double min, double max) {
		/*
		 * Note, this is an inclusive range. This is so tests like
		 * `bounded(constrain(0,0,1),0,1)` will return false. This is more helpful for
		 * determining edge-case behaviour than <= is.
		 */
		return (min < value) && (value < max);
	}

	/**
	 * To operate correctly, all PID parameters require the same sign This should
	 * align with the {@literal}reversed value
	 */
	private void checkSigns(int slot) {
		if (reversed) { /* all values should be below zero */
			if (p[slot] > 0)
				p[slot] *= -1;
			if (i[slot] > 0)
				i[slot] *= -1;
			if (d[slot] > 0)
				d[slot] *= -1;
			if (f[slot] > 0)
				f[slot] *= -1;
		} else { /* all values should be above zero */
			if (p[slot] < 0)
				p[slot] *= -1;
			if (i[slot] < 0)
				i[slot] *= -1;
			if (d[slot] < 0)
				d[slot] *= -1;
			if (f[slot] < 0)
				f[slot] *= -1;
		}
	}

	@Override
	public double getPreviousOutput() {
		return lastOutput;
	}

	@Override
	public void setActiveProfile(int slot) {
		if (slot >= 0 && slot < SLOT_COUNT) {
			this.slot = slot;
		} else {
			throw new IllegalArgumentException("Profile slot out of bounds: " + slot);
		}
	}

	@Override
	public int getActiveProfile() {
		return slot;
	}

	@Override
	public int getProfileCount() {
		return SLOT_COUNT;
	}
}
