/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.pid;

/**
 * The PID Controller API used by SwerveIO. This is a complete definition, and
 * every PID controller used with SwerveIO must implement this interface.
 *
 * <p>
 * All PID controllers should generally behave the same way, but this API makes
 * zero gaurantees. There is little reason to stray beyond the trusted and
 * well-tested {@link DefaultPIDController} controller, but SwerveIO is designed
 * to be as extensible as possible, providing interfaces for every pluggable
 * aspect of it.
 * </p>
 *
 * <p>
 * This PID controller interface supports multiple profile slots, which allows
 * you to set multiple gains on a single controller and switch between them at
 * will. PID controllers should provide at least 4 profile slots, but the actual
 * number of supported profiles may vary across implementations.
 * </p>
 *
 * <p>
 * It is <b>not</b> recommended to simply wrap hardware vendor-specific PID
 * controllers such as on the Spark Max or Talon for pivot motors, because these
 * controllers often don't provide the required functionality. Any
 * vendor-specific PID controllers are intended to be used for velocity loops on
 * the drive motors only.
 * </p>
 *
 * @author Jordan Bancino
 * @version 6.0.2
 * @since 1.2.3
 */
public interface PIDController {

  /**
   * Set the proportional gain of the PID loop coefficient. (The motor will
   * correct itself proportional to the offset of the measure compared to its
   * targeted measure.) It's only weakness will be when it treats positive and
   * negative offset equally, otherwise it narrows error to 0.
   *
   * @param p Proportional gain value. Must be positive. This gain is set on the
   *          active profile slot.
   */
  public default void setP(double p) {
    setP(getActiveProfile(), p);
  }

  /**
   * Set the proportional gain of the PID loop coefficient. (The motor will
   * correct itself proportional to the offset of the measure compared to its
   * targeted measure.) It's only weakness will be when it treats positive and
   * negative offset equally, otherwise it narrows error to 0.
   *
   * @param slot The profile slot to set the gain for.
   * @param p    Proportional gain value. Must be positive.
   */
  public void setP(int slot, double p);

  /**
   * Set the integral gain of the PID loop coefficient. (The motor will correct
   * itself based on past errors and integrates the history of offset to narrow
   * error down to 0.) The downside of using a purely integral system is that it's
   * slow to start, as it takes time to accumulate enough information to
   * accurately form it's coefficient.
   *
   * @param i Integral gain value. Must be positive. This gain is set on the
   *          active profile slot.
   */
  public default void setI(double i) {
    setI(getActiveProfile(), i);
  }

  /**
   * Set the integral gain of the PID loop coefficient. (The motor will correct
   * itself based on past errors and integrates the history of offset to narrow
   * error down to 0.) The downside of using a purely integral system is that it's
   * slow to start, as it takes time to accumulate enough information to
   * accurately form it's coefficient.
   * 
   * @param slot The profile slot to set the gain for.
   * @param i    Integral gain value. Must be positive.
   */
  public void setI(int slot, double i);

  /**
   * Set the derivative gain of the PID loop coefficient. (The motor will correct
   * error based on it's rate of change, not seeking to bring the error to 0, but
   * seeking to keep the error that the system is stable.) A derivative loop will
   * never bring your error to 0, but will simply keep your error from growing
   * larger by catching any change and correcting it.
   *
   * @param d Derivative gain value. Must be positive. This gain is set on the
   *          active profile slot.
   */
  public default void setD(double d) {
    setD(getActiveProfile(), d);
  }

  /**
   * Set the derivative gain of the PID loop coefficient. (The motor will correct
   * error based on it's rate of change, not seeking to bring the error to 0, but
   * seeking to keep the error that the system is stable.) A derivative loop will
   * never bring your error to 0, but will simply keep your error from growing
   * larger by catching any change and correcting it.
   *
   * @param slot The profile slot to set the gain for.
   * @param d    Derivative gain value. Must be positive.
   */
  public void setD(int slot, double d);

  /**
   * Sets the priority held in feed-forward augment. In a closed loop system, the
   * feed-forward predicts the outcome of the next output from the motor
   * controllers for better error correcting accuracy.
   *
   * @param f Feed-forward gain value. Must be positive. This gain is set on the
   *          active profile slot.
   */
  public default void setF(double f) {
    setF(getActiveProfile(), f);
  }

  /**
   * Sets the priority held in feed-forward augment. In a closed loop system, the
   * feed-forward predicts the outcome of the next output from the motor
   * controllers for better error correcting accuracy.
   *
   * @param slot The profile slot to set the gain for.
   * @param f    Feed-forward gain value. Must be positive.
   */
  public void setF(int slot, double f);

  /**
   * A convenience function for setting PID values all at once. See {@link #setP},
   * {@link #setI}, and {@link #setD}. This sets the gains on the active profile
   * slot.
   *
   * @param p The value to feed into {@link #setP}.
   * @param i The value to feed into {@link #setI}.
   * @param d The value to feed into {@link #setD}.
   */
  public default void setPID(double p, double i, double d) {
    setPID(getActiveProfile(), p, i, d);
  }

  /**
   * A convenience function for setting PID values all at once. See
   * {@link #setP(int,double)}, {@link #setI(int,double)}, and
   * {@link #setD(int,double)}.
   *
   * @param slot The profile slot to set the gains on.
   * @param p    The value to feed into {@link #setP(int,double)}.
   * @param i    The value to feed into {@link #setI(int,double)}.
   * @param d    The value to feed into {@link #setD(int,double)}.
   */
  public default void setPID(int slot, double p, double i, double d) {
    setP(slot, p);
    setI(slot, i);
    setD(slot, d);
  }

  /**
   * A convenience function for setting PID values all at once. See
   * {@link #setP(int,double)}, {@link #setI(int,double)},
   * {@link #setD(int,double)}, and {@link #setF(int,double)}. This sets the gains
   * on the active profile slot.
   *
   * @param p The value to feed into {@link #setP(int,double)}.
   * @param i The value to feed into {@link #setI(int,double)}.
   * @param d The value to feed into {@link #setD(int,double)}.
   * @param f The value to feed into {@link #setF(int,double)}.
   */
  public default void setPID(double p, double i, double d, double f) {
    setPID(getActiveProfile(), p, i, d, f);
  }

  /**
   * A convenience function for setting PID values all at once. See
   * {@link #setP(int,double)}, {@link #setI(int,double)},
   * {@link #setD(int,double)}, and {@link #setF(int,double)}.
   * 
   * @param slot The profile slot to set the gains on.
   * @param p    The value to feed into {@link #setP(int,double)}.
   * @param i    The value to feed into {@link #setI(int,double)}.
   * @param d    The value to feed into {@link #setD(int,double)}.
   * @param f    The value to feed into {@link #setF(int,double)}.
   */
  public default void setPID(int slot, double p, double i, double d, double f) {
    setPID(slot, p, i, d);
    setF(slot, f);
  }

  /**
   * Set an acceptable error. The PID controller should output nothing if the
   * setpoint and feedback error is within +- the provided error.
   * 
   * <p>
   * A simple implementation is to take the absolute value of the difference
   * between the setpoint and the feedback and check that that value is less than
   * this acceptable error.
   * </p>
   * 
   * @param acceptableError The amount of error allowed by this PID controller.
   *                        This value is in the units of the feedback device and
   *                        applies to both sides of the setpoint. For example,
   *                        when running a controller on a gyro, an acceptable
   *                        error value of 0.5 would allow for 0.5 degrees on both
   *                        sides of the setpoint, which means there is a total
   *                        allowable error of 1 degree.
   */
  public void setAcceptableError(double acceptableError);

  /**
   * Get the acceptable error that this PID controller allows.
   * 
   * @return See {@link #setAcceptableError(double)}.
   */
  public double getAcceptableError();

  /**
   * Set the maximum output value contributed by the I component of the system
   * This can be used to prevent large windup issues and make tuning simpler
   *
   * @param slot The profile slot to get the gains for.
   * @param max  Units are the same as the expected output value
   */
  public void setMaxIOutput(int slot, double max);

  /**
   * Set the maximum output value contributed by the I component of the system
   * This can be used to prevent large windup issues and make tuning simpler
   *
   * @param max Units are the same as the expected output value. This is set on
   *            the active profile slot.
   */
  public default void setMaxIOutput(double max) {
    setMaxIOutput(getActiveProfile(), max);
  }

  /**
   * Set the active profile slot on this PID controller. Future calls to methods
   * that don't accept a slot parameter use this slot number.
   *
   * @param slot A profile slot number, greater than or equal to zero.
   *
   * @throws IllegalArgumentException if {@code slot} is out of bounds for the
   *                                  controller. Consult your controller's
   *                                  documentation, but the bounds should be
   *                                  <b>at minimum</b> 4 slots, giving you slots
   *                                  0, 1, 2, and 3.
   */
  public void setActiveProfile(int slot);

  /**
   * Get the active profile slot on this PID controller.
   *
   * @return The active profile slot that this PID controller is pulling
   *         parameters from.
   */
  public int getActiveProfile();

  /**
   * Get the number of profile slots that this PID controller has. It should be at
   * least 4, but can be more depending on the implementation.
   * 
   * @return The number of profiles this PID controller supports.
   */
  public int getProfileCount();

  /**
   * Set the output limits of the pivot PID controller.
   *
   * @param slot The profile slot to set the limit values on.
   * @param min  The minimum value the PID should output.
   * @param max  The maximum value the PID should output.
   */
  public void setOutputLimits(int slot, double min, double max);

  /**
   * Set the output limits of the pivot PID controller.
   *
   * @param slot   The profile slot to set the limit on.
   * @param output Will produce an output limit of -output to +output.
   */
  public default void setOutputLimits(int slot, double output) {
    setOutputLimits(slot, -output, output);
  }

  /**
   * Set the output limits of the pivot PID controller. This limit is applied to
   * the active profile slot.
   *
   * @param output Will produce an output limit of -output to +output.
   */
  public default void setOutputLimits(double output) {
    setOutputLimits(getActiveProfile(), output);
  }

  /**
   * A simple toggle for reversing the PID loop's output. This is applied to all
   * profile slots.
   *
   * @param reversed Whether or not to flip the PID output.
   */
  public void setDirection(boolean reversed);

  /**
   * This is the main function of the PID loop. It performs all the calculations
   * to generate PID output. It uses the actively set profile slot's gains in the
   * calculations, so make sure the profile you want is set with
   * {@link #setActiveProfile(int)}
   *
   * @param feedback The output feedback; your current reference; where you are
   *                 currently.
   * @param setpoint The destination location. This is where you want to be.
   * @return The output of the PID calculation given the input.
   */
  public double getOutput(double feedback, double setpoint);

  /**
   * Get the previous output from the PID loop. This should store whatever the
   * previous call to the output function generates.
   *
   * @return The previous output of the PID calculation.
   */
  public double getPreviousOutput();

  /**
   * Reset the controller, erasing the previous output, feedback, and setpoints,
   * if stored. This may not do anything at all on certain implementations, but is
   * available for implementations to require when the system exits closed-loop
   * control for a period of time and then re-enters it. Optionally, if the active
   * profile slot changes, implementations may want to reset their internal state
   * as well.
   */
  public void reset();

  /**
   * Sets the ramp rate for the output. This is applied to all profile slots.
   *
   * @param rate The ramp rate for the output, in the same units as the output
   *             value. For each scan, the output value will jump no more than +
   *             or - this rate.
   */
  public void setOutputRampRate(double rate);

  /**
   * Set the output range of this PID controller. This effectively scales the
   * controller so it knows the bounds of what it will be fed. This is applied to
   * all profile slots.
   *
   * @param range The setpoint input range. In relation to swerve drive, this will
   *              be the counts per revolution of your pivot encoder, and will
   *              have no limit on the drive motor.
   */
  public void setSetpointRange(double range);

  /**
   * Set a filter on the output to reduce sharp oscillations. <br>
   * 0.1 is likely a sane starting value. Larger values use historical data more
   * heavily, with low values weigh newer data. 0 will disable, filtering, and use
   * only the most recent value. <br>
   * Increasing the filter strength will P and D oscillations, but force larger I
   * values and increase I term overshoot.<br>
   * Uses an exponential wieghted rolling sum filter, according to a simple <br>
   * <code>output*(1-strength)*sum(0..n){output*strength^n}</code> algorithm.
   * 
   * <p>
   * This value is applied to all profile slots.
   * </p>
   *
   * @param strength valid between [0..1), meaning [current output only..
   *                 historical output only)
   */
  public void setOutputFilter(double strength);
}
