/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;

import net.bancino.robotics.swerveio.geometry.Length.Unit;

/**
 * Test the length class.
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class LengthTest {

    private Length testLength1 = new Length(0, Unit.CENTIMETERS);
    private Length testLength2 = new Length(4, Unit.INCHES);
    private Length testLength3 = new Length((double) 1 / 3, Unit.FEET);
    private Length testLength4 = new Length(2, Unit.METERS);
    private Length testLength5 = new Length(8, Unit.INCHES);

    /**
     * Test {@code equals()}.
     */
    @Test
    public void testEquals() {
        assertEquals(testLength2, testLength2);
        assertEquals(testLength2, new Length(4, Unit.INCHES));
        assertEquals(testLength2, testLength3);

        assertNotEquals(testLength1, testLength2);
        assertNotEquals(testLength3, testLength4);
    }

    /**
     * Test that the constructor throws errors in invalid input and accepts valid
     * input.
     */
    @Test
    public void testBounds() {
        assertThrows(IllegalArgumentException.class, () -> new Length(-1, Unit.FEET));
        assertThrows(IllegalArgumentException.class, () -> new Length(1 - 1.000001, Unit.CENTIMETERS));

        assertDoesNotThrow(() -> new Length(1, Unit.FEET));
        assertDoesNotThrow(() -> new Length(0.00001, Unit.CENTIMETERS));
    }

    /**
     * Test various conversions.
     */
    @Test
    public void testConversions() {
        assertEquals(testLength2.get(Unit.INCHES), 4);
        assertEquals(testLength2.get(Unit.FEET), (double) 1 / 3);
        assertEquals(testLength3.get(Unit.INCHES), 4);
        assertEquals(testLength4.get(Unit.CENTIMETERS), 200);
    }

    /**
     * Test the mathematical functions.
     */
    @Test
    public void testMath() {
        assertEquals(testLength1, testLength2.minus(testLength3));
        assertEquals(testLength5, testLength3.plus(testLength2));
        assertEquals(testLength5, testLength2.times(new Length(2, Unit.INCHES)));
        assertEquals(testLength5, testLength2.dividedBy(new Length(0.5, Unit.INCHES)));
    }
}
