/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.geometry;

import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import org.junit.jupiter.api.Test;

/**
 * Test the module vector class.
 * 
 * @author Jordan Bancino
 * @version 6.1.0
 * @since 6.1.0
 */
public class ModuleVectorTest {

    private ModuleVector testVector1 = new ModuleVector(0, 0);
    private ModuleVector testVector2 = new ModuleVector(90, 0.25);
    private ModuleVector testVector3 = new ModuleVector(180, 0.5);

    /**
     * Test that the bounding checks are working properly. The module vector should
     * not allow values outside of the range [-360,+360] for angles and [-1,+1] for
     * speeds.
     */
    @Test
    public void testBounds() {
        assertDoesNotThrow(() -> new ModuleVector(0, 0));
        assertDoesNotThrow(() -> new ModuleVector(90, 0.5));
        assertDoesNotThrow(() -> new ModuleVector(358, 1));
        assertDoesNotThrow(() -> new ModuleVector(-45, -0.4));

        assertThrows(IllegalArgumentException.class, () -> new ModuleVector(-361, 0));
        assertThrows(IllegalArgumentException.class, () -> new ModuleVector(180, -1.001));
        assertThrows(IllegalArgumentException.class, () -> new ModuleVector(365, 1.1));
    }

    /**
     * Test that basic math operators work as intended. 
     */
    @Test
    public void testMath() {
        assertEquals(testVector3, testVector2.plus(testVector2));
        assertEquals(testVector2, testVector3.minus(testVector2));

    }

    /**
     * Test {@code equals()}.
     */
    @Test
    public void testEquals() {
        assertEquals(testVector1, testVector1);
        assertEquals(testVector1, new ModuleVector(0, 0));

        assertNotEquals(testVector1, testVector2);
        assertNotEquals(testVector2, testVector1);
    }
}
