/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

import com.ctre.phoenix.sensors.AbsoluteSensorRange;
import com.ctre.phoenix.sensors.CANCoder;
import com.ctre.phoenix.sensors.SensorInitializationStrategy;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * A wrapper for the CTRE CANCoder sensor. The encoder is automatically
 * configured with the appropriate values for use with SwerveIO.
 * 
 * @author Jordan Bancino
 * @version 5.0.2
 * @since 5.0.2
 */
@Loggable
public class PhoenixCANCoder implements Encoder {

    private final CANCoder encoder;

    @Log(as = "rawValue", atLevel = LogLevel.IMPORTANT)
    private double lastValue;

    /**
     * Construct a new CANCoder at the given CAN ID.
     * 
     * @param canId The CAN ID that this CANCoder is located at.
     */
    public PhoenixCANCoder(int canId) {
        encoder = new CANCoder(canId);

        /* Use absolute positioning 0 -> 360 */
        encoder.configSensorInitializationStrategy(SensorInitializationStrategy.BootToAbsolutePosition);
        encoder.configAbsoluteSensorRange(AbsoluteSensorRange.Unsigned_0_to_360);
        encoder.setPositionToAbsolute();

        /* 
         * Reverse the sensor direction so it counts clockwise is positive. 
         * (Default is counterclockwise is positive).
         */
        encoder.configSensorDirection(true);
    }

    @Override
    public double get() {
        lastValue = encoder.getAbsolutePosition();
        return lastValue;
    }

    @Override
    public double countsPerRevolution() {
        return 360;
    }

}
