/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

import com.ctre.phoenix.motorcontrol.FeedbackDevice;
import com.ctre.phoenix.motorcontrol.can.BaseMotorController;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * A wrapper for a CTRE Motor Controller sensor. This class provides SwerveIO
 * support for any encoder attached to a CTRE motor controller, whether it is
 * the internal brushless motor encoder, or an external attached encoder. Refer
 * to the CTRE Phoenix Java documentation for {@code FeedbackDevice} to see the
 * supported sensors.
 * 
 * <p>
 * Note that only the CTRE MagEncoder, Quad encoders, and analog encoders
 * support {@link #countsPerRevolution()}. Support for others can be added if we
 * know their counts per rev value.
 * </p>
 * 
 * @author Jordan Bancino
 * @version 5.0.1
 * @since 4.2.0
 */
@Loggable
public class PhoenixEncoder implements Encoder {

    private final BaseMotorController motor;
    private final FeedbackDevice feedbackDevice;

    @Log(as = "phoenixPidIdx")
    private final int slot;

    @Log(as = "rawValue", atLevel = LogLevel.IMPORTANT)
    private double lastValue;

    /**
     * Create a new Phoenix encoder with a motor controller and a feedback mode.
     * This constructor uses the default profile slot, which is 0.
     * 
     * @param motor          The motor controller to use.
     * @param feedbackDevice The encoder mode to set on the motor at the given PID
     *                       slot.
     */
    public PhoenixEncoder(BaseMotorController motor, FeedbackDevice feedbackDevice) {
        this(motor, feedbackDevice, 0);
    }

    /**
     * Create a new Phoenix encoder with a motor controller, the feedback mode, and
     * a PID slot.
     * 
     * @param motor          The motor controller to use.
     * @param feedbackDevice The encoder mode to set on the motor at the given PID
     *                       slot.
     * @param slot           The PID slot to fetch encoder values from.
     */
    public PhoenixEncoder(BaseMotorController motor, FeedbackDevice feedbackDevice, int slot) {
        this.motor = motor;

        /* Set the sensor to use on this profile slot. */
        motor.configSelectedFeedbackSensor(feedbackDevice, slot, 0);

        /*
         * Set the indicator light mode (Green == Increasing)
         * 
         * This isn't required for the TalonFX using the integrated encoder, but is
         * useful for other motors and controllers.
         */
        motor.setSensorPhase(true);

        this.feedbackDevice = feedbackDevice;
        this.slot = slot;
    }

    @Override
    public double get() {
        lastValue = motor.getSelectedSensorPosition(slot);
        return lastValue;
    }

    @Override
    public double countsPerRevolution() {
        switch (feedbackDevice) {
            case CTRE_MagEncoder_Absolute:
            case CTRE_MagEncoder_Relative:
            case QuadEncoder:
                return 4096;
            case IntegratedSensor:
                return 2048;
            case Analog:
                return 1024;
            default:
                throw new IllegalArgumentException("Unknown sensor: " + feedbackDevice);
        }
    }
}
