/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.gyro;

import edu.wpi.first.wpilibj.I2C;
import edu.wpi.first.wpilibj.SPI;
import edu.wpi.first.wpilibj.SerialPort;
import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

import com.kauailabs.navx.frc.AHRS;

/**
 * A Kauai Labs FRC NavX Gyro that implements the abstract gyro interface for
 * use with the swerve drive.
 * 
 * @author Jordan Bancino
 * @version 4.1.1
 * @since 2.0.0
 */
@Loggable
public class NavXGyro implements Gyro {

    private final AHRS gyro;

    @Log(atLevel = LogLevel.IMPORTANT)
    private double angle;

    /**
     * Create a gyro object on the given port.
     * 
     * @param port The SPI port that the gyro is on.
     */
    public NavXGyro(SPI.Port port) {
        this(new AHRS(port));
    }

    /**
     * Create a gyro object on the given port.
     * 
     * @param port The I2C port that the gyro is on.
     */
    public NavXGyro(I2C.Port port) {
        this(new AHRS(port));
    }

    /**
     * Create a gyro object on the given port.
     * 
     * @param port The SerialPort port that the gyro is on.
     */
    public NavXGyro(SerialPort.Port port) {
        this(new AHRS(port));
    }

    /**
     * Wrap an existing NavX AHRS object. This can be used if none of the other
     * constructors suit your needs. Because AHRS offers a number of constructors,
     * it doesn't make sense to duplicate them all here. This way, you can
     * instantiate your AHRS and configure it however you want before adding it to
     * SwerveIO.
     * 
     * @param ahrs The AHRS object to use.
     */
    public NavXGyro(AHRS ahrs) {
        if (ahrs != null) {
            gyro = ahrs;
        } else {
            throw new IllegalArgumentException("NavX AHRS cannot be null.");
        }
    }

    @Override
    public double getAngle() {
        double yaw = gyro.getYaw();
        if (yaw < 0) {
            yaw += 360;
        }
        angle = yaw;
        return angle;
    }

    @Override
    public void zero() {
        gyro.zeroYaw();
    }
}
