/*
 * SwerveIO - A versatile open-source FRC swerve drive library.
 *
 * Copyright (C) 2019-2022 Jordan Bancino <jordan@bancino.net>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published with this
 * source code. You should have received a copy of the GNU General Public
 * License with this program. If you did not, contact the distributors of
 * this program and view it here: https://bancino.net/licenses/gpl-3.0.txt
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 */
package net.bancino.robotics.swerveio.encoder;

import com.revrobotics.*;

import net.bancino.log.Log;
import net.bancino.log.LogLevel;
import net.bancino.log.Loggable;

/**
 * A wrapper for the RevRobotics SparkMax encoder interface. This class provides
 * SwerveIO support for any encoder attached to the Spark Max motor controller,
 * whether it is the internal brushless motor encoder, an alternate encoder, or
 * an analog encoder.
 * 
 * 
 * @author Jordan Bancino
 * @version 8.0.0
 * @since 1.0.0
 */
@Loggable
public class SparkMaxEncoder implements Encoder {

    private final Object encoder;

    @Log(as = "rawValue", atLevel = LogLevel.IMPORTANT)
    private double lastValue;

    /**
     * Create a new spark max encoder with a motor controller. This sets up the
     * encoder to read from the internal NEO encoder data port.
     * 
     * @param sparkMax The motor controller to use.
     */
    public SparkMaxEncoder(CANSparkMax sparkMax) {
        encoder = sparkMax.getEncoder();
        if (encoder == null) {
            throw new UnsupportedOperationException("No internal encoder attached to Spark Max.");
        }
    }

    /**
     * Create a new spark max encoder with a motor controller. This sets up the
     * encode to read from an alternate encoder connected to a spark max.
     * 
     * @param sparkMax     The motor controller to use.
     * @param encoderType  The type of encoder that is connected.
     * @param countsPerRev The counts per revolution of the encoder.
     */
    public SparkMaxEncoder(CANSparkMax sparkMax, SparkMaxAlternateEncoder.Type encoderType, int countsPerRev) {
        encoder = sparkMax.getAlternateEncoder(encoderType, countsPerRev);
        if (encoder == null) {
            throw new UnsupportedOperationException("No alternate encoder attached to Spark Max.");
        }
    }

    /**
     * Create a new spark max encoder with a motor controller. This sets up the
     * encoder to read from an analog encoder connected to the spark max.
     * 
     * @param sparkMax   The motor controller to use.
     * @param analogMode The mode of the analog sensor.
     */
    public SparkMaxEncoder(CANSparkMax sparkMax, SparkMaxAnalogSensor.Mode analogMode) {
        encoder = sparkMax.getAnalog(analogMode);
        if (encoder == null) {
            throw new UnsupportedOperationException("No analog encoder attached to Spark Max.");
        }
    }

    @Override
    public double get() {
        if (encoder instanceof RelativeEncoder) {
            lastValue = ((RelativeEncoder) encoder).getPosition();
        } else if (encoder instanceof SparkMaxAnalogSensor) {
            lastValue = ((SparkMaxAnalogSensor) encoder).getPosition();
        } else {
            throw new IllegalStateException("Something happened and I don't know what.");
        }
        return lastValue;
    }

    @Override
    public double countsPerRevolution() {
        if (encoder instanceof RelativeEncoder) {
            return ((RelativeEncoder) encoder).getCountsPerRevolution();
        } else {
            throw new IllegalArgumentException("SparkMaxAnalogSensor does not support counts per revolution.");
        }
    }
}